/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_COLOR_H
#define GLINT_COLOR_H

#include <glint/ColorCommon.h>

#include <cstdint>

namespace glint
{

class Color
{
public:
    Color() noexcept;
    Color(int red, int green, int blue, int alpha = 255);
    Color(float red, float green, float blue, float alpha = 1.0f);
    explicit Color(uint32_t c) noexcept;
    explicit Color(const uint8_t (& c)[4]) noexcept;
    explicit Color(const float (& c)[4]);
    explicit Color(const std::string& name);
    Color(const Color&) = default;
    Color& operator=(const Color&) = default;
    ~Color() = default;

    bool operator<(const Color& rhs) const;    // For use with STL containers
    bool operator==(const Color& rhs) const;   // For use with STL containers
    friend class std::hash<Color>;             // For use with STL containers

    inline int red() const noexcept;
    inline int green() const noexcept;
    inline int blue() const noexcept;
    inline int alpha() const noexcept;

    // Setters return previous value
    int set_red(int r);
    int set_green(int g);
    int set_blue(int b);
    int set_alpha(int a);

    inline float red_f() const noexcept;
    inline float green_f() const noexcept;
    inline float blue_f() const noexcept;
    inline float alpha_f() const noexcept;

    // Setters return previous value
    float set_red_f(float r);
    float set_green_f(float g);
    float set_blue_f(float b);
    float set_alpha_f(float a);

    inline uint32_t uint32() const noexcept;
    uint32_t set_uint32(uint32_t c) noexcept;

    void uint8_v(uint8_t (& c)[4]) const noexcept;
    void set_uint8_v(const uint8_t c[4]) noexcept;
    void set_uint8_v(const uint8_t c[4], uint8_t (& old)[4]) noexcept;

    void float_v(float (& c)[4]) const noexcept;
    void set_float_v(const float c[4]);
    void set_float_v(const float c[4], float (& old)[4]);

private:
    bool set_from_hex(const std::string& name);
    void set_from_SVG_name(const std::string& name);

    uint8_t red_;
    uint8_t green_;
    uint8_t blue_;
    uint8_t alpha_;
};

int
Color::
red() const noexcept
{
    return red_;
}

int
Color::
green() const noexcept
{
    return green_;
}

int
Color::
blue() const noexcept
{
    return blue_;
}

int
Color::
alpha() const noexcept
{
    return alpha_;
}

float
Color::
red_f() const noexcept
{
    return to_float(red_);
}

float
Color::
green_f() const noexcept
{
    return to_float(green_);
}

float
Color::
blue_f() const noexcept
{
    return to_float(blue_);
}

float
Color::
alpha_f() const noexcept
{
    return to_float(alpha_);
}

uint32_t
Color::
uint32() const noexcept
{
    return (alpha_ << 24) | (red_ << 16) | (green_ << 8) | blue_;
}

} // namespace glint

namespace std
{

template<>
class hash<glint::Color>
{
public:
    size_t operator()(const glint::Color &c) const
    {
        return (c.alpha_ << 24) | (c.red_ << 16) | (c.green_ << 8) | c.blue_;
    }
};

}

#endif
