/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_EXCEPTION_H
#define GLINT_EXCEPTION_H

#include <exception>
#include <string>
#include <memory>

namespace glint
{

namespace internal
{
class ExceptionImpl;
}

//
// Base class for all glint exceptions. It provides the methods for exception chaining.
//
class Exception : public std::exception, public std::nested_exception
{
public:
    Exception(const Exception&);
    Exception& operator=(const Exception&);
    virtual ~Exception() noexcept;

    //
    // Print the name of the exception. Derived classes must override this to print *only* the
    // name of the derived exception.
    // If the derived class sets a reason (or contains other details), it should *not* print the reason
    // or other details as part of what().
    //
    virtual const char* what() const noexcept = 0;

    //
    // Return the reason set by the constructor (empty string if none). Derived class should also include
    // any other state information, such as the value of data members or other relevant detail.
    //
    virtual std::string reason() const;

    //
    // Return a string containing the entire chain of exceptions, indented according to nesting depth.
    // The indent parameter controls the amount of indenting per level. Calling this is equivalent
    // to calling to_string(0, "    ").
    // If an exception has a history chain, the chain is shown as well, in oldest-to-newest order.
    //
    virtual std::string to_string(const std::string& indent = "    ") const;

    //
    // Same as the previous method, but allows setting of the indent level. 0 means no indent (outermost level).
    //
    virtual std::string to_string(int indent_level, const std::string& indent) const;

    //
    // Returns an exception_ptr to this exception. The earlier_exception must be a nullptr or an exception_ptr
    // to an exception that was remembered earlier. This allows a sequence of exceptions to be remembered without
    // having to throw them and is useful for example, in shutdown scenarios where any one of a sequence of steps
    // can fail, but we want to continue and try all the following steps and only throw after all of them have been
    // tried. In this case, each step that fails can add itself to the sequence of remembered exceptions, and finally
    // throw something like ShutdownException.
    //
    std::exception_ptr remember(std::exception_ptr earlier_exception);

    //
    // Returns the previous exception, nullptr if none.
    //
    std::exception_ptr get_earlier() const noexcept;

    //
    // Derived exceptions must implement this so the implemention of remember() (provided by this base class)
    // can return an exception_ptr to its own derived exception.
    //
    virtual std::exception_ptr self() const = 0;

protected:
    Exception(const std::shared_ptr<internal::ExceptionImpl>& derived);
    std::shared_ptr<internal::ExceptionImpl> pimpl() const noexcept;     // No need to reimplement in derived

private:
    std::shared_ptr<internal::ExceptionImpl> p_;
};

} // namespace glint

#endif
