/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_HSVCOLOR_H
#define GLINT_HSVCOLOR_H

#include <glint/Color.h>

#include <string>
#include <functional>

namespace glint
{

class HSVColor;

class HSVColor
{
public:
    HSVColor() noexcept;
    HSVColor(int h, int s, int v, int alpha = 255);
    HSVColor(float h, float s, float v, float alpha = 1.0f);
    HSVColor(const HSVColor&) = default;
    HSVColor& operator=(const HSVColor&) = default;
    ~HSVColor() = default;

    bool operator<(const HSVColor& rhs) const;    // For use with STL containers
    bool operator==(const HSVColor& rhs) const;   // For use with STL containers
    friend class std::hash<HSVColor>;             // For use with STL containers

    inline int hue() const noexcept;
    inline int saturation() const noexcept;
    inline int value() const noexcept;
    inline int alpha() const noexcept;

    // Setters return previous value
    int set_hue(int h);
    int set_saturation(int s);
    int set_value(int v);
    int set_alpha(int a);

    inline float hue_f() const noexcept;
    inline float saturation_f() const noexcept;
    inline float value_f() const noexcept;
    inline float alpha_f() const noexcept;

    // Setters return previous value
    float set_hue_f(float h);
    float set_saturation_f(float s);
    float set_value_f(float v);
    float set_alpha_f(float a);

private:
    uint16_t h_;       // Unfortunately, we need 8.5 bits to store 0 - 359
    uint8_t s_;
    uint8_t v_;
    uint8_t alpha_;
};

HSVColor RGB_to_HSV(const Color& c) noexcept;
Color HSV_to_RGB(const HSVColor& c) noexcept;

int
HSVColor::
hue() const noexcept
{
    return h_;
}

int
HSVColor::
saturation() const noexcept
{
    return s_;
}

int
HSVColor::
value() const noexcept
{
    return v_;
}

int
HSVColor::
alpha() const noexcept
{
    return alpha_;
}

float
HSVColor::
hue_f() const noexcept
{
    return h_ / 360.0f;
}

float
HSVColor::
saturation_f() const noexcept
{
    return to_float(s_);
}

float
HSVColor::
value_f() const noexcept
{
    return to_float(v_);
}

float
HSVColor::
alpha_f() const noexcept
{
    return to_float(alpha_);
}

} // namespace glint

namespace std
{

template<>
class hash<glint::HSVColor>
{
public:
    inline size_t operator()(const glint::HSVColor &c) const
    {
        return (c.s_ << 24) + (c.v_ << 16) + (c.alpha_ << 8) + c.h_;
    }
};

}

#endif
