/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_POINT_H
#define GLINT_POINT_H

#include <cmath>
#include <functional>
#include <climits>

namespace glint
{

template<typename> class PointT;

typedef PointT<int> Point;
typedef PointT<float> PointF;
typedef PointT<double> PointD;
typedef PointT<long double> PointLD;

template<typename T>
class PointT
{
public:
    inline PointT() noexcept;

    template<typename U, typename V>
    inline PointT(U x, V y) noexcept;

    template<typename U>
    inline PointT(const PointT<U>& other) noexcept;

    template<typename U>
    inline PointT& operator=(const PointT<U>& rhs) noexcept;

    ~PointT() = default;

    inline T x() const noexcept;
    inline T y() const noexcept;

    template<typename U>
    inline void set_x(U x) noexcept;

    template<typename U>
    inline void set_y(U x) noexcept;

    inline bool is_zero() const noexcept;

    template<typename U>
    inline bool operator==(const PointT<U>& other) const noexcept;

    template<typename U>
    inline bool operator!=(const PointT<U>& other) const noexcept;

    template<typename U>
    inline bool operator<(const PointT<U>& other) const noexcept;

    template<typename U>
    inline bool operator<=(const PointT<U>& other) const noexcept;

    template<typename U>
    inline bool operator>(const PointT<U>& other) const noexcept;

    template<typename U>
    inline bool operator>=(const PointT<U>& other) const noexcept;

    template<typename U>
    inline PointT operator+(const PointT<U>& other) const;

    template<typename U>
    inline PointT operator-(const PointT<U>& other) const;

    template<typename U>
    inline PointT operator*(U factor) const;

    template<typename U>
    inline PointT operator/(U divisor) const;

    template<typename U>
    inline PointT& operator+=(const PointT<U>& other);

    template<typename U>
    inline PointT& operator-=(const PointT<U>& other);

    template<typename U>
    inline PointT& operator*=(U factor);

    template<typename U>
    inline PointT& operator/=(U divisor);

    // Friend is needed so we can implement mixed-mode operations, for example, comparing
    // a Point<int> to a Point<float>.
    template<typename U> friend class PointT;

    friend class std::hash<PointT<T>>;   // For use with STL containers

private:
    T x_;
    T y_;
};

// Non-member operators

template<typename T, typename U>
const PointT<T> operator*(U factor, const PointT<T> p);

// -----------------------------------------------------
// Implementation
// -----------------------------------------------------

// Default constructor

template<typename T>
PointT<T>::
PointT() noexcept
    : x_(0), y_(0)
{
}

// Construct floating-point point

template<typename T>
template<typename U, typename V>
PointT<T>::
PointT(U x, V y) noexcept
    : x_(x), y_(y)
{
}

// Construct int point

template<>
template<typename U, typename V>
PointT<int>::
PointT(U x, V y) noexcept
    : x_(std::round(x)), y_(std::round(y))
{
}

// Copy-construct floating-point point from another point

template<typename T>
template<typename U>
PointT<T>::
PointT(const PointT<U>& other) noexcept
    : x_(other.x_), y_(other.y_)
{
}

// Copy-construct int point from another point

template<>
template<typename U>
PointT<int>::
PointT(const PointT<U>& other) noexcept
    : x_(std::round(other.x_)), y_(std::round(other.y_))
{
}

// Assign to floating-point point from floating-point point

template<typename T>
template<typename U>
PointT<T>&
PointT<T>::
operator=(const PointT<U>& other) noexcept
{
    x_ = other.x_;
    y_ = other.y_;
    return *this;
}

// Assign to int point from floating-point point

template<>
template<typename U>
PointT<int>&
PointT<int>::
operator=(const PointT<U>& other) noexcept
{
    x_ = std::round(other.x_);
    y_ = std::round(other.y_);
    return *this;
}

// Accessors

template<typename T>
T
PointT<T>::
x() const noexcept
{
    return x_;
}

template<typename T>
T
PointT<T>::
y() const noexcept
{
    return y_;
}

// Mutators

// Set x on float point

template<typename T>
template<typename U>
void
PointT<T>::
set_x(U x) noexcept
{
    x_ = x;
}

// Set x on int point

template<>
template<typename U>
void
PointT<int>::
set_x(U x) noexcept
{
    x_ = std::round(x);
}

// Set y on float point

template<typename T>
template<typename U>
void
PointT<T>::
set_y(U y) noexcept
{
    y_ = y;
}

// Set y on int point

template<>
template<typename U>
void
PointT<int>::
set_y(U y) noexcept
{
    y_ = std::round(y);
}

// is_zero

template<typename T>
bool
PointT<T>::
is_zero() const noexcept
{
    return x_ == 0 && y_ == 0;
}

// Comparison operators

// operator==

template<typename T>
template<typename U>
bool
PointT<T>::
operator==(const PointT<U>& other) const noexcept
{
    return (x_ == other.x_ && y_ == other.y_);
}

template<typename T>
template<typename U>
bool
PointT<T>::
operator!=(const PointT<U>& other) const noexcept
{
    return !(*this == other);
}

template<typename T>
template<typename U>
bool
PointT<T>::
operator<(const PointT<U>& other) const noexcept
{
    return (x_ < other.x_ || (x_ == other.x_ && y_ < other.y_));
}

template<typename T>
template<typename U>
bool
PointT<T>::
operator<=(const PointT<U>& other) const noexcept
{
    return !(*this > other);
}

template<typename T>
template<typename U>
bool
PointT<T>::
operator>(const PointT<U>& other) const noexcept
{
    return (x_ > other.x_ || (x_ == other.x_ && y_ > other.y_));
}

template<typename T>
template<typename U>
bool
PointT<T>::
operator>=(const PointT<U>& other) const noexcept
{
    return !(*this < other);
}

// Arithmetic operators

template<typename T>
template<typename U>
PointT<T>
PointT<T>::
operator+(const PointT<U>& other) const
{
    return PointT(x_ + other.x_, y_ + other.y_);
}

template<>
template<typename U>
PointT<int>
PointT<int>::
operator+(const PointT<U>& other) const
{
    return PointT(std::round(x_ + other.x_), std::round(y_ + other.y_));
}

template<typename T>
template<typename U>
PointT<T>
PointT<T>::
operator-(const PointT<U>& other) const
{
    return PointT(x_ - other.x_, y_ - other.y_);
}

template<>
template<typename U>
PointT<int>
PointT<int>::
operator-(const PointT<U>& other) const
{
    return PointT(std::round(x_ - other.x_), std::round(y_ - other.y_));
}

template<typename T>
template<typename U>
PointT<T>
PointT<T>::
operator*(U factor) const
{
    return PointT(x_ * factor, y_ * factor);
}

template<>
template<typename U>
PointT<int>
PointT<int>::
operator*(U factor) const
{
    return PointT(std::round(x_ * factor), std::round(y_ * factor));
}

template<typename T>
template<typename U>
PointT<T>
PointT<T>::
operator/(U divisor) const
{
    return PointT(x_ / divisor, y_ / divisor);
}

template<>
template<typename U>
PointT<int>
PointT<int>::
operator/(U divisor) const
{
    return PointT(std::round(x_ / divisor), std::round(y_ / divisor));
}

// Assignment operators

template<typename T>
template<typename U>
PointT<T>&
PointT<T>::
operator+=(const PointT<U>& other)
{
    return *this = *this + other;
}

template<typename T>
template<typename U>
PointT<T>&
PointT<T>::
operator-=(const PointT<U>& other)
{
    return *this = *this - other;
}

template<typename T>
template<typename U>
PointT<T>&
PointT<T>::
operator*=(U factor)
{
    return *this = *this * factor;
}

template<typename T>
template<typename U>
PointT<T>&
PointT<T>::
operator/=(U divisor)
{
    return *this = *this / divisor;
}

//
// Non-member operators
//

// operator*

template<typename T, typename U>
inline
const PointT<T>
operator*(U factor, const PointT<T> p)
{
    return p * factor;
}

} // namespace glint

namespace std
{

// Hash function for STL unordered containers. The function is agnostic to endianness and
// the sizes of size_t/int, float, double, and long double. To avoid having values
// such as Point(0,1) and Point(1,0) map to the same hash value, we swap the half-words of
// the hash of the y-coordinate before XORing it with the hash of the x-coordinate.

template<>
template<typename T>
class hash<glint::PointT<T>>
{
public:
    size_t operator()(const glint::PointT<T> &p) const
    {
        constexpr int shift = sizeof(size_t) * CHAR_BIT / 2;
        size_t y_hash = std::hash<T>()(p.y_);
        return std::hash<T>()(p.x_) ^ ((y_hash << shift) | (y_hash >> shift));
    }
};

}

#endif
