/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_RUNTIME_H
#define GLINT_RUNTIME_H

#include <glint/OpenGL.h>
#include <glint/util/DefinesPtrs.h>

#include <boost/utility.hpp>

#include <memory>

namespace glint
{

//
// Instance of a glint runtime with its own OpenGL context, X connection, etc.
// Each instance represents exactly on rectangular rendering area.
// A single address space (or thread) can have multiple instances active, for example,
// to render to multiple rendering rectangles.
//

namespace internal
{
class RuntimeImpl;
}

class GraphicsDisplay;

class Runtime : public util::DefinesPtrs<Runtime>, private boost::noncopyable
{
public:
    inline static UPtr create();
    void shutdown();

    int max_display_width() const noexcept;
    int max_display_height() const noexcept;

    GraphicsDisplay* create_display();                                                  // Full-screen
    GraphicsDisplay* create_display(int width, int height, int xpos = 0, int ypos = 0); // Specific size and position
    GraphicsDisplay* display() const;                                                   // Returns display

private:
    static UPtr runtime_create();
    static void check_GL_environment(GLEnv env);

    Runtime();   // Class is final, instantiation only via create()
    ~Runtime() noexcept;

    std::shared_ptr<internal::RuntimeImpl> p_;
    friend class std::default_delete<Runtime>;     // Because destructor is private
};

//
// Trick to ensure that the caller was compiled with the same GL headers (gl.h or gl2.h) as the glint library the
// caller linked with. The inline create() function here passes the value of gl_environment, which is
// defined in OpenGL.h into the glint run time. The value of the same gl_environment constant inside glint may
// different, if glint was compiled using different headers. If the caller's gl_environment does not match
// the library's environment, check_GL_environment throws a LogicException.
//

inline
Runtime::UPtr
Runtime::create()
{
    check_GL_environment(gl_environment);
    return runtime_create();
}

} // namespace glint

#endif
