/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_UNIFORMVARIABLE_H
#define GLINT_UNIFORMVARIABLE_H

#include <glint/ShaderVariable.h>
#include <glint/Types.h>

namespace glint
{

namespace internal
{

class UniformVariableImpl;

} // namespace internal

//
// UniformVariable is the common ancestor of all uniforms. This allows us to polymorphically
// store uniforms, for example, as the look-up value in a map.
//
// UniformVariable provides assignment functionality because shader uniforms are used
// mostly as write-only variables. This saves the caller having to first cast to
// the derived type with a dynamic_cast. As long as the value being assigned matches the
// type of the uniform at run time, the assignment will work. Otherwise, the caller
// gets a std::bad_cast exception.
//
// To read the value of a uniform (used rarely), the caller must down-cast to the concrete type.
//
// Note that for assignment from GLint and vector<GLint>, we have intermediate base classes.
// These cover assignment to IntUniform, IntArrayUniform, Sampler2DUniform,
// Sampler2DArrayUniform, SamplerCubeUniform, and SamplerCubeArrayUniform.
// That's because these uniforms the same underlying type (GLint or IntArray), so we can't
// identify the target type from the overload.
//

class UniformVariable : public ShaderVariable
{
public:
    void set(GLint rhs);
    UniformVariable& operator=(GLint rhs);

    void set(const std::vector<GLint>& rhs);
    UniformVariable& operator=(const std::vector<GLint>& rhs);

    void set(GLfloat rhs);
    UniformVariable& operator=(GLfloat rhs);

    void set(const FloatVec2& rhs);
    UniformVariable& operator=(const FloatVec2& rhs);

    void set(const FloatVec3& rhs);
    UniformVariable& operator=(const FloatVec3& rhs);

    void set(const FloatVec4& rhs);
    UniformVariable& operator=(const FloatVec4& rhs);

    void set(const FloatMat2& rhs);
    UniformVariable& operator=(const FloatMat2& rhs);

    void set(const FloatMat3& rhs);
    UniformVariable& operator=(const FloatMat3& rhs);

    void set(const FloatMat4& rhs);
    UniformVariable& operator=(const FloatMat4& rhs);

    void set(const FloatArray& rhs);
    UniformVariable& operator=(const FloatArray& rhs);

    void set(const FloatVec2Array& rhs);
    UniformVariable& operator=(const FloatVec2Array& rhs);

    void set(const FloatVec3Array& rhs);
    UniformVariable& operator=(const FloatVec3Array& rhs);

    void set(const FloatVec4Array& rhs);
    UniformVariable& operator=(const FloatVec4Array& rhs);

    void set(const FloatMat2Array& rhs);
    UniformVariable& operator=(const FloatMat2Array& rhs);

    void set(const FloatMat3Array& rhs);
    UniformVariable& operator=(const FloatMat3Array& rhs);

    void set(const FloatMat4Array& rhs);
    UniformVariable& operator=(const FloatMat4Array& rhs);

    void set(const IntVec2& rhs);
    UniformVariable& operator=(const IntVec2& rhs);

    void set(const IntVec3& rhs);
    UniformVariable& operator=(const IntVec3& rhs);

    void set(const IntVec4& rhs);
    UniformVariable& operator=(const IntVec4& rhs);

    void set(const IntVec2Array& rhs);
    UniformVariable& operator=(const IntVec2Array& rhs);

    void set(const IntVec3Array& rhs);
    UniformVariable& operator=(const IntVec3Array& rhs);

    void set(const IntVec4Array& rhs);
    UniformVariable& operator=(const IntVec4Array& rhs);

    void set(GLboolean rhs);
    UniformVariable& operator=(GLboolean rhs);

    void set(const BoolVec2& rhs);
    UniformVariable& operator=(const BoolVec2& rhs);

    void set(const BoolVec3& rhs);
    UniformVariable& operator=(const BoolVec3& rhs);

    void set(const BoolVec4& rhs);
    UniformVariable& operator=(const BoolVec4& rhs);

    void set(const BoolArray& rhs);
    UniformVariable& operator=(const BoolArray& rhs);

    void set(const BoolVec2Array& rhs);
    UniformVariable& operator=(const BoolVec2Array& rhs);

    void set(const BoolVec3Array& rhs);
    UniformVariable& operator=(const BoolVec3Array& rhs);

    void set(const BoolVec4Array& rhs);
    UniformVariable& operator=(const BoolVec4Array& rhs);

    virtual ~UniformVariable();

protected:
    UniformVariable(const std::shared_ptr<internal::UniformVariableImpl>& derived);
};

namespace internal
{

class IntUniformVariableBaseImpl;

} // namespace internal

class IntUniformVariableBase : public UniformVariable
{
public:
    GLint get() const;
    void set(GLint rhs);
    operator GLint() const;

    virtual ~IntUniformVariableBase();

protected:
    IntUniformVariableBase(const std::shared_ptr<internal::IntUniformVariableBaseImpl>& derived);
};

namespace internal
{

class IntArrayUniformVariableBaseImpl;

} // namespace internal

class IntArrayUniformVariableBase : public UniformVariable
{
public:
    std::vector<GLint> get() const;
    void set(const std::vector<GLint>& rhs);
    operator std::vector<GLint>() const;

    virtual ~IntArrayUniformVariableBase();

protected:
    IntArrayUniformVariableBase(const std::shared_ptr<internal::IntArrayUniformVariableBaseImpl>& derived);
};

//
// One derived class for each possible type of variable.
//

class FloatUniform : public UniformVariable
{
public:
    FloatUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatUniform();

    GLfloat get() const;
    void set(GLfloat v);

    operator GLfloat() const;
    FloatUniform& operator=(GLfloat rhs);
    FloatUniform& operator=(const FloatUniform& rhs);
};

class FloatVec2Uniform : public UniformVariable
{
public:
    FloatVec2Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatVec2Uniform();

    FloatVec2 get() const;
    void get(GLfloat& x, GLfloat& y) const;
    void set(const FloatVec2& v);
    void set(GLfloat x, GLfloat y);

    operator FloatVec2() const;
    FloatVec2Uniform& operator=(const FloatVec2& rhs);
    FloatVec2Uniform& operator=(const FloatVec2Uniform& rhs);
};

class FloatVec3Uniform : public UniformVariable
{
public:
    FloatVec3Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatVec3Uniform();

    FloatVec3 get() const;
    void get(GLfloat& x, GLfloat& y, GLfloat& z) const;
    void set(const FloatVec3& v);
    void set(GLfloat x, GLfloat y, GLfloat z);

    operator FloatVec3() const;
    FloatVec3Uniform& operator=(const FloatVec3& rhs);
    FloatVec3Uniform& operator=(const FloatVec3Uniform& rhs);
};

class FloatVec4Uniform : public UniformVariable
{
public:
    FloatVec4Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatVec4Uniform();

    FloatVec4 get() const;
    void get(GLfloat& x, GLfloat& y, GLfloat& z, GLfloat& w) const;
    void set(const FloatVec4& v);
    void set(GLfloat x, GLfloat y, GLfloat z, GLfloat w);

    operator FloatVec4() const;
    FloatVec4Uniform& operator=(const FloatVec4& rhs);
    FloatVec4Uniform& operator=(const FloatVec4Uniform& rhs);
};

class FloatMat2Uniform : public UniformVariable
{
public:
    FloatMat2Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatMat2Uniform();

    FloatMat2 get() const;
    void set(const FloatMat2& v);

    operator FloatMat2() const;
    FloatMat2Uniform& operator=(const FloatMat2& rhs);
    FloatMat2Uniform& operator=(const FloatMat2Uniform& rhs);
};

class FloatMat3Uniform : public UniformVariable
{
public:
    FloatMat3Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatMat3Uniform();

    FloatMat3 get() const;
    void set(const FloatMat3& v);

    operator FloatMat3() const;
    FloatMat3Uniform& operator=(const FloatMat3& rhs);
    FloatMat3Uniform& operator=(const FloatMat3Uniform& rhs);
};

class FloatMat4Uniform : public UniformVariable
{
public:
    FloatMat4Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatMat4Uniform();

    FloatMat4 get() const;
    void set(const FloatMat4& v);

    operator FloatMat4() const;
    FloatMat4Uniform& operator=(const FloatMat4& rhs);
    FloatMat4Uniform& operator=(const FloatMat4Uniform& rhs);
};

class FloatArrayUniform : public UniformVariable
{
public:
    FloatArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatArrayUniform();

    FloatArray get() const;
    void set(const FloatArray& v);

    operator FloatArray() const;
    FloatArrayUniform& operator=(const FloatArray& rhs);
    FloatArrayUniform& operator=(const FloatArrayUniform& rhs);
};

class FloatVec2ArrayUniform : public UniformVariable
{
public:
    FloatVec2ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatVec2ArrayUniform();

    FloatVec2Array get() const;
    void set(const FloatVec2Array& v);

    operator FloatVec2Array() const;
    FloatVec2ArrayUniform& operator=(const FloatVec2Array& rhs);
    FloatVec2ArrayUniform& operator=(const FloatVec2ArrayUniform& rhs);
};

class FloatVec3ArrayUniform : public UniformVariable
{
public:
    FloatVec3ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatVec3ArrayUniform();

    FloatVec3Array get() const;
    void set(const FloatVec3Array& v);

    operator FloatVec3Array() const;
    FloatVec3ArrayUniform& operator=(const FloatVec3Array& rhs);
    FloatVec3ArrayUniform& operator=(const FloatVec3ArrayUniform& rhs);
};

class FloatVec4ArrayUniform : public UniformVariable
{
public:
    FloatVec4ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatVec4ArrayUniform();

    FloatVec4Array get() const;
    void set(const FloatVec4Array& v);

    operator FloatVec4Array() const;
    FloatVec4ArrayUniform& operator=(const FloatVec4Array& rhs);
    FloatVec4ArrayUniform& operator=(const FloatVec4ArrayUniform& rhs);
};

class FloatMat2ArrayUniform : public UniformVariable
{
public:
    FloatMat2ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatMat2ArrayUniform();

    FloatMat2Array get() const;
    void set(const FloatMat2Array& v);

    operator FloatMat2Array() const;
    FloatMat2ArrayUniform& operator=(const FloatMat2Array& rhs);
    FloatMat2ArrayUniform& operator=(const FloatMat2ArrayUniform& rhs);
};

class FloatMat3ArrayUniform : public UniformVariable
{
public:
    FloatMat3ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatMat3ArrayUniform();

    FloatMat3Array get() const;
    void set(const FloatMat3Array& v);

    operator FloatMat3Array() const;
    FloatMat3ArrayUniform& operator=(const FloatMat3Array& rhs);
    FloatMat3ArrayUniform& operator=(const FloatMat3ArrayUniform& rhs);
};

class FloatMat4ArrayUniform : public UniformVariable
{
public:
    FloatMat4ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~FloatMat4ArrayUniform();

    FloatMat4Array get() const;
    void set(const FloatMat4Array& v);

    operator FloatMat4Array() const;
    FloatMat4ArrayUniform& operator=(const FloatMat4Array& rhs);
    FloatMat4ArrayUniform& operator=(const FloatMat4ArrayUniform& rhs);
};

class IntUniform : public IntUniformVariableBase
{
public:
    IntUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntUniform();

    IntUniform& operator=(GLint rhs);
    IntUniform& operator=(const IntUniform& rhs);
};

class IntVec2Uniform : public UniformVariable
{
public:
    IntVec2Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntVec2Uniform();

    IntVec2 get() const;
    void get(GLint& x, GLint& y) const;
    void set(const IntVec2& v);
    void set(GLint x, GLint y);

    operator IntVec2() const;
    IntVec2Uniform& operator=(const IntVec2& rhs);
    IntVec2Uniform& operator=(const IntVec2Uniform& rhs);
};

class IntVec3Uniform : public UniformVariable
{
public:
    IntVec3Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntVec3Uniform();

    IntVec3 get() const;
    void get(GLint& x, GLint& y, GLint& z) const;
    void set(const IntVec3& v);
    void set(GLint x, GLint y, GLint z);

    operator IntVec3() const;
    IntVec3Uniform& operator=(const IntVec3& rhs);
    IntVec3Uniform& operator=(const IntVec3Uniform& rhs);
};

class IntVec4Uniform : public UniformVariable
{
public:
    IntVec4Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntVec4Uniform();

    IntVec4 get() const;
    void get(GLint& x, GLint& y, GLint& z, GLint& w) const;
    void set(const IntVec4& v);
    void set(GLint x, GLint y, GLint z, GLint w);

    operator IntVec4() const;
    IntVec4Uniform& operator=(const IntVec4& rhs);
    IntVec4Uniform& operator=(const IntVec4Uniform& rhs);
};

class IntArrayUniform : public IntArrayUniformVariableBase
{
public:
    IntArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntArrayUniform();

    IntArrayUniform& operator=(const IntArray& rhs);
    IntArrayUniform& operator=(const IntArrayUniform& rhs);
};

class IntVec2ArrayUniform : public UniformVariable
{
public:
    IntVec2ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntVec2ArrayUniform();

    IntVec2Array get() const;
    void set(const IntVec2Array& v);

    operator IntVec2Array() const;
    IntVec2ArrayUniform& operator=(const IntVec2Array& rhs);
    IntVec2ArrayUniform& operator=(const IntVec2ArrayUniform& rhs);
};

class IntVec3ArrayUniform : public UniformVariable
{
public:
    IntVec3ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntVec3ArrayUniform();

    IntVec3Array get() const;
    void set(const IntVec3Array& v);

    operator IntVec3Array() const;
    IntVec3ArrayUniform& operator=(const IntVec3Array& rhs);
    IntVec3ArrayUniform& operator=(const IntVec3ArrayUniform& rhs);
};

class IntVec4ArrayUniform : public UniformVariable
{
public:
    IntVec4ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~IntVec4ArrayUniform();

    IntVec4Array get() const;
    void set(const IntVec4Array& v);

    operator IntVec4Array() const;
    IntVec4ArrayUniform& operator=(const IntVec4Array& rhs);
    IntVec4ArrayUniform& operator=(const IntVec4ArrayUniform& rhs);
};

class BoolUniform : public UniformVariable
{
public:
    BoolUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolUniform();

    GLboolean get() const;
    void set(GLboolean v);

    operator GLboolean() const;
    BoolUniform& operator=(GLboolean rhs);
    BoolUniform& operator=(const BoolUniform& rhs);
};

class BoolVec2Uniform : public UniformVariable
{
public:
    BoolVec2Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolVec2Uniform();

    BoolVec2 get() const;
    void get(GLboolean& x, GLboolean& y) const;
    void set(const BoolVec2& v);
    void set(GLboolean x, GLboolean y);

    operator BoolVec2() const;
    BoolVec2Uniform& operator=(const BoolVec2& rhs);
    BoolVec2Uniform& operator=(const BoolVec2Uniform& rhs);
};

class BoolVec3Uniform : public UniformVariable
{
public:
    BoolVec3Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolVec3Uniform();

    BoolVec3 get() const;
    void get(GLboolean& x, GLboolean& y, GLboolean& z) const;
    void set(const BoolVec3& v);
    void set(GLboolean x, GLboolean y, GLboolean z);

    operator BoolVec3() const;
    BoolVec3Uniform& operator=(const BoolVec3& rhs);
    BoolVec3Uniform& operator=(const BoolVec3Uniform& rhs);
};

class BoolVec4Uniform : public UniformVariable
{
public:
    BoolVec4Uniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolVec4Uniform();

    BoolVec4 get() const;
    void get(GLboolean& x, GLboolean& y, GLboolean& z, GLboolean& w) const;
    void set(const BoolVec4& v);
    void set(GLboolean x, GLboolean y, GLboolean z, GLboolean w);

    operator BoolVec4() const;
    BoolVec4Uniform& operator=(const BoolVec4& rhs);
    BoolVec4Uniform& operator=(const BoolVec4Uniform& rhs);
};

class BoolArrayUniform : public UniformVariable
{
public:
    BoolArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolArrayUniform();

    BoolArray get() const;
    void set(const BoolArray& v);

    operator BoolArray() const;
    BoolArrayUniform& operator=(const BoolArray& rhs);
    BoolArrayUniform& operator=(const BoolArrayUniform& rhs);
};

class BoolVec2ArrayUniform : public UniformVariable
{
public:
    BoolVec2ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolVec2ArrayUniform();

    BoolVec2Array get() const;
    void set(const BoolVec2Array& v);

    operator BoolVec2Array() const;
    BoolVec2ArrayUniform& operator=(const BoolVec2Array& rhs);
    BoolVec2ArrayUniform& operator=(const BoolVec2ArrayUniform& rhs);
};

class BoolVec3ArrayUniform : public UniformVariable
{
public:
    BoolVec3ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolVec3ArrayUniform();

    BoolVec3Array get() const;
    void set(const BoolVec3Array& v);

    operator BoolVec3Array() const;
    BoolVec3ArrayUniform& operator=(const BoolVec3Array& rhs);
    BoolVec3ArrayUniform& operator=(const BoolVec3ArrayUniform& rhs);
};

class BoolVec4ArrayUniform : public UniformVariable
{
public:
    BoolVec4ArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~BoolVec4ArrayUniform();

    BoolVec4Array get() const;
    void set(const BoolVec4Array& v);

    operator BoolVec4Array() const;
    BoolVec4ArrayUniform& operator=(const BoolVec4Array& rhs);
    BoolVec4ArrayUniform& operator=(const BoolVec4ArrayUniform& rhs);
};

class Sampler2DUniform : public IntUniformVariableBase
{
public:
    Sampler2DUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~Sampler2DUniform();

    Sampler2DUniform& operator=(GLint rhs);
    Sampler2DUniform& operator=(const Sampler2DUniform& rhs);
};

class Sampler2DArrayUniform : public IntArrayUniformVariableBase
{
public:
    Sampler2DArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~Sampler2DArrayUniform();

    Sampler2DArrayUniform& operator=(const Sampler2DArray& rhs);
    Sampler2DArrayUniform& operator=(const Sampler2DArrayUniform& rhs);
};

class SamplerCubeUniform : public IntUniformVariableBase
{
public:
    SamplerCubeUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~SamplerCubeUniform();

    SamplerCubeUniform& operator=(GLint rhs);
    SamplerCubeUniform& operator=(const SamplerCubeUniform& rhs);
};

class SamplerCubeArrayUniform : public IntArrayUniformVariableBase
{
public:
    SamplerCubeArrayUniform(GLuint prog_id, GLint location, const std::string& name, GLint size);
    virtual ~SamplerCubeArrayUniform();

    SamplerCubeArrayUniform& operator=(const SamplerCubeArray& rhs);
    SamplerCubeArrayUniform& operator=(const SamplerCubeArrayUniform& rhs);
};

} // namespace glint

#endif
