/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_GENV_INTERNAL_ABSTRACTGENV_H
#define GLINT_GENV_INTERNAL_ABSTRACTGENV_H

#include <glint/util/DefinesPtrs.h>

#include <boost/utility.hpp>

namespace glint
{

class GraphicsDisplay;

namespace genv
{

namespace internal
{

//
// AbstractGEnv is an interface. Classes that implement AbstractGEnv each exist in a separate library
// for each graphics environment. For example, we have one implementation of AbstractGEnv for GLX, one for
// EGL, etc. These concrete implementations each live in their own library, such as glint-linux-glx.so,
// glint-linux-egl.so, etc.
//
// The concrete GEnv implementations are expected to initialize their graphics environment in the constructor
// to the point where it is possible create an OpenGL context by calling create_display().
// The shutdown() method is called by the library so the concrete GEnv can reclaim resources.
// Any errors during initialization or finalization throw an exception.
//
// Note that the GEnv is expected to clean up even if shutdown() is not called, in which case, the GEnv
// is expected to do the cleanup from within its destructor. No exceptions will be thrown in that case (but
// failure to finalize correctly can still be logged from the destructor).
//

class AbstractGEnv : public util::DefinesPtrs<AbstractGEnv>, private boost::noncopyable
{
public:
    AbstractGEnv();
    virtual ~AbstractGEnv() noexcept;

    virtual void shutdown() = 0;

    virtual int max_display_width() const = 0;
    virtual int max_display_height() const = 0;

    virtual GraphicsDisplay* create_display(int width, int height, int xpos, int ypos) = 0;

    virtual std::string name() const = 0;
    virtual int major_version() const noexcept = 0; // GLX or EGL major version
    virtual int minor_version() const noexcept = 0; // GLX or EGL minor version

    virtual int gl_major_version() const noexcept = 0; // OpenGL major version
    virtual int gl_minor_version() const noexcept = 0; // OpenGL minor version

    virtual void make_current() = 0;
    virtual void release_current() = 0;
    virtual void swap_buffers() = 0;
};

//
// This entry point is provided by each concrete GEnv. The concrete GEnv creates an instance that derives
// from AbstractGEnv and returns it here.
// This design means that only one concrete GEnv can exist per process (which makes sense, seeing that we cannot
// have, say glX and glW simultaneously).
//
extern AbstractGEnv::UPtr create_concrete_genv();

} // namespace internal

} // namespace genv

} // namespace glint

#endif
