/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#ifndef GLINT_INTERNAL_PROGRAM_IMPL_H
#define GLINT_INTERNAL_PROGRAM_IMPL_H

#include <glint/Program.h>
#include <glint/UniformVariable.h>
#include <glint/util/ResourcePtr.h>

namespace glint
{

namespace internal
{

class VertexShaderImpl;
class FragmentShaderImpl;

class ProgramImpl : private boost::noncopyable
{
public:
    ProgramImpl(const VertexShader& vs, const FragmentShader& fs);
    ~ProgramImpl() noexcept;

    inline GLuint id() const noexcept;
    void use() const;
    bool in_use() const noexcept;
    static GLuint current_id() noexcept;

    inline GLuint vertex_shader_id() const;
    inline GLuint fragment_shader_id() const;

    UniformVariable& uniform(const std::string& name) const;
    const Program::UniformMap& uniforms() const;

private:
    void detach_shader(GLuint shader_id) noexcept;
    void init_uniforms() const;
    std::unique_ptr<UniformVariable> make_variable(GLint id, const std::string& name, GLenum type, GLint size) const;
    void throw_no_such_variable(const std::string& name) const;
    std::string get_log_msg(const std::string& caller) const;

    util::ResourcePtr<GLuint, std::function<void(GLuint)>> id_;       // Program ID
    util::ResourcePtr<GLuint, std::function<void(GLuint)>> vs_id_;    // Vertex shader ID
    util::ResourcePtr<GLuint, std::function<void(GLuint)>> fs_id_;    // Fragment shader ID
    mutable Program::UniformMap uniforms_;                            // Map of active uniforms for the program
    mutable bool initialized_;                                        // For lazy initialization of uniforms_
    mutable Program::UniformMap element_uniforms_;                    // For retrieving array element uniforms

    static GLuint current_prog_;                                      // ID of program last made current with use()
};

GLuint
ProgramImpl::
id() const noexcept
{
    return id_.get();
}

GLuint
ProgramImpl::
vertex_shader_id() const
{
    return vs_id_.get();
}

GLuint
ProgramImpl::
fragment_shader_id() const
{
    return fs_id_.get();
}

} // namespace internal

} // namespace glint

#endif
