#! /usr/bin/env python

#
# Copyright (C) 2012 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authored by: Michi Henning <michi.henning@canonical.com>
#

#
# Little helper program to generate glint.h.
#
# Usage:
#
# make_include_all_header.py [-h] [-v] [-p PREFIX] [-f FRONT_MATTER] dir header_file hdrs [hdrs ...]
#

import argparse
import os
import sys
import re

progname = sys.argv[0]

#
# Write a message prefixed by the program name
#
def write_msg(stream, msg):
    stream.write(progname + ": " + msg + "\n")

#
# Write a message to stderr prefixed by the program name
#
def error(msg):
    write_msg(sys.stderr, msg)

#
# Check that it is safe to overwrite the file by checking for the cookie at the start of the file.
# Return the file descriptor for the open file if so, throw otherwise.
#
def check_overwrite(header_file, cookie):
    try:
        fd = os.open(header_file, os.O_EXCL | os.O_RDWR)
        start = os.read(fd, len(cookie))
        if start != cookie:
            error("will not overwrite existing file \"" + header_file + "\"")
            sys.exit(1)
        return fd
    except OSError as e:
        error(e.strerror)
        raise

#
# Open the file for read/write, creating it if necessary. If the file exists, check if it can be overwritten.
# Return the file descriptor for the file, or throw if it can't be created or overwritten.
#
def open_file(filepath, cookie):
    try:
        fd = os.open(filepath, os.O_CREAT | os.O_EXCL | os.O_RDWR, 0666)
    except OSError as e:
        if e.errno == os.errno.EEXIST:
            fd = check_overwrite(filepath, cookie)
        else:
            error("cannot open " + filepath + ": " + e.strerror)
            raise
    return fd

#
# Write the front matter for into the file, namely the cookie, preamble (if any), and the double-include guard.
#
def write_front_matter(file, prefix, cookie, hdr_name, preamble):
    hdr_name = hdr_name.upper().replace(".", "_")
    prefix = prefix.upper().replace("/", "_")
    guard = ""
    if prefix != "":
        guard = prefix + "_"
    guard += hdr_name

    file.write(cookie)

    file.write("\n")
    file.write("#ifndef " + guard + "\n")
    file.write("#define " + guard + "\n")
    file.write("\n")

    if preamble != "":
        pa = open(preamble, 'r')
        pa_lines = pa.readlines()
        file.writelines(pa_lines)
        file.write("\n")
        pa.close()

#
# Write the include directives, possibly prefixed by a path.
#
def write_include_directives(file, prefix, includes):
    if prefix != "":
        prefix += "/"

    for i in includes:
        file.write("#include <" + prefix + i +">\n")
    if len(includes) != 0:
        file.write("\n")

#
# Write the stuff at the end of the file.
#
def write_end_matter(file):
    file.write("#endif\n")

#
# Generate the header file with name hdr_out in directory dir. all_hdrds is the list of header names for
# the include directives # (without any path prefix). cookie is the text at the beginning of the file to
# prevent accidental overwriting of existing files, prefix is any path prefix for the include directives,
# and preamble is the name of the file containing any text # that should be included in the file at the
# beginning, such as a copyright message. If verbose is set, print a message if the header is re-generated.
#
def generate_header(dir, hdr_out, all_hdrs, cookie, prefix, preamble, verbose):

    #
    # Construct header path name.
    #
    dir = dir.rstrip("/")
    fullname = dir
    if dir != "":
        fullname = fullname + "/"
    fullname = fullname + hdr_out

    #
    # Create the file or open it if it exists already and can be overwritten.
    #
    fd = open_file(fullname, cookie)
    file = os.fdopen(fd, 'w+')
    file.seek(0)

    #
    # Collect the existing included files into a list, without any prefix and store the
    # existing prefix (if any) in old_prefix.
    #
    lines = file.readlines()
    pat = re.compile(r'^#[ \t]*include[ \t]+<(.*/)?(.*)>')

    includes = []
    old_prefix = ""
    once = False
    for l in lines:
        mo = pat.match(l)
        if mo:
            includes += [ mo.group(2) ]
            if not once:
                old_prefix = mo.group(1)
                once = True

    #
    # If the preamble file was touched, we generate the header regardless.
    #
    force_regen = False
    if preamble != "":
        pa_time = os.path.getmtime(preamble)
        hdr_time = os.path.getmtime(fullname)
        if pa_time > hdr_time:
            force_regen = True

    #
    # If the prefix has changed or the headers aren't the same, we need to re-generate the file.
    #
    if not force_regen:
        prefix = prefix.rstrip("/")
        old_prefix = old_prefix.rstrip("/")

        if prefix == old_prefix:
            includes.sort()
            old_includes = sorted(all_hdrs)
            if includes == old_includes:
                file.close()
                sys.exit()    # Nothing to do.

    #
    # Something has changed, re-generate the file contents.
    #
    file.truncate(0)
    file.seek(0)
    write_front_matter(file, prefix, cookie, hdr_out, preamble)
    write_include_directives(file, prefix, all_hdrs)
    write_end_matter(file)
    file.close()
    if verbose:
        write_msg(sys.stdout, "re-generated " + fullname)

#
# Main program.
#

#
# Parse arguments.
#
parser = argparse.ArgumentParser(description = 'Generate a header that includes all headers pass on the command line')
parser.add_argument('-v', '--verbose', action='store_true', help = 'Print message if the file was re-generated')
parser.add_argument('-p', '--prefix', nargs=1, default=[""],
    help = 'The directory prefix for each include directive, such as "glint/"')
parser.add_argument('-f', '--front-matter', nargs=1, default=[""],
    help = 'Name of a file with the contents of any front matter to be generated into the header, such as a copyright notice')
parser.add_argument('dir', nargs = 1, help = 'The directory into which to generate the header')
parser.add_argument('header_file', nargs = 1, help = 'The name of the generated header file')
parser.add_argument('hdrs', nargs = '+',
    help = 'The names of the header files for which to generate include directives, separated by semicolons or spaces')
args = parser.parse_args()
progname = parser.prog

#
# Cookie to guard against accidental overwriting of a file.
#
cookie = "//\n// File generated by make_include_all_header.py. DO NOT EDIT: CONTENTS WILL BE OVERWRITTEN!\n//\n"

#
# Collect arguments and options
#
dir = args.dir[0]
hdr_out = args.header_file[0]
hdrs = args.hdrs
prefix = args.prefix[0]
preamble = args.front_matter[0]

#
# Convert the header file names into a list of one header per element.
#
all_hdrs = []
for i in hdrs:
    s = i.replace(";", " ");
    all_hdrs += s.split()

#
# Generate the header file.
#
try:
    generate_header(dir, hdr_out, all_hdrs, cookie, prefix, preamble, args.verbose)
except OSError:
    exit(1)   # Message printed already

sys.exit()  # All went well
