/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

//
// Generalised resource pointer that guarantees deallocation. Use for arbitrary
// pairs of allocate/deallocate functions, such as XCreateDisplay/XDestroyDisplay.
//
// The resource managed by this class must be default-constructible and assignable.
//
// This class essentially does what unique_ptr does, but it works with opaque types
// and resource allocation functions that do not return a pointer type.
//
// Note: Do not use reset() to set the resource to the "no resource allocated" state.
//       Instead, call dealloc() to do this. ResourcePtr has no idea
//       what a "not allocated" resource value looks like and therefore cannot test
//       for it. If you use reset() to install a "no resource allocated" value for
//       for the resource, the deleter will eventually be called with this value
//       as its argument. Whether this is benign or not depends on the deleter. For
//       example, XFree() must not be called with a nullptr argument.
//
// Note: Do not call get() or release() if no resource is currently allocated.
//       Doing so throws std::logic_error.
//
// Note: This class is thread safe.
//
// Here is an example that shows how to use this for a glXCreateContext/GLXDestroyContext pair.
// The returned GLXContext is a pointer to an opaque type; unique_ptr cannot be used
// for this, even with a custom deleter because the signatures of the allocator and deallocator
// don't match unique_ptr's expectations.
//
// ResourcePtr<GLXContext, std::function<void(GLXContext)>> context =
//   std::bind(&glXDestroyContext, display_, std::placeholders::_1);
//
// display_ is declared as
//
//   Display* display_;
//
// in this case.
//
// The deleter for the resource can return any type (including int, such as returned by XDestroyWindow()),
// and it must accept a single argument of the resource type (GLXContext in this example).
//
// glXDestroyContext() expects the display pointer as the first argument so, for this example, std::bind
// converts the binary glXDestroyContext() function into a unary function suitable as the deleter.
//
// Rather than mucking around with std::bind, it is often easier to use a lambda. For example:
//
// ResourcePtr<GLXContext, std::function<void(GLXContext)>> context =
//   [this](GLXContext c) { glXDestroyContext(this->display_.get(), c); };
//
// This calls a member function dealloc_GLXContext that, in turn calls glXDestroyContext, supplying the
// display parameter. display_ in this case is a unique_ptr that holds the return value from XOpenDisplay().
//

#ifndef GLINT_UTIL_RESOURCEPTR_H
#define GLINT_UTIL_RESOURCEPTR_H

#include <boost/thread/mutex.hpp>

namespace glint
{

namespace util
{

template<typename ResourceType, typename Deleter>
class ResourcePtr
{
public:
    ResourcePtr(Deleter d);
    ResourcePtr(ResourceType r, Deleter d);
    ~ResourcePtr() noexcept;

    ResourceType get() const;               // Returns the resource.
    void reset(ResourceType r);             // Assigns a new resource, deallocaing the previous one.
    ResourceType release();                 // Releases the resource without calling the deleter.
    void dealloc();                         // Deallocates current resource (if any).
    bool has_resource() const noexcept;     // Returns whether a resource is currently being managed.

private:
    ResourceType resource_;                 // The managed resource.
    Deleter delete_;                        // The deleter to call.
    bool initialized_;                      // True while we have a resource assigned.
    mutable boost::mutex m_;                // Protects this instance.

    typedef boost::lock_guard<decltype(m_)> AutoLock;
};

template<typename R, typename D>
ResourcePtr<R, D>::
ResourcePtr(D d)
    : delete_(d), initialized_(false)
{
}

template<typename R, typename D>
ResourcePtr<R, D>::
ResourcePtr(R r, D d)
    : resource_(r), delete_(d), initialized_(true)
{
}

template<typename R, typename D>
ResourcePtr<R, D>::
~ResourcePtr() noexcept
{
    try
    {
        dealloc();
    }
    catch (...)
    {
    }
}

template<typename R, typename D>
inline
R
ResourcePtr<R, D>::
get() const
{
    AutoLock lock(m_);

    if (!initialized_)
    {
        throw std::logic_error("get() called on ResourcePtr without resource");
    }
    return resource_;
}

template<typename R, typename D>
void
ResourcePtr<R, D>::
reset(R r)
{
    AutoLock lock(m_);

    bool has_old = initialized_;
    R old_resource;

    if (has_old)
    {
        old_resource = resource_;
    }
    resource_ = r;
    initialized_ = true;    // If the deleter throws, we still statisfy the postcondition: resource_ == r
    if (has_old)
    {

        delete_(old_resource);
    }
}

template<typename R, typename D>
inline
R
ResourcePtr<R, D>::
release()
{
    AutoLock lock(m_);

    if (!initialized_)
    {
        throw std::logic_error("release() called on ResourcePtr without resource");
    }
    initialized_ = false;
    return resource_;
}

template<typename R, typename D>
void
ResourcePtr<R, D>::
dealloc()
{
    AutoLock lock(m_);

    if (!initialized_)
    {
        return;
    }
    initialized_ = false;   // If the deleter throws, we will not try it again for the same resource
    delete_(resource_);
}

template<typename R, typename D>
bool
ResourcePtr<R, D>::
has_resource() const noexcept
{
    AutoLock lock(m_);

    return initialized_;
}

} // namespace util

} // namespace glint

#endif
