/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Color.h>
#include <glint/ColorNames.h>
#include <glint/GlintExceptions.h>

#include <boost/algorithm/string/case_conv.hpp>

#include <cstdint>
#include <cmath>
#include <cassert>
#include <sstream>
#include <iomanip>

using namespace glint;
using namespace std;

namespace glint
{

namespace
{

float
hex_to_int(const string& hexchars)
{
    istringstream s(hexchars);
    int i;
    s >> hex >> i;
    return i;
}

void
throw_bad_color(const char* range, const string& rm, const string& gm, const string& bm, const string& am)
{
    int errs = !rm.empty() + !gm.empty() + !bm.empty() + !am.empty();
    assert(errs > 0);

    ostringstream msg;
    msg << "Invalid color argument";
    if (errs != 1)
    {
        msg << "s";
    }
    msg << " outside [" << range << "] range: ";

    bool first = true;
    if (!rm.empty())
    {
        msg << rm;
        first = false;
    }
    if (!gm.empty())
    {
        msg << (first ? "" : ", ") << gm;
        first = false;
    }
    if (!bm.empty())
    {
        msg << (first ? "" : ", ") << bm;
        first = false;
    }
    if (!am.empty())
    {
        msg << (first ? "" : ", ") << am;
        first = false;
    }

    throw InvalidArgumentException(msg.str());
}

void
throw_if_bad_color(int red, int green, int blue, int alpha)
{
    if (!(value_in_range(red) && value_in_range(green) && value_in_range(blue) && value_in_range(alpha)))
    {
        throw_bad_color(int_range,
                        value_msg(red, "red"),
                        value_msg(green, "green"),
                        value_msg(blue, "blue"),
                        value_msg(alpha, "alpha"));
    }
}

void
throw_if_bad_color(float red, float green, float blue, float alpha)
{
    if (!(value_in_range(red) && value_in_range(green) && value_in_range(blue) && value_in_range(alpha)))
    {
        throw_bad_color(float_range,
                        value_msg(red, "red"),
                        value_msg(green, "green"),
                        value_msg(blue, "blue"),
                        value_msg(alpha, "alpha"));
    }
}

} // namespace

Color::
Color() noexcept
    : red_(0), green_(0), blue_(0), alpha_(0xff)
{
}

Color::
Color(int red, int green, int blue, int alpha)
    : red_(red), green_(green), blue_(blue), alpha_(alpha)
{
    throw_if_bad_color(red, green, blue, alpha);
}

Color::
Color(float red, float green, float blue, float alpha)
    : red_(to_int(red)), green_(to_int(green)), blue_(to_int(blue)), alpha_(to_int(alpha))
{
    throw_if_bad_color(red, green, blue, alpha);
}

Color::
Color(uint32_t c) noexcept
{
    set_uint32(c);
}

Color::
Color(const uint8_t (& c)[4]) noexcept
    : red_(c[0]), green_(c[1]), blue_(c[2]), alpha_(c[3])
{
}

Color::
Color(const float (& c)[4])
    : red_(to_int(c[0])), green_(to_int(c[1])), blue_(to_int(c[2])), alpha_(to_int(c[3]))
{
    throw_if_bad_color(c[0], c[1], c[2], c[3]);
}

Color::
Color(const string& name)
{
    if (set_from_hex(name))
    {
        return;
    }
    set_from_SVG_name(name);
}

bool
Color::
operator<(const Color& rhs) const
{
    uint32_t l = (alpha_ << 24) | (red_ << 16) | (green_ << 8) | blue_;
    uint32_t r = (rhs.alpha_ << 24) | (rhs.red_ << 16) | (rhs.green_ << 8) | rhs.blue_;
    return l < r;
}

bool
Color::
operator==(const Color& rhs) const
{
    return red_ == rhs.red_ && green_ == rhs.green_ && blue_ == rhs.blue_ && alpha_ == rhs.alpha_;
}

int
Color::
set_red(int v)
{
    throw_if_not_in_range(v, "red", int_range);
    int old = red_;
    red_ = v;
    return old;
}

int
Color::
set_green(int v)
{
    throw_if_not_in_range(v, "green", int_range);
    int old = green_;
    green_ = v;
    return old;
}

int
Color::
set_blue(int v)
{
    throw_if_not_in_range(v, "blue", int_range);
    int old = blue_;
    blue_ = v;
    return old;
}

int
Color::
set_alpha(int v)
{
    throw_if_not_in_range(v, "alpha", int_range);
    int old = alpha_;
    alpha_ = v;
    return old;
}

float
Color::
set_red_f(float v)
{
    throw_if_not_in_range(v, "red", float_range);
    float old = to_float(red_);
    red_ = to_int(v);
    return old;
}

float
Color::
set_green_f(float v)
{
    throw_if_not_in_range(v, "green", float_range);
    float old = to_float(green_);
    green_ = to_int(v);
    return old;
}

float
Color::
set_blue_f(float v)
{
    throw_if_not_in_range(v, "blue", float_range);
    float old = to_float(blue_);
    blue_ = to_int(v);
    return old;
}

float
Color::
set_alpha_f(float v)
{
    throw_if_not_in_range(v, "alpha", float_range);
    float old = to_float(alpha_);
    alpha_ = to_int(v);
    return old;
}

uint32_t
Color::
set_uint32(uint32_t c) noexcept
{
    uint32_t old = (alpha_ << 24) | (red_ << 16) | (green_ << 8) | blue_;
    alpha_ = c >> 24;
    red_ = (c >> 16) & 0xff;
    green_ = (c >> 8) & 0xff;
    blue_ = c & 0xff;
    return old;
}

void
Color::
uint8_v(uint8_t (& c)[4]) const noexcept
{
    c[0] = red_;
    c[1] = green_;
    c[2] = blue_;
    c[3] = alpha_;
}

void
Color::
set_uint8_v(const uint8_t c[4]) noexcept
{
    red_ = c[0];
    green_ = c[1];
    blue_ = c[2];
    alpha_ = c[3];
}

void
Color::
set_uint8_v(const uint8_t c[4], uint8_t (& old)[4]) noexcept
{
    old[0] = red_;
    old[1] = green_;
    old[2] = blue_;
    old[3] = alpha_;
    red_ = c[0];
    green_ = c[1];
    blue_ = c[2];
    alpha_ = c[3];
}

void
Color::
float_v(float (& c)[4]) const noexcept
{
    c[0] = to_float(red_);
    c[1] = to_float(green_);
    c[2] = to_float(blue_);
    c[3] = to_float(alpha_);
}

void
Color::
set_float_v(const float c[4])
{
    throw_if_bad_color(c[0], c[1], c[2], c[3]);
    red_ = to_int(c[0]);
    green_ = to_int(c[1]);
    blue_ = to_int(c[2]);
    alpha_ = to_int(c[3]);
}

void
Color::
set_float_v(const float c[4], float (& old)[4])
{
    throw_if_bad_color(c[0], c[1], c[2], c[3]);
    old[0] = to_float(red_);
    old[1] = to_float(green_);
    old[2] = to_float(blue_);
    old[3] = to_float(alpha_);
    red_ = to_int(c[0]);
    green_ = to_int(c[1]);
    blue_ = to_int(c[2]);
    alpha_ = to_int(c[3]);
}

void
Color::
set_from_SVG_name(const string& name)
{
    string lower(name);
    boost::algorithm::to_lower(lower);
    map<string, SVGColor>::const_iterator it = SVG_colors.find(lower.c_str());
    if (it != SVG_colors.end())
    {
        uint32_t c = static_cast<uint32_t>(it->second);
        blue_ = c & 0xff;
        green_ = (c >> 8) & 0xff;
        red_ = (c >> 16) & 0xff;
        alpha_ = 0xff;
    }
    else
    {
        string msg = "Invalid SVG color name: \"";
        msg += name + "\"";
        if (name.empty())
        {
            msg += " (empty string)";
        }
        throw InvalidArgumentException(msg);
    }
}

bool
Color::
set_from_hex(const string& name)
{
    if (name.empty() || name[0] != '#')
    {
        return false;
    }

    // String starts with '#', so it must be a valid color spec
    static const char* valid_chars = "0123456789aAbBcCdDeDfF";
    if (name.substr(1).find_first_not_of(valid_chars) != string::npos)
    {
        string msg = "Invalid color string (must only contain hex digits): \"";
        msg += name + "\"";
        throw InvalidArgumentException(msg);
    }

    switch (name.size())
    {
        case 4:
        case 5:
        {
            red_ = hex_to_int(name.substr(1, 1) + name.substr(1, 1));
            green_ = hex_to_int(name.substr(2, 1) + name.substr(2, 1));
            blue_ = hex_to_int(name.substr(3, 1) + name.substr(3, 1));
            alpha_ = name.size() == 4 ? 0xff : hex_to_int(name.substr(4, 1) + name.substr(4, 1));
            break;
        }
        case 7:
        case 9:
        {
            red_ = hex_to_int(name.substr(1, 2));
            green_ = hex_to_int(name.substr(3, 2));
            blue_ = hex_to_int(name.substr(5, 2));
            alpha_ = name.size() == 7 ? 0xff : hex_to_int(name.substr(7, 2));
            break;
        }
        default:
        {
            string msg = "Invalid color string (incorrect number of hex digits): \"";
            msg += name + "\"";
            throw InvalidArgumentException(msg);
        }
    }
    return true;
}

} // namespace glint
