/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/HSVColor.h>

#include <cassert>

using namespace glint;
using namespace std;

namespace glint
{

namespace
{

inline
bool
degrees_in_range(int d)
{
    return d >= 0 && d < 360;
}

inline
bool
degrees_in_range(float d)
{
    return d >= 0.0f && d <= 1.0f;
}

inline
int
float_to_degrees(float degrees)
{
    // Round to closest degree
    return degrees == 1.0f ? 0.0f : floor((degrees + 1.0 / 360.0 / 2) * 360);
}

inline
float
degrees_to_float(int degrees)
{
    return degrees / 360.0f;
}

void
throw_bad_HSV(const char* range, const string& hm, const string& sm, const string& vm, const string& am)
{
    ostringstream msg;

    if (!hm.empty())
    {
        msg << hm;
    }

    int errs = !sm.empty() + !vm.empty() + !am.empty();

    assert(!hm.empty() || errs != 0);

    if (errs != 0)
    {
        msg << (hm.empty() ? "I" : "; i");

        msg << "nvalid color argument";
        if (errs != 1)
        {
            msg << "s";
        }
        msg << " outside [" << range << "] range: ";

        bool first = true;
        if (!sm.empty())
        {
            msg << (first ? "" : ", ") << sm;
            first = false;
        }
        if (!vm.empty())
        {
            msg << (first ? "" : ", ") << vm;
            first = false;
        }
        if (!am.empty())
        {
            msg << (first ? "" : ", ") << am;
            first = false;
        }
    }

    throw InvalidArgumentException(msg.str());
}

void
throw_if_bad_value(int h, int s, int v, int alpha)
{
    if (!(degrees_in_range(h) && value_in_range(s) && value_in_range(v) && value_in_range(alpha)))
    {
        stringstream hue_msg;
        if (!degrees_in_range(h))
        {
            hue_msg << "Invalid hue outside [" << degree_range << "] range: " << h;
        }
        throw_bad_HSV(int_range,
                      hue_msg.str(),
                      value_msg(s, "saturation"),
                      value_msg(v, "value"),
                      value_msg(alpha, "alpha"));
    }
}

void
throw_if_bad_value(float h, float s, float v, float alpha)
{
    if (!(degrees_in_range(h) && value_in_range(s) && value_in_range(v) && value_in_range(alpha)))
    {
        stringstream hue_msg;
        if (!degrees_in_range(h))
        {
            hue_msg << "Invalid hue outside [" << float_range << "] range: " << h;
        }
        throw_bad_HSV(float_range,
                      hue_msg.str(),
                      value_msg(s, "saturation"),
                      value_msg(v, "value"),
                      value_msg(alpha, "alpha"));
    }
}

template<typename T>
void
throw_if_bad_degrees(T degrees, const char* range)
{
    if (!(degrees_in_range(degrees)))
    {
        ostringstream msg;
        msg << "Invalid hue argument outside [" << range << "] range: " << degrees;
        throw InvalidArgumentException(msg.str());
    }
}

} // namespace

HSVColor::
HSVColor() noexcept
    : h_(0), s_(0), v_(0), alpha_(0xff)
{
}

// Create HSV from int values. s, v, and alpha must be in the range [0-255]. h must be in the range [0-359].
// Values outside these ranges thrown IllegalArgumentException.

HSVColor::
HSVColor(int h, int s, int v, int alpha)
    : h_(h), s_(s), v_(v), alpha_(alpha)
{
    throw_if_bad_value(h, s, v, alpha);
}

// Create HSV from float values. h, s, v, and alpha must be in the range [0.0-1.0]. If h == 1.0, it is set to 0.0.
// Values outside these ranges thrown IllegalArgumentException.

HSVColor::
HSVColor(float h, float s, float v, float alpha)
    : h_(float_to_degrees(h)), s_(to_int(s)), v_(to_int(v)), alpha_(to_int(alpha))
{
    throw_if_bad_value(h, s, v, alpha);
}

bool
HSVColor::
operator<(const HSVColor& rhs) const
{
    if (h_ != rhs.h_)
    {
        return h_ < rhs.h_;
    }
    uint32_t l = (s_ << 16) | (v_ << 8) | alpha_;
    uint32_t r = (rhs.s_ << 16) | (rhs.v_ << 8) | rhs.alpha_;
    return l < r;
}

bool
HSVColor::
operator==(const HSVColor& rhs) const
{
    return h_ == rhs.h_ && s_ == rhs.s_ && v_ == rhs.v_ && alpha_ == rhs.alpha_;
}

int
HSVColor::
set_hue(int h)
{
    throw_if_bad_degrees(h, degree_range);
    int old = h_;
    h_ = h;
    return old;
}

int
HSVColor::
set_saturation(int s)
{
    throw_if_not_in_range(s, "saturation", int_range);
    int old = s_;
    s_ = s;
    return old;
}

int
HSVColor::
set_value(int v)
{
    throw_if_not_in_range(v, "value", int_range);
    int old = v_;
    v_ = v;
    return old;
}

int
HSVColor::
set_alpha(int a)
{
    throw_if_not_in_range(a, "alpha", int_range);
    int old = alpha_;
    alpha_ = a;
    return old;
}

float
HSVColor::
set_hue_f(float h)
{
    throw_if_not_in_range(h, "hue", float_range);
    float old = degrees_to_float(h_);
    h_ = float_to_degrees(h);
    return old;
}


float
HSVColor::
set_saturation_f(float s)
{
    throw_if_not_in_range(s, "saturation", float_range);
    float old = to_float(s_);
    s_ = to_int(s);
    return old;
}

float
HSVColor::
set_value_f(float v)
{
    throw_if_not_in_range(v, "value", float_range);
    float old = to_float(v_);
    v_ = to_int(v);
    return old;
}

float
HSVColor::
set_alpha_f(float a)
{
    throw_if_not_in_range(a, "alpha", float_range);
    float old = to_float(alpha_);
    alpha_ = to_int(a);
    return old;
}

// Convert RGB to HSV. Algorithm adapted from http://www.cs.rit.edu/~ncs/color/t_convert.html, by Nan C. Schaller.

void
RGB_to_HSV(float r, float g, float b, float& h, float& s, float& v) noexcept
{
    float min = std::min(r, std::min(g, b));
    float max = std::max(r, std::max(g, b));
    v = max;

    float delta = max - min;
    if (delta == 0)                   // Black, grey, or white
    {
        h = 0;
        s = 0;
        return;
    }

    assert(max != 0);
    s = delta / max;

    float hue;
    if (r == max)
    {
        hue = (g - b) / delta;        // Between yellow and magenta
    }
    else if (g == max)
    {
        hue = 2 + (b - r) / delta;    // Between cyan and yellow
    }
    else
    {
        hue = 4 + (r - g) / delta;    // Between magenta and cyan
    }
    hue *= 60;
    if (hue < 0)
    {
        hue += 360;
    }
    h = degrees_to_float(hue);
}

HSVColor
RGB_to_HSV(const Color& c) noexcept
{
    float red = c.red_f();
    float green = c.green_f();
    float blue = c.blue_f();

    float hue, saturation, value;
    RGB_to_HSV(red, green, blue, hue, saturation, value);
    return HSVColor(hue, saturation, value, c.alpha_f());
}

// Convert HSV to RGB. Algorithm adapted from http://www.cs.rit.edu/~ncs/color/t_convert.html, by Nan C. Schaller.

void
HSV_to_RGB(float h, float s, float v, float& r, float& g, float& b) noexcept
{
    if (s == 0.0f)
    {
        r = g = b = v;  // white, gray, or black
        return;
    }

    float hue = float_to_degrees(h); // Normalizes to [0 - 359] range

    hue /= 60;
    int sector = std::floor(hue);         // Find 60-degree sector (0-5)

    float f = hue - sector;               // Fractional part of hue
    float p = v * (1 - s);
    float q = v * (1 - s * f);
    float t = v * (1 - s * (1 - f));

    assert(sector >= 0 && sector < 6);
    switch (sector)
    {
        case 0:         // Red to yellow
        {
            r = v;
            g = t;
            b = p;
            break;
        }
        case 1:         // Yellow to green
        {
            r = q;
            g = v;
            b = p;
            break;
        }
        case 2:         // Green to cyan
        {
            r = p;
            g = v;
            b = t;
            break;
        }
        case 3:         // Cyan to blue
        {
            r = p;
            g = q;
            b = v;
            break;
        }
        case 4:         // Blue to magenta
        {
            r = t;
            g = p;
            b = v;
            break;
        }
        default:         // Magenta to red
        {
            r = v;
            g = p;
            b = q;
            break;
        }
    }
}

Color
HSV_to_RGB(const HSVColor& c) noexcept
{
    float hue = c.hue_f();
    float saturation = c.saturation_f();
    float value = c.value_f();

    float red, green, blue;
    HSV_to_RGB(hue, saturation, value, red, green, blue);
    return Color(red, green, blue, c.alpha_f());
}

} // namespace glint
