/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Runtime.h>
#include <glint/GlintExceptions.h>
#include <glint/internal/RuntimeImpl.h>

#include <cassert>
#include <sstream>

using namespace std;

namespace glint
{

class GraphicsDisplay;

using namespace internal;

//
// Factory method to instantiate a new glint runtime.
//

Runtime::UPtr
Runtime::runtime_create()
{
    return UPtr(new Runtime);
}

namespace
{

string
env_name(GLEnv env)
{
    assert(env == GLEnv::OpenGL || env == GLEnv::OpenGLES);
    return env == GLEnv::OpenGL ? "GL" : "GLES2";
}

} // namespace

//
// If the API client was compiled with gl.h (OpenGL) headers, it must link with a glint library that also
// used those headers when it was compiled. Similarly, if the API client was compiled with gl2.h (GLES) headers,
// the headers used when the glint library was compiled must match.
//
// This check throws a LogicException if there is a mismatch. It relies on the inline Runtime::create() function
// that is defined in Runtime.h which, in turn, relies on the definition of the gl_environment constant in
// OpenGL.h. The value of this constant will differ for the caller and the glint library if they were compiled
// using different headers.
//

void
Runtime::check_GL_environment(GLEnv env)
{
    if (env != gl_environment)
    {
        ostringstream msg;
        msg << "Mismatch between glint library and calling code: glint library was compiled with "
            << env_name(gl_environment) << " headers, but calling code was compiled with "
            << env_name(env) << " headers";
        throw LogicException(msg.str());
    }
}

void
Runtime::
shutdown()
{
    p_->shutdown();
}

int
Runtime::max_display_width() const noexcept
{
    return p_->max_display_width();
}

int
Runtime::max_display_height() const noexcept
{
    return p_->max_display_height();
}

GraphicsDisplay*
Runtime::create_display()
{
    return p_->create_display();
}

GraphicsDisplay*
Runtime::create_display(int width, int height, int xpos, int ypos)
{
    return p_->create_display(width, height, xpos, ypos);
}

GraphicsDisplay*
Runtime::display() const
{
    return p_->display();
}

Runtime::Runtime()
    : p_(make_shared<RuntimeImpl>())
{
}

Runtime::~Runtime() noexcept
{
}

} // namespace glint
