/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */


#include <glint/genv/egl/internal/EGLGEnv.h>
#include <glint/genv/egl/EGLException.h>
#include <glint/OpenGL.h>
#include <glint/genv/egl/internal/EGLGraphicsDisplay.h>

#include <boost/algorithm/string.hpp>

#include <sstream>

using namespace glint::internal;
using namespace boost::algorithm;
using namespace std;

namespace glint
{

namespace genv
{

namespace egl
{

namespace internal
{

//
// The constructor instantiates a concrete windowing environment. This initializes
// the windowing environment to the the point where it has a connection to the display server
// (if any), and can be asked about the screen dimensions.
//

EGLGEnv::
EGLGEnv()
    : wenv_(genv::internal::create_concrete_wenv())
    , egl_display_([this](EGLDisplay d) { this->terminate_egl(d); })
    , context_([this](EGLContext c) { this->destroy_egl_context(c); })
    , surface_([this](EGLSurface s) { this->destroy_egl_surface(s); })
    , win_(nullptr)
{
    throw_if_wrong_EGL_version();
}

EGLGEnv::
~EGLGEnv() noexcept
{
    try
    {
        shutdown();
    }
    catch (...)
    {
    }
}

void
EGLGEnv::
shutdown()
{
    if (!surface_.has_resource())   // Calling shutdown twice does nothing
    {
        return;
    }

    exception_ptr ep;

    if (!eglMakeCurrent(egl_display_.get(), EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT))
    {
        ep = make_exception_ptr(EGLException("eglMakeCurrent() failed"));
    }

    try
    {
        surface_.dealloc();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    try
    {
        context_.dealloc();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    try
    {
        egl_display_.dealloc();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    try
    {
        wenv_->shutdown();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }
    wenv_.reset(nullptr);

    try
    {
        gdisplay_.reset(nullptr);
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    if (ep)
    {
        rethrow_exception(ep);
    }
}

int
EGLGEnv::
max_display_width() const noexcept
{
    return wenv_->max_display_width();
}

int
EGLGEnv::
max_display_height() const noexcept
{
    return wenv_->max_display_height();
}

//
// Called to create a new display, which is a rectangular region connected to an OpenGL context.
// Instantiating the GraphicsDisplay initializes the graphics environment to the point where it
// has established the visual parameters (bit depth, double-buffering, etc) that are needed
// for window creation. Once the window is created, we complete the initialization by connecting
// the window and the graphics context.
//

GraphicsDisplay*
EGLGEnv::
create_display(int width, int height, int xpos, int ypos)
{
    assert(!win_);
    create_egl_context();
    win_ = dynamic_cast<genv::ConcreteWindow*>(wenv_->create_window(visual_id_, width, height, xpos, ypos));
    assert(win_);
    create_egl_surface();
    make_current();
    throw_if_wrong_GLES_version();
    gdisplay_.reset(new EGLGraphicsDisplay(*this, *wenv_));
    return gdisplay_.get();
}

std::string
EGLGEnv::
name() const
{
    return "EGL";
}

int
EGLGEnv::
major_version() const noexcept
{
    return egl_major_;
}

int
EGLGEnv::
minor_version() const noexcept
{
    return egl_minor_;
}

int
EGLGEnv::
gl_major_version() const noexcept
{
    assert(gdisplay_);
    return gl_major_;
}

int
EGLGEnv::
gl_minor_version() const noexcept
{
    assert(gdisplay_);
    return gl_minor_;
}

void
EGLGEnv::
make_current()
{
    if (!eglMakeCurrent(egl_display_.get(), surface_.get(), surface_.get(), context_.get()))
    {
        throw EGLException("eglMakeCurrent() failed");
    }
}

void
EGLGEnv::
release_current()
{
    if (!eglMakeCurrent(egl_display_.get(), EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT))
    {
        throw EGLException("could not release current: eglMakeCurrent() failed");
    }
}

void
EGLGEnv::
swap_buffers()
{
    if (!eglSwapBuffers(egl_display_.get(), surface_.get()))
    {
        throw EGLException("eglSwapBuffers() failed");
    }
}

void
EGLGEnv::
throw_if_wrong_EGL_version()
{
    egl_display_.reset(eglGetDisplay(EGL_DEFAULT_DISPLAY));
    if (egl_display_.get() == EGL_NO_DISPLAY)
    {
        throw EGLException("eglGetDisplay() failed");
    }

    if (!eglInitialize(egl_display_.get(), &egl_major_, &egl_minor_))
    {
        throw EGLException("eglInitialize() failed");
    }

    if (egl_major_ < 1 || (egl_major_ == 1 && egl_minor_ < 3))
    {
        ostringstream reason;
        reason << "Unsupported EGL version: " << egl_major_ << "." << egl_minor_ << " (minimum required: 1.3)";
        throw EGLException(reason.str());
    }
}

void
EGLGEnv::
terminate_egl(EGLDisplay d)
{
    if (d == EGL_NO_DISPLAY)
    {
        return;
    }

    exception_ptr ep;

    if (!eglMakeCurrent(d, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT))
    {
        ep = make_exception_ptr(EGLException("eglMakeCurrent() failed"));
    }

    if (!eglTerminate(d))
    {
        EGLException e("eglTerminate() failed");
        ep = e.remember(ep);
    }

    if (ep)
    {
        rethrow_exception(ep);
    }
}

void
EGLGEnv::
create_egl_context()
{
    EGLint const attr[] = {
        EGL_RED_SIZE,     8,
        EGL_GREEN_SIZE,   8,
        EGL_BLUE_SIZE,    8,
        EGL_ALPHA_SIZE,   8,
        EGL_DEPTH_SIZE,  24,
        EGL_STENCIL_SIZE, 8,
        EGL_RENDERABLE_TYPE,
        EGL_OPENGL_ES2_BIT,
        EGL_NONE
    };

    EGLint num_config;
    if (!eglChooseConfig(egl_display_.get(), attr, &egl_config_, 1, &num_config))
    {
        throw EGLException("eglChooseConfig() failed");
    }

    if (!eglGetConfigAttrib(egl_display_.get(), egl_config_, EGL_NATIVE_VISUAL_ID, &visual_id_))
    {
        throw EGLException("eglGetConfigAttrib() failed");
    }

    const EGLint context_attr [] =
    {
        EGL_CONTEXT_CLIENT_VERSION, 2,
        EGL_NONE
    };
    context_.reset(eglCreateContext(egl_display_.get(), egl_config_, EGL_NO_CONTEXT, context_attr));
    if (context_.get() == EGL_NO_CONTEXT)
    {
        throw EGLException("eglCreateContext() failed");
    }
}

void
EGLGEnv::
destroy_egl_context(EGLContext c)
{
    if (!eglDestroyContext(egl_display_.get(), c))
    {
        throw EGLException("eglDestroyContext() failed");
    }
}

void
EGLGEnv::
create_egl_surface()
{
    surface_.reset(eglCreateWindowSurface(egl_display_.get(), egl_config_, win_->window(), nullptr));
    if (surface_.get() == EGL_NO_SURFACE)
    {
        throw EGLException("eglCreateWindowSurface() failed");
    }
}

void
EGLGEnv::
destroy_egl_surface(EGLContext c)
{
    if (!eglDestroySurface(egl_display_.get(), c))
    {
        throw EGLException("eglDestroySurface() failed");
    }
}

void
EGLGEnv::
throw_if_wrong_GLES_version()
{
    // vs has the format OpenGL<space>ES<space><version number><space><vendor-specific information>
    //
    const GLubyte* vs = glGetString(GL_VERSION);
    if (vs == nullptr)
    {
        throw EGLException("glGetString() failed");
    }
    string vstring(reinterpret_cast<const char*>(vs));

    // Split string into tokens separated by spaces.
    //
    vector<string> split_vector;
    split(split_vector, vstring, is_any_of(" "), token_compress_on);
    if (split_vector.size() < 3)
    {
        throw EGLException("malformed version string: \"" + vstring + "\"");
    }

    // Split version token (such as 5.7) at the period.
    //
    vector<string> versions;
    split(versions, split_vector[2], is_any_of("."), token_compress_off);
    if (versions.size() < 2)
    {
        throw EGLException("malformed version string: \"" + vstring + "\"");
    }
    try
    {
        gl_major_ = stoi(versions[0]);
        gl_minor_ = stoi(versions[1]);
    }
    catch (...)
    {
        throw EGLException("malformed version string: \"" + vstring + "\"");
    }

    if (gl_major_ < 2)
    {
        ostringstream reason;
        reason << "Unsupported GLES version: " << gl_major_ << "." << gl_minor_ << " (minimum required: 2.0)";
        throw EGLException(reason.str());
    }
}

} // namespace internal

} // namespace egl

namespace internal
{

//
// Library entry point so the glint library can instantiate an EGLGEnv without compile-time dependencies.
//

AbstractGEnv::UPtr
create_concrete_genv()
{
    return AbstractGEnv::UPtr(new glint::genv::egl::internal::EGLGEnv);
}

} // namespace internal

} // namespace genv

} // namespace glint
