/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/OpenGL.h>

#include <glint/genv/glx/internal/GLXGEnv.h>

#include <glint/genv/glx/GLXException.h>
#include <glint/genv/glx/internal/GLXGraphicsDisplay.h>

#include <boost/algorithm/string.hpp>

#include <sstream>

using namespace glint::internal;
using namespace boost::algorithm;
using namespace std;

namespace glint
{

namespace genv
{

namespace glx
{

namespace internal
{

//
// The constructor instantiates a concrete windowing environment. This initializes
// the windowing environment to the the point where it has a connection to the display server
// (if any), and can be asked about the screen dimensions.
//

GLXGEnv::
GLXGEnv()
    : wenv_(genv::internal::create_concrete_wenv())
    , context_([this](GLXContext c) { this->destroy_glx_context(c); })
    , vi_(nullptr, XFree)
    , win_(nullptr)
{
    throw_if_wrong_GLX_version();
    setup_visual();
    create_glx_context();
}

GLXGEnv::
~GLXGEnv() noexcept
{
    try
    {
        shutdown();
    }
    catch (...)
    {
    }
}

void
GLXGEnv::
shutdown()
{
    if (!vi_)
    {
        return;
    }
    vi_.reset(nullptr);

    exception_ptr ep;

    if (!glXMakeCurrent(wenv_->display(), None, nullptr))
    {
        ep = make_exception_ptr(GLXException("glXMakeCurrent() failed"));
    }

    try
    {
        context_.dealloc();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    try
    {
        wenv_->shutdown();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }
    wenv_.reset(nullptr);

    gdisplay_.reset(nullptr);

    if (ep)
    {
        rethrow_exception(ep);
    }
}

int
GLXGEnv::
max_display_width() const noexcept
{
    return wenv_->max_display_width();
}

int
GLXGEnv::
max_display_height() const noexcept
{
    return wenv_->max_display_height();
}

//
// Called to create a new display, which is a rectangular region connected to an OpenGL context.
// Instantiating the GraphicsDisplay initializes the graphics environment to the point where it
// has established the visual parameters (bit depth, double-buffering, etc) that are needed
// for window creation. Once the window is created, we complete the initialization by connecting
// the window and the graphics context.
//

GraphicsDisplay*
GLXGEnv::
create_display(int width, int height, int xpos, int ypos)
{
    assert(!win_);
    win_ = dynamic_cast<x11::X11Window*>(wenv_->create_window(vi_->visualid, width, height, xpos, ypos));
    assert(win_);
    make_current();
    throw_if_wrong_GL_version();
    if (glewInit() != GL_NO_ERROR)
    {
        throw GLXException("glewInit() failed");
    }
    gdisplay_.reset(new GLXGraphicsDisplay(*this, *wenv_));
    return gdisplay_.get();
}

std::string
GLXGEnv::
name() const
{
    return "GLX";
}

int
GLXGEnv::
major_version() const noexcept
{
    return glx_major_;
}

int
GLXGEnv::
minor_version() const noexcept
{
    return glx_minor_;
}

int
GLXGEnv::
gl_major_version() const noexcept
{
    assert(gdisplay_);
    return gl_major_;
}

int
GLXGEnv::
gl_minor_version() const noexcept
{
    assert(gdisplay_);
    return gl_minor_;
}


void
GLXGEnv::
make_current()
{
    if (!glXMakeCurrent(win_->display(), win_->window(), context_.get()))
    {
        throw GLXException("glXMakeCurrent() failed");
    }
}

void
GLXGEnv::
release_current()
{
    if (!glXMakeCurrent(wenv_->display(), None, nullptr))
    {
        throw GLXException("could release current: glXMakeCurrent() failed");
    }
}

void
GLXGEnv::
swap_buffers()
{
    glXSwapBuffers(win_->display(), win_->window());
}

void
GLXGEnv::
throw_if_wrong_GLX_version()
{
    // Can't run without glx.
    //
    int error_base;
    int event_base;
    if (!glXQueryExtension(wenv_->display(), &error_base, &event_base))
    {
        throw GLXException("No GLX extension available");
    }

    if (!glXQueryVersion(wenv_->display(), &glx_major_, &glx_minor_))
    {
        throw GLXException("glXQueryVersion() failed");
    }

    // We require at least 1.3.
    //
    if (glx_major_ < 1 || (glx_major_ == 1 && glx_minor_ < 4))
    {
        ostringstream reason;
        reason << "Unsupported GLX version: " << glx_major_ << "." << glx_minor_
               << " (minimum required: 1.4)";
        throw GLXException(reason.str());
    }
}

void
GLXGEnv::
setup_visual()
{
    // Get visual info capable of supporting OpenGL.
    //
    int attr[] =
    {
        GLX_RGBA,
        GLX_DOUBLEBUFFER,
        GLX_RED_SIZE,     8,
        GLX_GREEN_SIZE,   8,
        GLX_BLUE_SIZE,    8,
        GLX_ALPHA_SIZE,   8,
        GLX_DEPTH_SIZE,   24,
        GLX_STENCIL_SIZE, 8,
        None
    };
    vi_.reset(glXChooseVisual(wenv_->display(), XDefaultScreen(wenv_->display()), attr));
    if (!vi_.get())
    {
        throw GLXException("glXChooseVisual() failed");
    }
}

void
GLXGEnv::
create_glx_context()
{
    context_.reset(glXCreateContext(wenv_->display(), vi_.get(), nullptr, True));
    if (!context_.get())
    {
        throw GLXException("glXCreateContext() failed");
    }
}
void
GLXGEnv::
destroy_glx_context(GLXContext c)
{
    glXDestroyContext(wenv_->display(), c);
}

void
GLXGEnv::
throw_if_wrong_GL_version()
{
    // vs has the format <major>.<minor> or <major>.<minor>.<release>.
    // This is possibly followed by a space and vendor-specific information.
    //
    const GLubyte* vs = glGetString(GL_VERSION);
    if (vs == nullptr)
    {
        throw GLXException("glGetString() failed");
    }
    string vstring(reinterpret_cast<const char*>(vs));

    // Split string into tokens separated by spaces.
    //
    vector<string> split_vector;
    split(split_vector, vstring, is_any_of(" "), token_compress_on);
    if (split_vector.empty())
    {
        throw GLXException("malformed version string: \"" + vstring + "\"");
    }

    // Split version token (such as 5.7) at the period.
    //
    vector<string> versions;
    split(versions, split_vector[0], is_any_of("."), token_compress_off);
    if (versions.size() < 2)
    {
        throw GLXException("malformed version string: \"" + vstring + "\"");
    }
    try
    {
        gl_major_ = stoi(versions[0]);
        gl_minor_ = stoi(versions[1]);
    }
    catch (...)
    {
        throw GLXException("malformed version string: \"" + vstring + "\"");
    }

    if (gl_major_ < 2)
    {
        ostringstream reason;
        reason << "Unsupported GL version: " << gl_major_ << "." << gl_minor_ << " (minimum required: 2.0)";
        throw GLXException(reason.str());
    }
}

} // namespace internal

} // namespace GLX

namespace internal
{

//
// Library entry point so the glint library can instantiate an GLXGEnv without compile-time dependencies.
//

AbstractGEnv::UPtr
create_concrete_genv()
{
    return AbstractGEnv::UPtr(new glint::genv::glx::internal::GLXGEnv);
}

} // namespace internal

} // namespace genv

} // namespace glint
