/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/genv/x11/internal/X11WEnv.h>
#include <glint/genv/x11/X11Exception.h>
#include <glint/genv/x11/X11Window.h>

#include <cassert>

namespace glint
{

namespace genv
{

namespace x11
{

namespace internal
{

using namespace genv::internal;
using namespace std;

X11WEnv::
X11WEnv()
    : display_(nullptr, XCloseDisplay)
    , vi_(nullptr, [this](XVisualInfo* vi) { this->free_xvisual(vi); })
    , cmap_([this](Colormap m) { this->free_colormap(m); })
    , x11win_([this](Window w) { this->destroy_window(w); })
{
    connect_display();
    set_screen_size();
}

X11WEnv::
~X11WEnv() noexcept
{
    try
    {
        shutdown();
    }
    catch (...)
    {
    }
}

void
X11WEnv::
shutdown()
{
    if (!vi_)
    {
        return;
    }

    exception_ptr ep;

    try
    {
        x11win_.dealloc();
    }
    catch (const glint::Exception&)
    {
        ep = make_exception_ptr(current_exception());
    }

    try
    {
        cmap_.dealloc();
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    try
    {
        vi_.reset(nullptr);
    }
    catch (glint::Exception& e)
    {
        ep = e.remember(ep);
    }

    display_.reset(nullptr);

    if (ep)
    {
        rethrow_exception(ep);
    }
}

Display*
X11WEnv::
display() const noexcept
{
    assert(display_);
    return display_.get();
}

int
X11WEnv::
max_display_width() const noexcept
{
    return max_width_;
}

int
X11WEnv::
max_display_height() const noexcept
{
    return max_height_;
}

namespace
{

void init_win_attr(XSetWindowAttributes& a, unsigned long& m, Colormap cm)
{
    m = 0;
    a.background_pixel = 0;
    m |= CWBackPixel;
    a.border_pixel = 0;
    m |= CWBorderPixel;
    a.colormap = cm;
    m |= CWColormap;
    a.override_redirect = true;
    m |= CWOverrideRedirect;
}

} // namespace

AbstractWindow*
X11WEnv::
create_window(int visual_id, int width, int height, int xpos, int ypos)
{
    assert(!vi_); // Must not be called more than once.

    XVisualInfo visual_template;
    int num_visuals;
    visual_template.visualid = visual_id;
    vi_.reset(XGetVisualInfo(display_.get(), VisualIDMask, &visual_template, &num_visuals));
    if (!vi_)
    {
        throw X11Exception("XGetVisualInfo() failed");
    }

    auto root_window = RootWindow(display_.get(), vi_->screen);
    cmap_.reset(XCreateColormap(display_.get(), root_window, vi_->visual, AllocNone));
    if (!cmap_.get())
    {
        throw X11Exception("XCreateColormap() failed");
    }
    XSetWindowAttributes win_attr;
    unsigned long mask;
    init_win_attr(win_attr, mask, cmap_.get());
    x11win_.reset(XCreateWindow(display_.get(),
                                root_window,
                                xpos,
                                ypos,
                                width,
                                height,
                                0,
                                vi_->depth,
                                InputOutput,
                                vi_->visual,
                                mask,
                                &win_attr));
    if (!x11win_.get())
    {
        throw X11Exception("XCreateWindow() failed");
    }

    // BUGFIX: The call to XGetGeometry() should not be necessary (and, indeed, is not necessary for GLX).
    //         However, for EGL, without this call, eglCreateWindowSurface() fails with an X11 BadDrawable error.
    //         It looks like XGetGeometry() has a side-effect that eglCreateWindowSurface() depends on.
    //         The problem exists on the following hardware/software combinations. It may be present for
    //         other environments as well:
    //
    //         - Precise and Quantal
    //         - Intel 3000 and 4000 graphics hardware
    //         - Mesa 8.0.2 and 9.0
    //         - Gallium (with VMWare) and Intel 965 drivers
    {
        Window win;
        unsigned int border_width;
        unsigned int bit_depth;
        unsigned int w;
        unsigned int h;
        int xpos;
        int ypos;
        if (!XGetGeometry(display_.get(), x11win_.get(), &win, &xpos, &ypos, &w, &h, &border_width, &bit_depth))
        {
            throw X11Exception("XGetGeometry() failed");
        }
    }

    win_.reset(new X11Window(display_.get(), x11win_.get()));

    // We set the window to be on top of the stacking order but leave it invisible.
    // The caller can call set_visibility() once it has initialized the window contents.
    //
    set_stacking_order(AbstractWEnv::StackingOrder::top);

    return win_.get();
}

glint::AbstractWindow*
X11WEnv::
window() const noexcept
{
    assert(x11win_.has_resource());
    return win_.get();
}

int
X11WEnv::
width() const noexcept
{
    assert(x11win_.has_resource());

    XWindowAttributes attr;
    if (!XGetWindowAttributes(display_.get(), x11win_.get(), &attr))
    {
        throw X11Exception("XGetWindowAttributes() failed");
    }
    return attr.width;
}

int
X11WEnv::
height() const noexcept
{
    assert(x11win_.has_resource());

    XWindowAttributes attr;
    if (!XGetWindowAttributes(display_.get(), x11win_.get(), &attr))
    {
        throw X11Exception("XGetWindowAttributes() failed");
    }
    return attr.height;
}

int
X11WEnv::
xpos() const noexcept
{
    assert(x11win_.has_resource());

    XWindowAttributes attr;
    if (!XGetWindowAttributes(display_.get(), x11win_.get(), &attr))
    {
        throw X11Exception("XGetWindowAttributes() failed");
    }
    return attr.x;
}

int
X11WEnv::
ypos() const noexcept
{
    assert(x11win_.has_resource());

    XWindowAttributes attr;
    if (!XGetWindowAttributes(display_.get(), x11win_.get(), &attr))
    {
        throw X11Exception("XGetWindowAttributes() failed");
    }
    return attr.y;
}

AbstractWEnv::Vis
X11WEnv::
set_visibility(AbstractWEnv::Vis v)
{
    assert(x11win_.has_resource());

    auto old_vis = visibility();
    if (v == old_vis)
    {
        return v;
    }
    if (v == AbstractWEnv::Vis::visible)
    {
        if (!XMapWindow(display_.get(), x11win_.get()))
        {
            throw X11Exception("XMapWindow() failed");
        }
    }
    else
    {
        if (!XUnmapWindow(display_.get(), x11win_.get()))
        {
            throw X11Exception("XUnmapWindow() failed");
        }
    }
    return old_vis;
}

AbstractWEnv::Vis
X11WEnv::
visibility()
{
    assert(x11win_.has_resource());

    XWindowAttributes attr;
    if (!XGetWindowAttributes(display_.get(), x11win_.get(), &attr))
    {
        throw X11Exception("XGetWindowAttributes() failed");
    }
    return attr.map_state == IsViewable ? AbstractWEnv::Vis::visible : AbstractWEnv::Vis::invisible;
}

void
X11WEnv::
set_stacking_order(AbstractWEnv::StackingOrder order)
{
    assert(x11win_.has_resource());

    if (order == AbstractWEnv::StackingOrder::top)
    {
        if (!XRaiseWindow(display_.get(), x11win_.get()))
        {
            throw X11Exception("XRaiseWindow() failed");
        }
    }
    else
    {
        if (!XLowerWindow(display_.get(), x11win_.get()))
        {
            throw X11Exception("XLowerWindow() failed");
        }
    }
}

void
X11WEnv::
connect_display()
{
    display_.reset(XOpenDisplay(nullptr));
    if (!display_)
    {
        throw X11Exception("XOpenDisplay() failed");
    }
}

void
X11WEnv::
free_xvisual(XVisualInfo* vi)
{
    if (!XFree(vi))
    {
        throw X11Exception("Cannot deallocate XVisualInfo: XFree() failed");
    }
}

void
X11WEnv::
free_colormap(Colormap cm)
{
    if (!XFreeColormap(display_.get(), cm))
    {
        throw X11Exception("XFreeColormap() failed");
    }
}

void
X11WEnv::
destroy_window(Window w)
{
    if (!XDestroyWindow(display_.get(), w))
    {
        throw X11Exception("XDestroyWindow() failed");
    }
}

void
X11WEnv::
set_screen_size()
{
    max_width_ = XDisplayWidth(display_.get(), XDefaultScreen(display_.get()));
    max_height_ = XDisplayHeight(display_.get(), XDefaultScreen(display_.get()));
}

} // namespace internal

} // namespace x11

namespace internal
{

AbstractWEnv::UPtr
create_concrete_wenv()
{
    return AbstractWEnv::UPtr(new x11::internal::X11WEnv);
}

} // namespace internal

} // namespace genv

} // namespace glint
