/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/internal/ExceptionImpl.h>
#include <glint/Exception.h>

using namespace std;

namespace glint
{

namespace internal
{

ExceptionImpl::
ExceptionImpl(const string& reason)
    : reason_(reason)
{
}

ExceptionImpl::
~ExceptionImpl() noexcept = default;

string
ExceptionImpl::
reason() const
{
    return reason_;
}

namespace
{

//
// Follow the nested exceptions that were rethrown along the call stack, printing them into s.
//
void
follow_nested(string& s, const nested_exception& nested, int indent_level, const std::string& indent)
{
    // Work out amount of indent.
    //
    string margin;
    for (int i = 0; i < indent_level; ++i)
    {
        margin += indent;
    }
    s += margin;

    // Add the what() string if this is a std::exception.
    //
    const std::exception* std_exception = dynamic_cast<const std::exception*>(&nested);
    if (std_exception)
    {
        s += std_exception->what();
    }

    // Add the reason if this is a glint::Exception.
    //
    const glint::Exception* glint_exception = dynamic_cast<const glint::Exception*>(&nested);
    if (glint_exception)
    {
        string reason = glint_exception->reason();
        if (!reason.empty())
        {
            s += ": " + reason;
        }
    }

    // Append info about unknown std::exception and std::nested_exception.
    //
    if (!glint_exception)
    {
        if (std_exception)
        {
            s += " (derived from std::exception and std::nested_exception)";
        }
        else
        {
            s += "std::nested_exception";
        }
    }

    // Add the details of any nested exception.
    //
    if (nested.nested_ptr())
    {
        s += ":\n";
        try
        {
            nested.rethrow_nested();
        }
        catch (const std::nested_exception& e)
        {
            follow_nested(s, e, indent_level + 1, indent);    // Follow the chain.
        }
        catch (const std::exception& e)
        {
            s += margin + indent;
            s += e.what();                                    // Can show only what() for std::exception.
        }
        catch (...)
        {
            s += margin + indent;
            s += "unknown exception";                         // Best we can do for an exception whose type we don't know.
        }
    }
}

//
// Follow the history chain and print each exception in the chain.
//

void
follow_history(string& s, int& count, const glint::Exception& e, int indent_level, const std::string& indent)
{
    string margin;
    for (int i = 0; i < indent_level; ++i)
    {
        margin += indent;
    }

    if (!e.get_earlier())
    {
        count = 1;  // We have reached the oldest exception; set exception generation count and terminate recursion.
    }
    else
    {
        try
        {
            rethrow_exception(e.get_earlier());
        }
        catch (const glint::Exception& e)
        {
            // Recurse along the chain until we hit the end, then, as we pop back up the levels, we increment the
            // count and print it as a generation number for the exception information.
            // A bit like the "kicks" in "Inception", except that the deepest level is level 1...
            //
            follow_history(s, count, e, indent_level, indent);
        }
        ++count;
    }

    //
    // Show info for this exception.
    //
    s += "\n";
    s += margin + "Exception #";
    s += to_string(count) + ":\n";
    follow_nested(s, e, indent_level + 1, indent);
}

} // namespace

string
ExceptionImpl::
to_string(const nested_exception& nested, int indent_level, const string& indent) const
{
    string s;

    // Printed this and any nested exceptions.
    //
    follow_nested(s, nested, indent_level, indent);

    // Check whether there is an exception history and print each exception in the history.
    //
    const glint::Exception* glint_exception = dynamic_cast<const glint::Exception*>(&nested);
    if (glint_exception && glint_exception->get_earlier())
    {
        string margin;
        for (int i = 0; i < indent_level; ++i)
        {
            margin += indent;
        }
        s += "\n";
        s += margin + indent + "Exception history:";
        try
        {
            rethrow_exception(glint_exception->get_earlier());
        }
        catch (const glint::Exception& e)
        {
            int count;
            follow_history(s, count, e, indent_level + 2, indent);
        }
    }
    return s;
}

exception_ptr
ExceptionImpl::
remember(const Exception* env, exception_ptr earlier_exception)
{
    // Doesn't prevent loops, but protects against accidental self-assignment.
    //
    if (previous_ != earlier_exception)
    {
        previous_ = earlier_exception;
    }
    return env->self();
}

exception_ptr
ExceptionImpl::
get_earlier() const noexcept
{
    return previous_;
}

} // namespace internal

} // namespace glint
