/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/internal/ProgramImpl.h>
#include <glint/Shader.h>
#include <glint/internal/ShaderImpl.h>
#include <glint/GLUtil.h>
#include <glint/GlintExceptions.h>
#include <glint/UniformVariable.h>

#include <cassert>
#include <sstream>

using namespace std;

namespace glint
{

namespace internal
{

namespace
{

void
delete_program(GLuint program_id) noexcept
{
    // The only way this could fail is if the program ID is incorrect, that is, if glCreateProgram() failed.
    glDeleteProgram(program_id);
    clear_gl_error();
}

} // namespace

GLuint ProgramImpl::current_prog_ = 0;

ProgramImpl::
ProgramImpl(const VertexShader& vs, const FragmentShader& fs)
    : id_(&delete_program)
    , vs_id_([this](GLuint s) { this->detach_shader(s); })
    , fs_id_([this](GLuint s) { this->detach_shader(s); })
    , initialized_(false)
{
    // Create the program, attach both shaders, and link.

    GLuint prog_id = glCreateProgram();
    if (prog_id == 0)
    {
        throw OpenGLException("glCreateProgram() failed", 0);
    }
    id_.reset(prog_id);

    GLuint vertex_id = vs.id();
    glAttachShader(prog_id, vertex_id);
    check_gl("cannot attach vertex shader: glAttachShader() failed");
    vs_id_.reset(vertex_id);

    GLuint fragment_id = fs.id();
    glAttachShader(prog_id, fragment_id);
    check_gl("cannot attach fragment shader: glAttachShader() failed");
    fs_id_.reset(fragment_id);

    glLinkProgram(prog_id);
    GLenum err = clear_gl_error();
    if (err != GL_NO_ERROR)
    {
        throw OpenGLException(get_log_msg("glLinkProgram()"), err);
    }
}

ProgramImpl::
~ProgramImpl() noexcept
{
    current_prog_ = 0;
}

void
ProgramImpl::
use() const
{
    glUseProgram(id_.get());

    // Many implementations delay reporting an error until glUseProgram() is called, even if the
    // error can be diagnosed at link time. We check for a log message once more here, even
    // though we shouldn't have to according to the specification.
    GLint err = clear_gl_error();
    if (err != GL_NO_ERROR)
    {
        throw OpenGLException(get_log_msg("glUseProgram()"), err);
    }

    // We release the memory for the compiler when the first program is installed. The idea is to
    // compile all the shaders up-front and, once they are compiled, to use them in programs.
    if (current_prog_ == 0 && Shader::supports_compiled_shaders())
    {
        // This should never fail. If it does, it's not significant because
        // glReleaseShaderCompiler() is only a hint and may be a no-op in the implementation.
        // In case there was a failure, we clear the error status so we don't incorrectly
        // conclude later that something else failed when, in fact, it didn't.
        glReleaseShaderCompiler();
        clear_gl_error();
    }

    current_prog_ = id_.get();
}

bool
ProgramImpl::
in_use() const noexcept
{
    return id_.get() == current_prog_;
}

GLuint
ProgramImpl::
current_id() noexcept
{
    return current_prog_;
}

//
// Return a reference to the uniform with the given name. If no such variable exists, throw a LogicException.
// The caller is expected to use a dynamic_cast to down-cast the reference to the correct concrete type.
//

UniformVariable&
ProgramImpl::
uniform(const string& name) const
{
    if (!initialized_)
    {
        init_uniforms();
        initialized_ = true;
    }

    auto u = uniforms_.find(name);
    if (u != uniforms_.end())
    {
        return *u->second;
    }

    // If the variable name isn't in the map, it might still be valid because, for arrays, it is possible
    // retrieve a specific array element, such as "my_array[1]". So, we try the name as is. If already present,
    // we return the variable for the name. If not already present, we try and locate the variable and, if found,
    // add an entry for it.

    const auto e = element_uniforms_.find(name);
    if (e != element_uniforms_.end())
    {
        return *e->second;
    }

    GLint loc = glGetUniformLocation(id_.get(), name.c_str());
    if (loc == -1)
    {
        throw_no_such_variable(name);
    }

    // Variable was found. It must have a name ending in "[<digits>]", otherwise it would have been found
    // in the uniforms_ map. Split off the substring preceding the subscript and look for that in the map,
    // so we can get the element type.

    string::size_type pos = name.find('[');
    if (pos == string::npos)
    {
        // This should be impossible, according to the GLES 2.0 specification.
        // We check anyway, just to be safe.
        throw_no_such_variable(name);
    }
    string stem = name.substr(0, pos);
    u = uniforms_.find(stem);
    if (u == uniforms_.end())
    {
        // Again, should be impossible, but we check in case some implementations have a broken parser.
        throw_no_such_variable(name);
    }
    GLenum type = u->second->type();
    return *element_uniforms_.insert(make_pair(name, make_variable(loc, name, type, 1))).first->second;
}

//
// Return a const reference to the uniform map.
//

const Program::UniformMap&
ProgramImpl::
uniforms() const
{
    if (!initialized_)
    {
        init_uniforms();
        initialized_ = true;
    }
    return uniforms_;
}

//
// This helper method creates a C++ instance that acts as a proxy for a uniform variable in a shader.
// All the proxy classes derive from UniformVariable. Each derived proxy class provides a type-safe
// API for a uniform variable of a particular type. Uniform variables can either be single values
// or arrays. This means there are two proxy classes for each possible uniform type, one to provide
// the API for a single value, and one to provide the API for an array of values (using a vector).
//

unique_ptr<UniformVariable>
ProgramImpl::
make_variable(GLint location, const string& name, GLenum type, GLint size) const
{
    UniformVariable* p;
    switch (type)
    {
        case GL_FLOAT:
        {
            if (size == 1)
            {
                p = new FloatUniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_FLOAT_VEC2:
        {
            if (size == 1)
            {
                p = new FloatVec2Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatVec2ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_FLOAT_VEC3:
        {
            if (size == 1)
            {
                p = new FloatVec3Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatVec3ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_FLOAT_VEC4:
        {
            if (size == 1)
            {
                p = new FloatVec4Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatVec4ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_FLOAT_MAT2:
        {
            if (size == 1)
            {
                p = new FloatMat2Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatMat2ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_FLOAT_MAT3:
        {
            if (size == 1)
            {
                p = new FloatMat3Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatMat3ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_FLOAT_MAT4:
        {
            if (size == 1)
            {
                p = new FloatMat4Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new FloatMat4ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_INT:
        {
            if (size == 1)
            {
                p = new IntUniform(id_.get(), location, name, size);
            }
            else
            {
                p = new IntArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_INT_VEC2:
        {
            if (size == 1)
            {
                p = new IntVec2Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new IntVec2ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_INT_VEC3:
        {
            if (size == 1)
            {
                p = new IntVec3Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new IntVec3ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_INT_VEC4:
        {
            if (size == 1)
            {
                p = new IntVec4Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new IntVec4ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_BOOL:
        {
            if (size == 1)
            {
                p = new BoolUniform(id_.get(), location, name, size);
            }
            else
            {
                p = new BoolArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_BOOL_VEC2:
        {
            if (size == 1)
            {
                p = new BoolVec2Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new BoolVec2ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_BOOL_VEC3:
        {
            if (size == 1)
            {
                p = new BoolVec3Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new BoolVec3ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_BOOL_VEC4:
        {
            if (size == 1)
            {
                p = new BoolVec4Uniform(id_.get(), location, name, size);
            }
            else
            {
                p = new BoolVec4ArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_SAMPLER_2D:
        {
            if (size == 1)
            {
                p = new Sampler2DUniform(id_.get(), location, name, size);
            }
            else
            {
                p = new Sampler2DArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        case GL_SAMPLER_CUBE:
        {
            if (size == 1)
            {
                p = new SamplerCubeUniform(id_.get(), location, name, size);
            }
            else
            {
                p = new SamplerCubeArrayUniform(id_.get(), location, name, size);
            }
            break;
        }
        default:
        {
            assert(false);
        }
    }
    return unique_ptr<UniformVariable>(p);
}

void
ProgramImpl::
throw_no_such_variable(const string& name) const
{
    ostringstream msg;
    msg << "no uniform variable \"" << name << "\" found for program id " << id();
    throw LogicException(msg.str());
}

string
ProgramImpl::
get_log_msg(const string& caller) const
{
    GLint log_size;
    glGetProgramiv(id_.get(), GL_INFO_LOG_LENGTH, &log_size);

    // If we cannot get more detailed information
    // about the error, we just indicate that the call didn't work.
    if (clear_gl_error() != GL_NO_ERROR || log_size == 0)
    {
        return caller + " failed";
    }

    unique_ptr<char[]> log_msg(new char[log_size]);
    glGetProgramInfoLog(id_.get(), log_size, nullptr, log_msg.get());
    if (clear_gl_error() != GL_NO_ERROR)
    {
        return caller + " failed";
    }

    string msg(caller + " failed: ");
    msg += log_msg.get();
    return msg;
}

void
ProgramImpl::
detach_shader(GLuint shader_id) noexcept
{
    // The only way this could fail is if the program ID or shader ID is incorrect, or if the shader wasn't
    // attached to begin with.
    glDetachShader(id_.get(), shader_id);
    clear_gl_error();
}

//
// Get all the active uniforms for this program and initialize the variable map.
//

void
ProgramImpl::
init_uniforms() const
{
    GLint num_uniforms;
    glGetProgramiv(id_.get(), GL_ACTIVE_UNIFORMS, &num_uniforms);
    check_gl("cannot GL_ACTIVE_UNIFORMS: glGetProgramiv() failed");
    if (num_uniforms == 0)
    {
        return;
    }

    GLint max_len;
    glGetProgramiv(id_.get(), GL_ACTIVE_UNIFORM_MAX_LENGTH, &max_len);
    check_gl("cannot read GL_ACTIVE_UNIFORM_MAX_LENGTH: glGetProgramiv() failed");
    if (max_len == 0)
    {
        return; // Could be an assert(false), but we don't assert on things we can't control.
    }

    unique_ptr<char[]> buf(new char[max_len]);
    for (int i = 0; i < num_uniforms; ++i)
    {
        GLint size;
        GLenum type;
        glGetActiveUniform(id_.get(), i, max_len, nullptr, &size, &type, buf.get());
        check_gl("cannot get uniform with with index " + to_string(i) + ": glGetActiveUniform() failed");
        GLint loc = glGetUniformLocation(id_.get(), buf.get());
        if (loc == -1)
        {
            throw_no_such_variable(buf.get()); // Should never happen, unless the implementation is broken.
        }
        uniforms_.insert(make_pair(buf.get(), make_variable(loc, buf.get(), type, size)));
    }
}

} // namespace internal

} // namespace glint
