/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/internal/RuntimeImpl.h>
#include <glint/Runtime.h>
#include <glint/GlintExceptions.h>
#include <glint/GLUtil.h>

#include <sstream>

using namespace std;

namespace glint
{

namespace internal
{

RuntimeImpl::
RuntimeImpl()
    : display_created_(ATOMIC_FLAG_INIT)
    , display_(nullptr)
    , finalized_(ATOMIC_FLAG_INIT)
    , genv_(glint::genv::internal::create_concrete_genv())
{
}

RuntimeImpl::
~RuntimeImpl() noexcept
{
    try
    {
        shutdown();
    }
    catch (...)
    {
        return;
    }
}

void
RuntimeImpl::
shutdown()
{
    if (!finalized_.test_and_set())
    {
        try
        {
            genv_->shutdown();
        }
        catch (...)
        {
            ShutdownException e("Shutdown failure");
            e.remember(current_exception());
            throw e;
        }
    }
}

int
RuntimeImpl::
max_display_width() const noexcept
{
    return genv_->max_display_width();
}

int
RuntimeImpl::
max_display_height() const noexcept
{
    return genv_->max_display_height();
}

GraphicsDisplay*
RuntimeImpl::
create_display()
{
    return create_display(max_display_width(), max_display_height(), 0, 0);
}

GraphicsDisplay*
RuntimeImpl::
create_display(int width, int height, int xpos, int ypos)
{
    if (width <= 0)
    {
        string reason = "Invalid window width: " + to_string(width) + " (width must be greater than zero)";
        throw InvalidArgumentException(reason);
    }
    if (height <= 0)
    {
        string reason = "Invalid window height: " + to_string(height) + " (height must be greater than zero)";
        throw InvalidArgumentException(reason);
    }
    if (xpos < 0)
    {
        string reason = "Invalid x position: " + to_string(xpos) + " (x position cannot be negative)";
        throw InvalidArgumentException(reason);
    }
    if (ypos < 0)
    {
        string reason = "Invalid y position: " + to_string(ypos) + " (y position cannot be negative)";
        throw InvalidArgumentException(reason);
    }
    if (xpos + width > genv_->max_display_width())
    {
        ostringstream reason;
        reason << "Invalid x position: " << to_string(xpos) << " (x position plus width (" << width
               << ") must not exceed display width (" << genv_->max_display_width() << "))";
        throw InvalidArgumentException(reason.str());
    }
    if (ypos + height > genv_->max_display_height())
    {
        ostringstream reason;
        reason << "Invalid y position: " << to_string(ypos) << " (y position plus height (" << height
               << ") must not exceed display height (" << genv_->max_display_height() << "))";
        throw InvalidArgumentException(reason.str());
    }

    if (!display_created_.test_and_set())
    {
        try
        {
            display_ = genv_->create_display(width, height, xpos, ypos);
            init_GL();
            return display_;
        }
        catch (const glint::Exception& e)
        {
            display_created_.clear();
            throw;
        }
    }
    else
    {
        throw LogicException("Cannot create more than one display");
    }
}

GraphicsDisplay*
RuntimeImpl::
display() const
{
    if (!display_created_.test_and_set())
    {
        display_created_.clear();
        throw LogicException("Cannot access display before a display was created with create_display()");
    }
    return display_;
}

void
RuntimeImpl::
init_GL()
{
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);   // Black
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glViewport(0, 0, display_->width(), display_->height());
    check_gl("Cannot clear buffers and set viewport");
}

} // namespace internal

} // namespace glint
