/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/internal/ShaderImpl.h>
#include <glint/GLUtil.h>
#include <glint/GlintExceptions.h>
#include <glint/OpenGL.h>

using namespace std;

namespace glint
{

namespace internal
{

bool
ShaderImpl::
supports_compiled_shaders()
{
    if (gl_environment == GLEnv::OpenGL)
    {
        return true;    // OpenGL provides shader compilation from 2.0 onwards
    }

    // GLES may not support compiled shaders in some implementations.
    GLboolean b;
    glGetBooleanv(GL_SHADER_COMPILER, &b);
    check_gl("cannot read GL_SHADER_COMPILER");
    return b;
}

namespace
{

void
check_compiler()
{
    if (!ShaderImpl::supports_compiled_shaders())
    {
        throw LogicException("shader compilation is not supported by this implementation");
    }
}

void
compile_source(const string& source, GLuint id)
{
    check_compiler();

    const char* s = source.c_str();
    glShaderSource(id, 1, &s, nullptr);
    check_gl("cannot attach source to vertex shader");

    glCompileShader(id);

    GLint success;
    glGetShaderiv(id, GL_COMPILE_STATUS, &success);
    check_gl("cannot retrieve GL_COMPILE_STATUS");
    if (success)
    {
        return;
    }

    GLint log_size;
    glGetShaderiv(id, GL_INFO_LOG_LENGTH, &log_size);

    // GL_INFO_LOG_LENGTH is optional. If we cannot get more detailed information
    // about the error, we just indicate that the compilation didn't work.
    if (clear_gl_error() != GL_NO_ERROR || log_size == 0)
    {
        throw LogicException("shader compilation failed");
    }

    unique_ptr<char[]> log_msg(new char[log_size]);
    glGetShaderInfoLog(id, log_size, nullptr, log_msg.get());

    string msg("shader compilation failed: ");
    msg += log_msg.get();
    throw LogicException(msg);
}

void
delete_shader(GLuint shader_id) noexcept
{
    // This can't fail unless the implementation is broken, or someone has deleted the shader behind our back.
    glDeleteShader(shader_id);
    clear_gl_error();
}

} // namespace

VertexShaderImpl::
VertexShaderImpl(const string& source)
    : id_(&delete_shader)
{
    GLuint id = glCreateShader(GL_VERTEX_SHADER);
    if (id == 0)
    {
        throw OpenGLException("cannot create vertex shader", GL_NO_ERROR);
    }
    id_.reset(id);
    compile_source(source, id);
}

FragmentShaderImpl::
FragmentShaderImpl(const string& source)
    : id_(&delete_shader)
{
    GLuint id = glCreateShader(GL_FRAGMENT_SHADER);
    if (id == 0)
    {
        throw OpenGLException("cannot create fragment shader", GL_NO_ERROR);
    }
    id_.reset(id);
    compile_source(source, id_.get());
}

} // namespace internal

} // namespace glint
