/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/internal/ShaderVariableImpl.h>
#include <glint/internal/ProgramImpl.h>
#include <glint/GLUtil.h>
#include <glint/GlintExceptions.h>

#include <cassert>
#include <sstream>

using namespace std;

namespace glint
{

namespace internal
{

ShaderVariableImpl::
ShaderVariableImpl(GLuint prog_id, GLint location, const string& name, GLenum type, GLint size) noexcept
    : prog_id_(prog_id), location_(location), name_(name), type_(type), size_(size)
{
    assert(prog_id != 0);
    assert(location >= 0);
    assert(!name.empty());
    assert(size > 0);
    assert(type == GL_FLOAT || type == GL_FLOAT_VEC2 || type == GL_FLOAT_VEC3 || type == GL_FLOAT_VEC4 ||
           type == GL_INT   || type == GL_INT_VEC2   || type == GL_INT_VEC3   || type == GL_INT_VEC4   ||
           type == GL_BOOL  || type == GL_BOOL_VEC2  || type == GL_BOOL_VEC3  || type == GL_BOOL_VEC4  ||
           type == GL_FLOAT_MAT2 || type == GL_FLOAT_MAT3 || type == GL_FLOAT_MAT4 ||
           type == GL_SAMPLER_2D || type == GL_SAMPLER_CUBE);
}

ShaderVariableImpl::
~ShaderVariableImpl() = default;

void
ShaderVariableImpl::
check(const char* op) const
{
    GLenum err = clear_gl_error();
    if (err != GL_NO_ERROR)
    {
        ostringstream msg;
        msg << "cannot " << op << " uniform variable \"" << name_ << "\" at location " << to_string(location_)
            << "; size = " << to_string(size_) << ", program id = " << to_string(prog_id_);
        if (ProgramImpl::current_id() != prog_id_)
        {
            // Setting a variable in a program that is not in use is a logic error,
            // instead of an unexpected failure.
            msg << ": program is not currently in use";
            throw LogicException(msg.str());
        }
        throw OpenGLException(msg.str(), err);
    }
}

void
ShaderVariableImpl::
check_get() const
{
    check("get");
}

void
ShaderVariableImpl::
check_set() const
{
    check("set");
}

void
ShaderVariableImpl::
check_size(size_t vector_size) const
{
    if (vector_size != static_cast<size_t>(size()))
    {
        ostringstream msg;
        msg << "cannot set array uniform variable \"" << name_ << "\" at location " << to_string(location_)
            << " for program id " << to_string(prog_id_) << ": expected size is " << size_ << " elements";
        msg << " but vector contains " << vector_size << " element";
        if (vector_size != 1)
        {
            msg << "s";
        }
        throw LogicException(msg.str());
    }
}

} // namespace internal

} // namespace glint
