/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/internal/UniformVariableImpl.h>

#include <algorithm>
#include <cassert>

using namespace std;

namespace glint
{

namespace internal
{

// UniformVariable

UniformVariableImpl::
UniformVariableImpl(GLuint prog_id, GLint location, const string& name, GLenum type, GLint size)
    : ShaderVariableImpl(prog_id, location, name, type, size)
{
}

UniformVariableImpl::
~UniformVariableImpl() noexcept
{
}

// IntUniformVariableBase

IntUniformVariableBaseImpl::
IntUniformVariableBaseImpl(GLuint prog_id, GLint location, const string& name, GLenum type, GLint size)
    : UniformVariableImpl(prog_id, location, name, type, size)
{
}

IntUniformVariableBaseImpl::
~IntUniformVariableBaseImpl() noexcept
{
}

GLint
IntUniformVariableBaseImpl::
get() const
{
    GLint v;
    glGetUniformiv(program_id(), location(), &v);
    check_get();
    return v;
}

void
IntUniformVariableBaseImpl::
set(GLint v)
{
    glUniform1i(location(), v);
    check_set();
}

// IntArrayUniformVariableBase

IntArrayUniformVariableBaseImpl::
IntArrayUniformVariableBaseImpl(GLuint prog_id, GLint location, const string& name, GLenum type, GLint size)
    : UniformVariableImpl(prog_id, location, name, type, size)
{
}

IntArrayUniformVariableBaseImpl::
~IntArrayUniformVariableBaseImpl() noexcept
{
}

vector<GLint>
IntArrayUniformVariableBaseImpl::
get() const
{
    IntArray v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &v[i]);
    }
    check_get();
    return v;
}

void
IntArrayUniformVariableBaseImpl::
set(const vector<GLint>& v)
{
    check_size(v.size());
    glUniform1iv(location(), size(), &v[0]);
    check_set();
}

// Float

FloatUniformImpl::
FloatUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT, size)
{
}

FloatUniformImpl::
~FloatUniformImpl() noexcept
{
}

GLfloat
FloatUniformImpl::
get() const
{
    GLfloat v;
    glGetUniformfv(program_id(), location(), &v);
    check_get();
    return v;
}

void
FloatUniformImpl::
set(GLfloat v)
{
    glUniform1f(location(), v);
    check_set();
}

// FloatVec2

FloatVec2UniformImpl::
FloatVec2UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_VEC2, size)
{
}

FloatVec2UniformImpl::
~FloatVec2UniformImpl() noexcept
{
}

FloatVec2
FloatVec2UniformImpl::
get() const
{
    FloatVec2 v;
    glGetUniformfv(program_id(), location(), &v[0]);
    check_get();
    return v;
}

void
FloatVec2UniformImpl::
set(const FloatVec2& v)
{
    glUniform2fv(location(), size(), &v[0]);
    check_set();
}

// FloatVec3

FloatVec3UniformImpl::
FloatVec3UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_VEC3, size)
{
}

FloatVec3UniformImpl::
~FloatVec3UniformImpl() noexcept
{
}

FloatVec3
FloatVec3UniformImpl::
get() const
{
    FloatVec3 v;
    glGetUniformfv(program_id(), location(), &v[0]);
    check_get();
    return v;
}

void
FloatVec3UniformImpl::
set(const FloatVec3& v)
{
    glUniform3fv(location(), size(), &v[0]);
    check_set();
}

// FloatVec4

FloatVec4UniformImpl::
FloatVec4UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_VEC4, size)
{
}

FloatVec4UniformImpl::
~FloatVec4UniformImpl() noexcept
{
}

FloatVec4
FloatVec4UniformImpl::
get() const
{
    FloatVec4 v;
    glGetUniformfv(program_id(), location(), &v[0]);
    check_get();
    return v;
}

void
FloatVec4UniformImpl::
set(const FloatVec4& v)
{
    glUniform4fv(location(), size(), &v[0]);
    check_set();
}

// FloatMat2

FloatMat2UniformImpl::
FloatMat2UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_MAT2, size)
{
}

FloatMat2UniformImpl::
~FloatMat2UniformImpl() noexcept
{
}

FloatMat2
FloatMat2UniformImpl::
get() const
{
    FloatMat2 v;
    glGetUniformfv(program_id(), location(), &v[0][0]);
    check_get();
    return v;
}

void
FloatMat2UniformImpl::
set(const FloatMat2& v)
{
    glUniformMatrix2fv(location(), size(), GL_FALSE, &v[0][0]);
    check_set();
}

// FloatMat3

FloatMat3UniformImpl::
FloatMat3UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_MAT3, size)
{
}

FloatMat3UniformImpl::
~FloatMat3UniformImpl() noexcept
{
}

FloatMat3
FloatMat3UniformImpl::
get() const
{
    FloatMat3 v;
    glGetUniformfv(program_id(), location(), &v[0][0]);
    check_get();
    return v;
}

void
FloatMat3UniformImpl::
set(const FloatMat3& v)
{
    glUniformMatrix3fv(location(), size(), GL_FALSE, &v[0][0]);
    check_set();
}

// FloatMat4

FloatMat4UniformImpl::
FloatMat4UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_MAT4, size)
{
}

FloatMat4UniformImpl::
~FloatMat4UniformImpl() noexcept
{
}

FloatMat4
FloatMat4UniformImpl::
get() const
{
    FloatMat4 v;
    glGetUniformfv(program_id(), location(), &v[0][0]);
    check_get();
    return v;
}

void
FloatMat4UniformImpl::
set(const FloatMat4& v)
{
    glUniformMatrix4fv(location(), size(), GL_FALSE, &v[0][0]);
    check_set();
}

// FloatArray

FloatArrayUniformImpl::
FloatArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT, size)
{
}

FloatArrayUniformImpl::
~FloatArrayUniformImpl() noexcept
{
}

FloatArray
FloatArrayUniformImpl::
get() const
{
    FloatArray v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i]);
    }
    check_get();
    return v;
}

void
FloatArrayUniformImpl::
set(const FloatArray& v)
{
    check_size(v.size());
    glUniform1fv(location(), size(), &v[0]);
    check_set();
}

// FloatVec2Array

FloatVec2ArrayUniformImpl::
FloatVec2ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_VEC2, size)
{
}

FloatVec2ArrayUniformImpl::
~FloatVec2ArrayUniformImpl() noexcept
{
}

FloatVec2Array
FloatVec2ArrayUniformImpl::
get() const
{
    FloatVec2Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i][0]);
    }
    check_get();
    return v;
}

void
FloatVec2ArrayUniformImpl::
set(const FloatVec2Array& v)
{
    check_size(v.size());
    glUniform2fv(location(), size(), &v[0][0]);
    check_set();
}

// FloatVec3Array

FloatVec3ArrayUniformImpl::
FloatVec3ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_VEC3, size)
{
}

FloatVec3ArrayUniformImpl::
~FloatVec3ArrayUniformImpl() noexcept
{
}

FloatVec3Array
FloatVec3ArrayUniformImpl::
get() const
{
    FloatVec3Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i][0]);
    }
    check_get();
    return v;
}

void
FloatVec3ArrayUniformImpl::
set(const FloatVec3Array& v)
{
    check_size(v.size());
    glUniform3fv(location(), size(), &v[0][0]);
    check_set();
}

// FloatVec4Array

FloatVec4ArrayUniformImpl::
FloatVec4ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_VEC4, size)
{
}

FloatVec4ArrayUniformImpl::
~FloatVec4ArrayUniformImpl() noexcept
{
}

FloatVec4Array
FloatVec4ArrayUniformImpl::
get() const
{
    FloatVec4Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i][0]);
    }
    check_get();
    return v;
}

void
FloatVec4ArrayUniformImpl::
set(const FloatVec4Array& v)
{
    check_size(v.size());
    glUniform4fv(location(), size(), &v[0][0]);
    check_set();
}

// FloatMat2Array

FloatMat2ArrayUniformImpl::
FloatMat2ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_MAT2, size)
{
}

FloatMat2ArrayUniformImpl::
~FloatMat2ArrayUniformImpl() noexcept
{
}

FloatMat2Array
FloatMat2ArrayUniformImpl::
get() const
{
    FloatMat2Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i][0][0]);
    }
    check_get();
    return v;
}

void
FloatMat2ArrayUniformImpl::
set(const FloatMat2Array& v)
{
    check_size(v.size());
    glUniformMatrix2fv(location(), size(), GL_FALSE, &v[0][0][0]);
    check_set();
}

// FloatMat3Array

FloatMat3ArrayUniformImpl::
FloatMat3ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_MAT3, size)
{
}

FloatMat3ArrayUniformImpl::
~FloatMat3ArrayUniformImpl() noexcept
{
}

FloatMat3Array
FloatMat3ArrayUniformImpl::
get() const
{
    FloatMat3Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i][0][0]);
    }
    check_get();
    return v;
}

void
FloatMat3ArrayUniformImpl::
set(const FloatMat3Array& v)
{
    check_size(v.size());
    glUniformMatrix3fv(location(), size(), GL_FALSE, &v[0][0][0]);
    check_set();
}

// FloatMat4Array

FloatMat4ArrayUniformImpl::
FloatMat4ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_FLOAT_MAT4, size)
{
}

FloatMat4ArrayUniformImpl::
~FloatMat4ArrayUniformImpl() noexcept
{
}

FloatMat4Array
FloatMat4ArrayUniformImpl::
get() const
{
    FloatMat4Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformfv(program_id(), location() + i, &v[i][0][0]);
    }
    check_get();
    return v;
}

void
FloatMat4ArrayUniformImpl::
set(const FloatMat4Array& v)
{
    check_size(v.size());
    glUniformMatrix4fv(location(), size(), GL_FALSE, &v[0][0][0]);
    check_set();
}

// Int

IntUniformImpl::
IntUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : IntUniformVariableBaseImpl(prog_id, location, name, GL_INT, size)
{
}

IntUniformImpl::
~IntUniformImpl() noexcept
{
}

// IntVec2

IntVec2UniformImpl::
IntVec2UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_INT_VEC2, size)
{
}

IntVec2UniformImpl::
~IntVec2UniformImpl() noexcept
{
}

IntVec2
IntVec2UniformImpl::
get() const
{
    IntVec2 v;
    glGetUniformiv(program_id(), location(), &v[0]);
    check_get();
    return v;
}

void
IntVec2UniformImpl::
set(const IntVec2& v)
{
    glUniform2iv(location(), size(), &v[0]);
    check_set();
}

// IntVec3

IntVec3UniformImpl::
IntVec3UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_INT_VEC3, size)
{
}

IntVec3UniformImpl::
~IntVec3UniformImpl() noexcept
{
}

IntVec3
IntVec3UniformImpl::
get() const
{
    IntVec3 v;
    glGetUniformiv(program_id(), location(), &v[0]);
    check_get();
    return v;
}

void
IntVec3UniformImpl::
set(const IntVec3& v)
{
    glUniform3iv(location(), size(), &v[0]);
    check_set();
}

// IntVec4

IntVec4UniformImpl::
IntVec4UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_INT_VEC4, size)
{
}

IntVec4UniformImpl::
~IntVec4UniformImpl() noexcept
{
}

IntVec4
IntVec4UniformImpl::
get() const
{
    IntVec4 v;
    glGetUniformiv(program_id(), location(), &v[0]);
    check_get();
    return v;
}

void
IntVec4UniformImpl::
set(const IntVec4& v)
{
    glUniform4iv(location(), size(), &v[0]);
    check_set();
}

// IntArray

IntArrayUniformImpl::
IntArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : IntArrayUniformVariableBaseImpl(prog_id, location, name, GL_INT, size)
{
}

IntArrayUniformImpl::
~IntArrayUniformImpl() noexcept
{
}

// IntVec2Array

IntVec2ArrayUniformImpl::
IntVec2ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_INT_VEC2, size)
{
}

IntVec2ArrayUniformImpl::
~IntVec2ArrayUniformImpl() noexcept
{
}

IntVec2Array
IntVec2ArrayUniformImpl::
get() const
{
    IntVec2Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &v[i][0]);
    }
    check_get();
    return v;
}

void
IntVec2ArrayUniformImpl::
set(const IntVec2Array& v)
{
    check_size(v.size());
    glUniform2iv(location(), size(), &v[0][0]);
    check_set();
}

// IntVec3Array

IntVec3ArrayUniformImpl::
IntVec3ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_INT_VEC3, size)
{
}

IntVec3ArrayUniformImpl::
~IntVec3ArrayUniformImpl() noexcept
{
}

IntVec3Array
IntVec3ArrayUniformImpl::
get() const
{
    IntVec3Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &v[i][0]);
    }
    check_get();
    return v;
}

void
IntVec3ArrayUniformImpl::
set(const IntVec3Array& v)
{
    check_size(v.size());
    glUniform3iv(location(), size(), &v[0][0]);
    check_set();
}

// IntVec4Array

IntVec4ArrayUniformImpl::
IntVec4ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_INT_VEC4, size)
{
}

IntVec4ArrayUniformImpl::
~IntVec4ArrayUniformImpl() noexcept
{
}

IntVec4Array
IntVec4ArrayUniformImpl::
get() const
{
    IntVec4Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &v[i][0]);
    }
    check_get();
    return v;
}

void
IntVec4ArrayUniformImpl::
set(const IntVec4Array& v)
{
    check_size(v.size());
    glUniform4iv(location(), size(), &v[0][0]);
    check_set();
}

// Bool

// Bools are GLboolean (usually unsigned char) in GLES, but are integers in GLSL. We hide the difference
// here at the cost of having to copy the values, so the caller can deal uniformly with things as GLboolean.

BoolUniformImpl::
BoolUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL, size)
{
}

BoolUniformImpl::
~BoolUniformImpl() noexcept
{
}

GLboolean
BoolUniformImpl::
get() const
{
    GLint v;
    glGetUniformiv(program_id(), location(), &v);
    check_get();
    return v;
}

void
BoolUniformImpl::
set(GLboolean v)
{
    glUniform1i(location(), v);
    check_set();
}

// BoolVec2

BoolVec2UniformImpl::
BoolVec2UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL_VEC2, size)
{
}

BoolVec2UniformImpl::
~BoolVec2UniformImpl() noexcept
{
}

BoolVec2
BoolVec2UniformImpl::
get() const
{
    IntVec2 i;
    glGetUniformiv(program_id(), location(), &i[0]);
    check_get();
    BoolVec2 v;
    v[0] = i[0];
    v[1] = i[1];
    return v;
}

void
BoolVec2UniformImpl::
set(const BoolVec2& v)
{
    IntVec2 i;
    i[0] = v[0];
    i[1] = v[1];
    glUniform2iv(location(), size(), &i[0]);
    check_set();
}

// BoolVec3

BoolVec3UniformImpl::
BoolVec3UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL_VEC3, size)
{
}

BoolVec3UniformImpl::
~BoolVec3UniformImpl() noexcept
{
}

BoolVec3
BoolVec3UniformImpl::
get() const
{
    IntVec3 i;
    glGetUniformiv(program_id(), location(), &i[0]);
    check_get();
    BoolVec3 v;
    v[0] = i[0];
    v[1] = i[1];
    v[2] = i[2];
    return v;
}

void
BoolVec3UniformImpl::
set(const BoolVec3& v)
{
    IntVec3 i;
    i[0] = v[0];
    i[1] = v[1];
    i[2] = v[2];
    glUniform3iv(location(), size(), &i[0]);
    check_set();
}

// BoolVec4

BoolVec4UniformImpl::
BoolVec4UniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL_VEC4, size)
{
}

BoolVec4UniformImpl::
~BoolVec4UniformImpl() noexcept
{
}

BoolVec4
BoolVec4UniformImpl::
get() const
{
    IntVec4 i;
    glGetUniformiv(program_id(), location(), &i[0]);
    check_get();
    BoolVec4 v;
    v[0] = i[0];
    v[1] = i[1];
    v[2] = i[2];
    v[3] = i[3];
    return v;
}

void
BoolVec4UniformImpl::
set(const BoolVec4& v)
{
    IntVec4 i;
    i[0] = v[0];
    i[1] = v[1];
    i[2] = v[2];
    i[3] = v[3];
    glUniform4iv(location(), size(), &i[0]);
    check_set();
}

// BoolArray

BoolArrayUniformImpl::
BoolArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL, size)
{
}

BoolArrayUniformImpl::
~BoolArrayUniformImpl() noexcept
{
}

BoolArray
BoolArrayUniformImpl::
get() const
{
    IntArray a;
    a.resize(size());
    BoolArray v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &a[i]);
        v[i] = a[i];
    }
    check_get();
    return v;
}

void
BoolArrayUniformImpl::
set(const BoolArray& v)
{
    check_size(v.size());
    IntArray a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        a[i] = v[i];
    }
    glUniform1iv(location(), size(), &a[0]);
    check_set();
}

// BoolVec2Array

BoolVec2ArrayUniformImpl::
BoolVec2ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL_VEC2, size)
{
}

BoolVec2ArrayUniformImpl::
~BoolVec2ArrayUniformImpl() noexcept
{
}

BoolVec2Array
BoolVec2ArrayUniformImpl::
get() const
{
    IntVec2Array a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &a[i][0]);
    }
    check_get();
    BoolVec2Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        v[i][0] = a[i][0];
        v[i][1] = a[i][1];
    }
    return v;
}

void
BoolVec2ArrayUniformImpl::
set(const BoolVec2Array& v)
{
    check_size(v.size());
    IntVec2Array a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        a[i][0] = v[i][0];
        a[i][1] = v[i][1];
    }
    glUniform2iv(location(), size(), &a[0][0]);
    check_set();
}

// BoolVec3Array

BoolVec3ArrayUniformImpl::
BoolVec3ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL_VEC3, size)
{
}

BoolVec3ArrayUniformImpl::
~BoolVec3ArrayUniformImpl() noexcept
{
}

BoolVec3Array
BoolVec3ArrayUniformImpl::
get() const
{
    IntVec3Array a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &a[i][0]);
    }
    check_get();
    BoolVec3Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        v[i][0] = a[i][0];
        v[i][1] = a[i][1];
        v[i][2] = a[i][2];
    }
    return v;
}

void
BoolVec3ArrayUniformImpl::
set(const BoolVec3Array& v)
{
    check_size(v.size());
    IntVec3Array a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        a[i][0] = v[i][0];
        a[i][1] = v[i][1];
        a[i][2] = v[i][2];
    }
    glUniform3iv(location(), size(), &a[0][0]);
    check_set();
}

// BoolVec4Array

BoolVec4ArrayUniformImpl::
BoolVec4ArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : UniformVariableImpl(prog_id, location, name, GL_BOOL_VEC4, size)
{
}

BoolVec4ArrayUniformImpl::
~BoolVec4ArrayUniformImpl() noexcept
{
}

BoolVec4Array
BoolVec4ArrayUniformImpl::
get() const
{
    IntVec4Array a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        glGetUniformiv(program_id(), location() + i, &a[i][0]);
    }
    check_get();
    BoolVec4Array v;
    v.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        v[i][0] = a[i][0];
        v[i][1] = a[i][1];
        v[i][2] = a[i][2];
        v[i][3] = a[i][3];
    }
    return v;
}

void
BoolVec4ArrayUniformImpl::
set(const BoolVec4Array& v)
{
    check_size(v.size());
    IntVec4Array a;
    a.resize(size());
    for (int i = 0; i < size(); ++i)
    {
        a[i][0] = v[i][0];
        a[i][1] = v[i][1];
        a[i][2] = v[i][2];
        a[i][3] = v[i][3];
    }
    glUniform4iv(location(), size(), &a[0][0]);
    check_set();
}

// Sampler2D

Sampler2DUniformImpl::
Sampler2DUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : IntUniformVariableBaseImpl(prog_id, location, name, GL_SAMPLER_2D, size)
{
}

Sampler2DUniformImpl::
~Sampler2DUniformImpl() noexcept
{
}

// Sampler2DArray

Sampler2DArrayUniformImpl::
Sampler2DArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : IntArrayUniformVariableBaseImpl(prog_id, location, name, GL_SAMPLER_2D, size)
{
}

Sampler2DArrayUniformImpl::
~Sampler2DArrayUniformImpl() noexcept
{
}

// SamplerCube

SamplerCubeUniformImpl::
SamplerCubeUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : IntUniformVariableBaseImpl(prog_id, location, name, GL_SAMPLER_CUBE, size)
{
}

SamplerCubeUniformImpl::
~SamplerCubeUniformImpl() noexcept
{
}

// SamplerCubeArray

SamplerCubeArrayUniformImpl::
SamplerCubeArrayUniformImpl(GLuint prog_id, GLint location, const string& name, GLint size)
    : IntArrayUniformVariableBaseImpl(prog_id, location, name, GL_SAMPLER_CUBE, size)
{
}

SamplerCubeArrayUniformImpl::
~SamplerCubeArrayUniformImpl() noexcept
{
}

} // namespace internal

} // namespace glint
