/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Color.h>
#include <glint/GlintExceptions.h>
#include <gtest/gtest.h>

#include <sstream>
#include <vector>
#include <unordered_map>

using namespace glint;
using namespace std;

//
// When we set a color component from a float, the Color implementation
// rounds to the closest int. We use this function to test that, after setting
// a value from a float, the float that is returned is within 1/2 of 1/255.0f,
// that is, was mapped to the correct integer value.
//

bool
in_range(float expected_val, float set_val)
{
    const float half_range = 1 / 255.0f / 2;
    float low = set_val - half_range;
    float high = set_val + half_range;
    return expected_val >= low && expected_val < high;
}

TEST(Color, constructors)
{
    {
        Color c;

        EXPECT_EQ(0, c.red());
        EXPECT_EQ(0, c.green());
        EXPECT_EQ(0, c.blue());
        EXPECT_EQ(255, c.alpha());

        EXPECT_EQ(0.0f, c.red_f());
        EXPECT_EQ(0.0f, c.green_f());
        EXPECT_EQ(0.0f, c.blue_f());
        EXPECT_EQ(1.0f, c.alpha_f());
    }

    for (int i = 0; i < 256; ++i)
    {
        Color c(i, i, i, i);
        EXPECT_EQ(i, c.red());
        EXPECT_EQ(i / 255.0f, c.red_f());
        EXPECT_EQ(i, c.green());
        EXPECT_EQ(i / 255.0f, c.green_f());
        EXPECT_EQ(i, c.blue());
        EXPECT_EQ(i / 255.0f, c.blue_f());
        EXPECT_EQ(i, c.alpha());
        EXPECT_EQ(i / 255.0f, c.alpha_f());
    }

    for (int i = 0; i < 256; ++i)
    {
        float f = i / 255.0f;
        Color c(f, f, f, f);
        EXPECT_EQ(i, c.red());
        EXPECT_EQ(i / 255.0f, c.red_f());
        EXPECT_EQ(i, c.green());
        EXPECT_EQ(i / 255.0f, c.red_f());
        EXPECT_EQ(i, c.blue());
        EXPECT_EQ(i / 255.0f, c.blue_f());
        EXPECT_EQ(i, c.alpha());
        EXPECT_EQ(i / 255.0f, c.alpha_f());
    }

    {
        Color c(1, 2, 3);
        EXPECT_EQ(1, c.red());
        EXPECT_EQ(2, c.green());
        EXPECT_EQ(3, c.blue());
        EXPECT_EQ(255, c.alpha());
    }

    {
        Color c(0.1f, 0.2f, 0.3f);
        EXPECT_TRUE(in_range(0.1f, c.red_f()));
        EXPECT_TRUE(in_range(0.2f, c.green_f()));
        EXPECT_TRUE(in_range(0.3f, c.blue_f()));
        EXPECT_TRUE(in_range(1.0f, c.alpha_f()));
    }

    {
        uint32_t v = (4 << 24) | (3 << 16) | (2 << 8) | 1;
        Color c(v);
        EXPECT_EQ(3, c.red());
        EXPECT_EQ(2, c.green());
        EXPECT_EQ(1, c.blue());
        EXPECT_EQ(4, c.alpha());
    }

    {
        uint8_t v[4] = { 1, 2, 3, 4 };
        Color c(v);
        EXPECT_EQ(1, c.red());
        EXPECT_EQ(2, c.green());
        EXPECT_EQ(3, c.blue());
        EXPECT_EQ(4, c.alpha());
    }

    {
        float v[4] = { 0.1f, 0.2f, 0.3f, 0.4f };
        Color c(v);
        EXPECT_TRUE(in_range(0.1f, c.red_f()));
        EXPECT_TRUE(in_range(0.2f, c.green_f()));
        EXPECT_TRUE(in_range(0.3f, c.blue_f()));
        EXPECT_TRUE(in_range(0.4f, c.alpha_f()));
    }

    for (int blue = 0; blue <= 0xf; ++blue)
    {
        for (int green = 0; green <= 0xf; ++green)
        {
            for (int red = 0; red <= 0xf; ++red)
            {
                ostringstream s;
                s << "#" << hex << setfill('0') << red << green << blue;
                Color c(s.str());
                EXPECT_EQ((red << 4) | red, c.red());
                EXPECT_EQ((green << 4) | green, c.green());
                EXPECT_EQ((blue << 4) | blue, c.blue());
                EXPECT_EQ(255, c.alpha());
            }
        }
    }

    for (int blue = 0; blue <= 0xf; ++blue)
    {
        for (int green = 0; green <= 0xf; ++green)
        {
            for (int red = 0; red <= 0xf; ++red)
            {
                for (int alpha = 0; alpha <= 0xf; ++alpha)
                {
                    ostringstream s;
                    s << "#" << hex << red << green << blue << alpha;
                    Color c(s.str());
                    EXPECT_EQ((red << 4) | red, c.red());
                    EXPECT_EQ((green << 4) | green, c.green());
                    EXPECT_EQ((blue << 4) | blue, c.blue());
                    EXPECT_EQ((alpha << 4) | alpha, c.alpha());
                }
            }
        }
    }

    // We can't exhaustively test 2^24 color strings, so we check at the boundaries
    // and a few values in between.
    {
        vector<string> v =
        {
            "#000000",
            "#000001", "#0000aa", "#0000ff",
            "#000100", "#00ab00", "#00ff00",
            "#010000", "#ac0000", "#ff0000",
            "#ffffff"
        };
        for (auto col_str : v)
        {
            Color c(col_str);
            int red, green, blue;
            {
                istringstream s(col_str.substr(1, 2));
                s >> hex >> red;
            }
            {
                istringstream s(col_str.substr(3, 2));
                s >> hex >> green;
            }
            {
                istringstream s(col_str.substr(5, 2));
                s >> hex >> blue;
            }
            EXPECT_EQ(red, c.red());
            EXPECT_EQ(green, c.green());
            EXPECT_EQ(blue, c.blue());
            EXPECT_EQ(255, c.alpha());
        }
    }

    // We can't exhaustively test 2^32 color strings, so we check at the boundaries
    // and a few values in between.
    {
        vector<string> v =
        {
            "#00000000",
            "#00000001", "#000000aa", "#000000ff",
            "#00000100", "#0000ab00", "#0000ff00",
            "#00010000", "#00ac0000", "#00ff0000",
            "#01000000", "#ad000000", "#ff000000",
            "#ffffffff"
        };
        for (auto col_str : v)
        {
            Color c(col_str);
            int red, green, blue, alpha;
            {
                istringstream s(col_str.substr(1, 2));
                s >> hex >> red;
            }
            {
                istringstream s(col_str.substr(3, 2));
                s >> hex >> green;
            }
            {
                istringstream s(col_str.substr(5, 2));
                s >> hex >> blue;
            }
            {
                istringstream s(col_str.substr(7, 2));
                s >> hex >> alpha;
            }
            EXPECT_EQ(red, c.red());
            EXPECT_EQ(green, c.green());
            EXPECT_EQ(blue, c.blue());
            EXPECT_EQ(alpha, c.alpha());
        }
    }

    {
        Color c("aliceblue");
        EXPECT_EQ(0xf0, c.red());
        EXPECT_EQ(0xf8, c.green());
        EXPECT_EQ(0xff, c.blue());
        EXPECT_EQ(0xff, c.alpha());
    }

    {
        Color c("ALICEBLUE");
        EXPECT_EQ(0xf0, c.red());
        EXPECT_EQ(0xf8, c.green());
        EXPECT_EQ(0xff, c.blue());
        EXPECT_EQ(0xff, c.alpha());
    }
}

TEST(Color, constructor_exceptions)
{
    try
    {
        Color c(-1, 0, 0 , 0);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range:"
                  " -1 (red)",
                  e.to_string());
    }

    try
    {
        Color c(1, -1, 256 , 300);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0-255] range:"
                  " -1 (green), 256 (blue), 300 (alpha)",
                  e.to_string());
    }

    try
    {
        Color c(-1.0f, 0.0f, 0.0f, 0.0f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range:"
                  " -1.000000 (red)",
                  e.to_string());
    }

    try
    {
        Color c(1.0f, -0.0f, -1.01f, 1.01f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0.0-1.0] range:"
                  " -1.010000 (blue), 1.010000 (alpha)",
                  e.to_string());
    }

    try
    {
        float v[4] = { -0.1f, 1.2f, -0.3f, 1.4f };
        Color c(v);
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0.0-1.0] range:"
                  " -0.100000 (red), 1.200000 (green), -0.300000 (blue), 1.400000 (alpha)",
                  e.to_string());
    }

    try
    {
        Color c("");
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid SVG color name: \"\" (empty string)",
                  e.to_string());
    }

    try
    {
        Color c("#");
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color string (incorrect number of hex digits): \"#\"",
                  e.to_string());
    }

    try
    {
        Color c("#A");
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color string (incorrect number of hex digits): \"#A\"",
                  e.to_string());
    }

    try
    {
        Color c("#ababababA");
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color string (incorrect number of hex digits): \"#ababababA\"",
                  e.to_string());
    }

    try
    {
        Color c("#abagbabab");
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color string (must only contain hex digits): \"#abagbabab\"",
                  e.to_string());
    }

    try
    {
        Color c("Octarine (The Color of Magic)");
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid SVG color name: \"Octarine (The Color of Magic)\"",
                  e.to_string());
    }
}

TEST(Color, mutators)
{
    {
        Color c(1, 2, 3, 4);

        EXPECT_EQ(1, c.set_red(11));
        EXPECT_EQ(11, c.red());
        EXPECT_EQ(2, c.set_green(12));
        EXPECT_EQ(12, c.green());
        EXPECT_EQ(3, c.set_blue(13));
        EXPECT_EQ(13, c.blue());
        EXPECT_EQ(4, c.set_alpha(14));
        EXPECT_EQ(14, c.alpha());

        EXPECT_EQ(11, c.red());
        EXPECT_EQ(12, c.green());
        EXPECT_EQ(13, c.blue());
        EXPECT_EQ(14, c.alpha());
    }

    {
        Color c(0.1f, 0.2f, 0.3f, 0.4f);

        EXPECT_TRUE(in_range(0.1f, c.set_red_f(0.11f)));
        EXPECT_TRUE(in_range(0.11f, c.red_f()));
        EXPECT_TRUE(in_range(0.2f, c.set_green_f(0.12f)));
        EXPECT_TRUE(in_range(0.12f, c.green_f()));
        EXPECT_TRUE(in_range(0.3f, c.set_blue_f(0.13f)));
        EXPECT_TRUE(in_range(0.13f, c.blue_f()));
        EXPECT_TRUE(in_range(0.4f, c.set_alpha_f(0.14f)));
        EXPECT_TRUE(in_range(0.14f, c.alpha_f()));

        EXPECT_TRUE(in_range(0.11f, c.red_f()));
        EXPECT_TRUE(in_range(0.12f, c.green_f()));
        EXPECT_TRUE(in_range(0.13f, c.blue_f()));
        EXPECT_TRUE(in_range(0.14f, c.alpha_f()));

        EXPECT_EQ(28, c.red());
        EXPECT_EQ(31, c.green());
        EXPECT_EQ(33, c.blue());
        EXPECT_EQ(36, c.alpha());
    }

    {
        Color c(0, 0, 0, 0);
        uint32_t v = (1 << 24) | (2 << 16) | (3 << 8) | 4;
        EXPECT_EQ(0, c.set_uint32(v));
        EXPECT_EQ(v, c.uint32());
        EXPECT_EQ(1, c.alpha());
        EXPECT_EQ(2, c.red());
        EXPECT_EQ(3, c.green());
        EXPECT_EQ(4, c.blue());
    }

    {
        Color c(0, 1, 2, 3);
        uint8_t v[4] = { 1, 2, 3, 4 };
        uint8_t old[4];
        c.set_uint8_v(v, old);
        EXPECT_EQ(0, old[0]);
        EXPECT_EQ(1, old[1]);
        EXPECT_EQ(2, old[2]);
        EXPECT_EQ(3, old[3]);

        uint8_t v2[4] = { 5, 6, 7, 8 };
        c.set_uint8_v(v2);
        EXPECT_EQ(5, c.red());
        EXPECT_EQ(6, c.green());
        EXPECT_EQ(7, c.blue());
        EXPECT_EQ(8, c.alpha());

        c.uint8_v(v);
        EXPECT_EQ(5, v[0]);
        EXPECT_EQ(6, v[1]);
        EXPECT_EQ(7, v[2]);
        EXPECT_EQ(8, v[3]);
    }

    {
        Color c(0.0f, 0.1f, 0.2f, 0.3f);
        float v[4] = { 0.1f, 0.2f, 0.3f, 0.4f };
        float old[4];
        c.set_float_v(v, old);
        EXPECT_TRUE(in_range(0.0f, old[0]));
        EXPECT_TRUE(in_range(0.1f, old[1]));
        EXPECT_TRUE(in_range(0.2f, old[2]));
        EXPECT_TRUE(in_range(0.3f, old[3]));

        float v2[4] = { 0.5f, 0.6f, 0.7f, 0.8f };
        c.set_float_v(v2);
        EXPECT_TRUE(in_range(0.5f, c.red_f()));
        EXPECT_TRUE(in_range(0.6f, c.green_f()));
        EXPECT_TRUE(in_range(0.7f, c.blue_f()));
        EXPECT_TRUE(in_range(0.8f, c.alpha_f()));

        c.float_v(v);
        EXPECT_TRUE(in_range(0.5f, v[0]));
        EXPECT_TRUE(in_range(0.6f, v[1]));
        EXPECT_TRUE(in_range(0.7f, v[2]));
        EXPECT_TRUE(in_range(0.8f, v[3]));
    }
}

TEST(Color, mutator_exceptions)
{
    Color c;

    try
    {
        c.set_red(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (red)",
                  e.to_string());
    }

    try
    {
        c.set_red(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (red)",
                  e.to_string());
    }

    try
    {
        c.set_green(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (green)",
                  e.to_string());
    }

    try
    {
        c.set_green(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (green)",
                  e.to_string());
    }

    try
    {
        c.set_blue(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (blue)",
                  e.to_string());
    }

    try
    {
        c.set_blue(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (blue)",
                  e.to_string());
    }

    try
    {
        c.set_alpha(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (alpha)",
                  e.to_string());
    }

    try
    {
        c.set_alpha(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (alpha)",
                  e.to_string());
    }

    try
    {
        c.set_red_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (red)",
                  e.to_string());
    }

    try
    {
        c.set_red_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (red)",
                  e.to_string());
    }

    try
    {
        c.set_green_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (green)",
                  e.to_string());
    }

    try
    {
        c.set_green_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (green)",
                  e.to_string());
    }

    try
    {
        c.set_blue_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (blue)",
                  e.to_string());
    }

    try
    {
        c.set_blue_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (blue)",
                  e.to_string());
    }

    try
    {
        c.set_alpha_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (alpha)",
                  e.to_string());
    }

    try
    {
        c.set_alpha_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (alpha)",
                  e.to_string());
    }

    try
    {
        Color c;
        float v[4] = { 1.1f, -0.2f, 1.3f, -0.4f };
        c.set_float_v(v);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0.0-1.0] range:"
                  " 1.100000 (red), -0.200000 (green), 1.300000 (blue), -0.400000 (alpha)",
                  e.to_string());
    }

    float old[4] = { 1, 2, 3, 4 };
    try
    {
        Color c;
        float v[4] = { 1.1f, -0.2f, 1.3f, -0.4f };
        float old[4] = { 1, 2, 3, 4 };
        c.set_float_v(v, old);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ(old[0], 1);
        EXPECT_EQ(old[1], 2);
        EXPECT_EQ(old[2], 3);
        EXPECT_EQ(old[3], 4);
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0.0-1.0] range:"
                  " 1.100000 (red), -0.200000 (green), 1.300000 (blue), -0.400000 (alpha)",
                  e.to_string());
    }
}

TEST(Color, less)
{
    EXPECT_FALSE(Color(0, 0, 0, 0) < Color(0, 0, 0, 0));

    EXPECT_TRUE(Color(0, 0, 0, 0) < Color(1, 0, 0, 0));
    EXPECT_TRUE(Color(0, 0, 0, 0) < Color(0, 1, 0, 0));
    EXPECT_TRUE(Color(0, 0, 0, 0) < Color(0, 0, 1, 0));
    EXPECT_TRUE(Color(0, 0, 0, 0) < Color(0, 0, 0, 1));

    EXPECT_FALSE(Color(1, 0, 0, 0) < Color(0, 0, 0, 0));
    EXPECT_FALSE(Color(0, 1, 0, 0) < Color(0, 0, 0, 0));
    EXPECT_FALSE(Color(0, 0, 1, 0) < Color(0, 0, 0, 0));
    EXPECT_FALSE(Color(0, 0, 0, 1) < Color(0, 0, 0, 0));
}

TEST(Color, equals)
{
    EXPECT_TRUE(Color(0, 0, 0, 0) == Color(0, 0, 0, 0));

    EXPECT_FALSE(Color(0, 0, 0, 0) == Color(1, 0, 0, 0));
    EXPECT_FALSE(Color(0, 0, 0, 0) == Color(0, 1, 0, 0));
    EXPECT_FALSE(Color(0, 0, 0, 0) == Color(0, 0, 1, 0));
    EXPECT_FALSE(Color(0, 0, 0, 0) == Color(0, 0, 0, 1));

    EXPECT_FALSE(Color(1, 0, 0, 0) == Color(0, 0, 0, 0));
    EXPECT_FALSE(Color(0, 1, 0, 0) == Color(0, 0, 0, 0));
    EXPECT_FALSE(Color(0, 0, 1, 0) == Color(0, 0, 0, 0));
    EXPECT_FALSE(Color(0, 0, 0, 1) == Color(0, 0, 0, 0));
}

TEST(Color, hash)
{
    unordered_map<Color, string> m;
    m[Color()] = "zero";
    m[Color(1, 0, 0, 0)] = "one";
    EXPECT_EQ(m.find(Color())->second, "zero");
    EXPECT_EQ(m.find(Color(1, 0, 0, 0))->second, "one");
}
