/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/GlintExceptions.h>
#include <glint/internal/GlintExceptionsImpl.h>

#include <GLES2/gl2.h>
#include <gtest/gtest.h>

using namespace std;
using namespace glint;

//
// Check Exception base class functionality (copy, assignment, accessor methods, etc.)
//

TEST(Exception, basic)
{
    OpenGLException e("Hello", 0);
    EXPECT_EQ("Hello (GL_NO_ERROR)", e.reason());
    EXPECT_STREQ("glint::OpenGLException", e.what());

    OpenGLException e2(e);
    EXPECT_EQ(e2.reason(), e.reason());
    EXPECT_STREQ(e2.what(), e.what());

    OpenGLException e3("blah", 0);
    e2 = e3;
    EXPECT_EQ(e2.reason(), e3.reason());
    EXPECT_STREQ(e2.what(), e.what());

    EXPECT_EQ("glint::OpenGLException: blah (GL_NO_ERROR)", e3.to_string());
    EXPECT_EQ("glint::OpenGLException: blah (GL_NO_ERROR)", e3.to_string(0, "    "));
    EXPECT_EQ("        glint::OpenGLException: blah (GL_NO_ERROR)", e3.to_string(1, "        "));
    EXPECT_EQ("                glint::OpenGLException: blah (GL_NO_ERROR)", e3.to_string(2, "        "));
}

TEST(Exception, empty_reason)
{
    {
        InvalidArgumentException e("");
        EXPECT_EQ("", e.reason());
        EXPECT_STREQ("glint::InvalidArgumentException", e.what());
    }

    {
        OpenGLException e("", 0);
        EXPECT_EQ("(GL_NO_ERROR)", e.reason());
        EXPECT_STREQ("glint::OpenGLException", e.what());
    }
}

//
// A few helper functions so we can test that the nesting works correctly.
//

void a()
{
    throw bad_alloc();
}

void b()
{
    try
    {
        a();
    }
    catch (...)
    {
        throw InvalidArgumentException("a() threw");
    }
}

void c()
{
    try
    {
        b();
    }
    catch (const nested_exception&)
    {
        throw InvalidArgumentException("b() threw");
    }
}

class E : public std::exception, public nested_exception
{
public:
    virtual const char* what() const noexcept
    {
        return "E";
    }
    E() = default;
    ~E() noexcept {}
};

void throw_E()
{
    try
    {
        c();
    }
    catch (...)
    {
        throw E();
    }
}

void d()
{
    try
    {
        throw_E();
    }
    catch (...)
    {
        throw InvalidArgumentException("throw_E() threw");
    }
}

struct N : public nested_exception
{
};

void throw_N()
{
    try
    {
        b();
    }
    catch (...)
    {
        throw N();
    }
}

void f()
{
    try
    {
        throw_N();
    }
    catch (...)
    {
        throw InvalidArgumentException("throw_N() threw");
    }
}

void throw_unknown()
{
    throw 42;
}

void g()
{
    try
    {
        throw_unknown();
    }
    catch (...)
    {
        throw InvalidArgumentException("throw_unknown threw");
    }
}

TEST(Exception, nesting)
{
    // Check basic chaining works and terminates correctly on a std::exception.
    //
    try
    {
        c();
    }
    catch (const Exception& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: b() threw:\n"
                  "    glint::InvalidArgumentException: a() threw:\n"
                  "        std::bad_alloc", e.to_string());
    }

    // Check that we follow the chain for exceptions we don't know, but that derive
    // from both std::exception and std::nested_exception.
    //
    try
    {
        d();
    }
    catch (const Exception& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: throw_E() threw:\n"
                  "    E (derived from std::exception and std::nested_exception):\n"
                  "        glint::InvalidArgumentException: b() threw:\n"
                  "            glint::InvalidArgumentException: a() threw:\n"
                  "                std::bad_alloc", e.to_string());
    }

    // Check that we follow the chain for exceptions that are derived from std::nested_exception
    // but not std::exception.
    //
    try
    {
        f();
    }
    catch (const glint::Exception& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: throw_N() threw:\n"
                  "    std::nested_exception:\n"
                  "        glint::InvalidArgumentException: a() threw:\n"
                  "            std::bad_alloc", e.to_string());
    }

    // Check that we are correctly dealing with unknown exceptions
    //
    try
    {
        g();
    }
    catch (const glint::Exception& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: throw_unknown threw:\n"
                  "    unknown exception", e.to_string());
    }
}

//
// Test the history chaining.
//

TEST(Exception, history)
{
    //
    // Check that remember() and get_earlier() return the correct exception.
    //
    {
        InvalidArgumentException e("");
        EXPECT_EQ(nullptr, e.get_earlier());
        exception_ptr ep = make_exception_ptr(e);

        InvalidArgumentException e2("");
        exception_ptr ep2 = e2.remember(ep);
        EXPECT_EQ(e2.get_earlier(), ep);
    }

    {
        InvalidArgumentException e("");
        exception_ptr ep = make_exception_ptr(e);

        InvalidArgumentException e2("");
        exception_ptr ep2 = e2.remember(ep);
        EXPECT_EQ(e2.get_earlier(), ep);
    }

    //
    // Check that we are following the history chain.
    //
    {
        exception_ptr ep;
        try
        {
            throw InvalidArgumentException("Step 1");
        }
        catch (Exception& e)
        {
            ep = e.remember(ep);
        }
        InvalidArgumentException e2("Step 2");
        ep = e2.remember(ep);

        try
        {
            ShutdownException e("Finalization problem");
            e.remember(ep);
            throw e;
        }
        catch (const Exception& e)
        {
            string reason = e.to_string();
            EXPECT_EQ("glint::ShutdownException: Finalization problem\n"
                      "    Exception history:\n"
                      "        Exception #1:\n"
                      "            glint::InvalidArgumentException: Step 1\n"
                      "        Exception #2:\n"
                      "            glint::InvalidArgumentException: Step 2", e.to_string());
        }
    }

    //
    // Same test, but this time with nested exceptions in the history.
    //
    {
        exception_ptr ep;
        try
        {
            c();
        }
        catch (Exception& e)
        {
            ep = e.remember(ep);
        }
        try
        {
            f();
        }
        catch (Exception &e)
        {
            ep = e.remember(ep);
        }

        try
        {
            ShutdownException e("Finalization problem");
            e.remember(ep);
            throw e;
        }
        catch (const Exception& e)
        {
            string reason = e.to_string();
            EXPECT_EQ("glint::ShutdownException: Finalization problem\n"
                      "    Exception history:\n"
                      "        Exception #1:\n"
                      "            glint::InvalidArgumentException: b() threw:\n"
                      "                glint::InvalidArgumentException: a() threw:\n"
                      "                    std::bad_alloc\n"
                      "        Exception #2:\n"
                      "            glint::InvalidArgumentException: throw_N() threw:\n"
                      "                std::nested_exception:\n"
                      "                    glint::InvalidArgumentException: a() threw:\n"
                      "                        std::bad_alloc", e.to_string());
        }
    }

    //
    // Same test, but this time with history in a nested exception.
    //
    {
        exception_ptr ep;
        try
        {
            c();
        }
        catch (Exception& e)
        {
            ep = e.remember(ep);
        }
        try
        {
            f();
        }
        catch (Exception &e)
        {
            ep = e.remember(ep);
        }

        try
        {
            try
            {
                ShutdownException e("Finalization problem");
                e.remember(ep);
                throw e;
            }
            catch (Exception &e)
            {
                throw ShutdownException("Cannot finalize");
            }
        }
        catch (const Exception& e)
        {
            string reason = e.to_string();
            cout << reason << endl;
            EXPECT_EQ("glint::ShutdownException: Finalization problem\n"
                      "    Exception history:\n"
                      "        Exception #1:\n"
                      "            glint::InvalidArgumentException: b() threw:\n"
                      "                glint::InvalidArgumentException: a() threw:\n"
                      "                    std::bad_alloc\n"
                      "        Exception #2:\n"
                      "            glint::InvalidArgumentException: throw_N() threw:\n"
                      "                std::nested_exception:\n"
                      "                    glint::InvalidArgumentException: a() threw:\n"
                      "                        std::bad_alloc", e.to_string());
        }
    }
}

//
// Tests for the state of concrete derived exceptions follow.
//

TEST(OpenGLException, state)
{
    // Check that we correctly mention the error code.
    //
    {
        OpenGLException e("without error", GL_NO_ERROR);
        EXPECT_EQ("glint::OpenGLException: without error (GL_NO_ERROR)", e.to_string());
        OpenGLException e2("blah", GL_NO_ERROR);
        e2 = e;
        EXPECT_EQ(e.reason(), e2.reason());
    }

    {
        OpenGLException e("with error code", 42);
        EXPECT_EQ("glint::OpenGLException: with error code (Unknown GL error: 42 (0x2a))", e.to_string());
        EXPECT_EQ(e.gl_error(), 42);
        EXPECT_THROW(rethrow_exception(e.self()), OpenGLException);
    }

    {
        OpenGLException e("with enum", GL_INVALID_ENUM);
        EXPECT_EQ("glint::OpenGLException: with enum (GL_INVALID_ENUM)", e.to_string());
        EXPECT_EQ(e.gl_error(), GL_INVALID_ENUM);
        EXPECT_THROW(rethrow_exception(e.self()), OpenGLException);
    }
}

TEST(InvalidArgumentException, state)
{
    {
        InvalidArgumentException e("bad arg");
        EXPECT_STREQ("glint::InvalidArgumentException", e.what());
        EXPECT_THROW(rethrow_exception(e.self()), InvalidArgumentException);
        InvalidArgumentException e2("blah");
        e2 = e;
        EXPECT_EQ(e.reason(), e2.reason());
    }
}

TEST(LogicException, state)
{
    {
        LogicException e("You shouldn't have done that!");
        EXPECT_STREQ("glint::LogicException", e.what());
        EXPECT_THROW(rethrow_exception(e.self()), LogicException);
        LogicException e2("blah");
        e2 = e;
        EXPECT_EQ(e.reason(), e2.reason());
    }
}

TEST(ShutdownException, state)
{
    {
        ShutdownException e("Need some kicks");
        EXPECT_STREQ("glint::ShutdownException", e.what());
        EXPECT_THROW(rethrow_exception(e.self()), ShutdownException);
        ShutdownException e2("blah");
        e2 = e;
        EXPECT_EQ(e.reason(), e2.reason());
    }
}

TEST(FileException, state)
{
    {
        FileException e("File error", 0);
        EXPECT_EQ("File error", e.reason());
        EXPECT_STREQ("glint::FileException", e.what());
        EXPECT_EQ(0, e.error());
        EXPECT_THROW(rethrow_exception(e.self()), FileException);
        FileException e2("blah", 0);
        e2 = e;
        EXPECT_EQ(e.reason(), e2.reason());
        EXPECT_EQ(e.error(), e2.error());
    }

    {
        FileException e("File error", 42);
        EXPECT_EQ("File error (errno = 42)", e.reason());
        EXPECT_STREQ("glint::FileException", e.what());
        EXPECT_EQ(42, e.error());
        EXPECT_THROW(rethrow_exception(e.self()), FileException);
        FileException e2("blah", 0);
        e2 = e;
        EXPECT_EQ(e.reason(), e2.reason());
        EXPECT_EQ(e.error(), e2.error());
    }
}

// Dynamic allocataion to get around bogus function coverage reports by gcov.

TEST(Exceptions, dynamic)
{
    {
        OpenGLException* ep = new OpenGLException("Hello", 0);
        delete ep;
        internal::OpenGLExceptionImpl* eip = new internal::OpenGLExceptionImpl("Hello", 0);
        delete eip;
    }

    {
        InvalidArgumentException* ep = new InvalidArgumentException("Hello");
        delete ep;
        internal::InvalidArgumentExceptionImpl* eip = new internal::InvalidArgumentExceptionImpl("Hello");
        delete eip;
    }

    {
        LogicException* ep = new LogicException("Hello");
        delete ep;
        internal::LogicExceptionImpl* eip = new internal::LogicExceptionImpl("Hello");
        delete eip;
    }

    {
        ShutdownException* ep = new ShutdownException("Hello");
        delete ep;
        internal::ShutdownExceptionImpl* eip = new internal::ShutdownExceptionImpl("Hello");
        delete eip;
    }

    {
        FileException* ep = new FileException("Hello", 0);
        delete ep;
        internal::FileExceptionImpl* eip = new internal::FileExceptionImpl("Hello", 0);
        delete eip;
    }
}
