/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/HSVColor.h>
#include <glint/GlintExceptions.h>
#include <gtest/gtest.h>

#include <sstream>
#include <vector>
#include <unordered_map>

using namespace glint;
using namespace std;

//
// When we set a color component from a float, the Color implementation
// rounds to the closest int. We use this function to test that, after setting
// a value from a float, the float that is returned is within 1/2 of 1/255.0f,
// that is, was mapped to the correct integer value.
//

bool
in_range(float expected_val, float set_val)
{
    const float half_range = 1 / 255.0f / 2;
    float low = set_val - half_range;
    float high = set_val + half_range;
    return expected_val >= low && expected_val < high;
}

TEST(HSVColor, constructors)
{
    {
        HSVColor c;

        EXPECT_EQ(0, c.hue());
        EXPECT_EQ(0, c.saturation());
        EXPECT_EQ(0, c.value());
        EXPECT_EQ(255, c.alpha());

        EXPECT_EQ(0.0f, c.hue_f());
        EXPECT_EQ(0.0f, c.saturation_f());
        EXPECT_EQ(0.0f, c.value_f());
        EXPECT_EQ(1.0f, c.alpha_f());
    }

    // Check that hue of 1.0 changes to 0.0.

    {
        HSVColor c(1.0f, 0.1f, 0.2f, 0.3f);
        EXPECT_EQ(0.0f, c.hue_f());
        EXPECT_TRUE(in_range(0.1f, c.saturation_f()));
        EXPECT_TRUE(in_range(0.2f, c.value_f()));
        EXPECT_TRUE(in_range(0.3f, c.alpha_f()));
    }

    for (int h = 0; h < 360; ++h)
    {
        for (int i = 0; i < 256; ++i)
        {
            HSVColor c(h, i, i, i);
            EXPECT_EQ(h, c.hue());
            EXPECT_EQ(h / 360.0f, c.hue_f());
            EXPECT_EQ(i, c.saturation());
            EXPECT_EQ(i / 255.0f, c.saturation_f());
            EXPECT_EQ(i, c.value());
            EXPECT_EQ(i / 255.0f, c.value_f());
            EXPECT_EQ(i, c.alpha());
            EXPECT_EQ(i / 255.0f, c.alpha_f());
        }
    }

    for (int h = 0; h < 360; ++h)
    {
        for (int i = 0; i < 256; ++i)
        {
            float f = i / 255.0f;
            HSVColor c(h / 360.0f, f, f, f);
            EXPECT_EQ(h, c.hue());
            EXPECT_EQ(h / 360.0f, c.hue_f());
            EXPECT_EQ(i, c.saturation());
            EXPECT_EQ(i / 255.0f, c.saturation_f());
            EXPECT_EQ(i, c.value());
            EXPECT_EQ(i / 255.0f, c.value_f());
            EXPECT_EQ(i, c.alpha());
            EXPECT_EQ(i / 255.0f, c.alpha_f());
        }
    }

    {
        HSVColor c(1, 2, 3);
        EXPECT_EQ(1, c.hue());
        EXPECT_EQ(2, c.saturation());
        EXPECT_EQ(3, c.value());
        EXPECT_EQ(255, c.alpha());
    }

    {
        HSVColor c(0.1f, 0.2f, 0.3f);
        EXPECT_TRUE(in_range(0.1f, c.hue_f()));
        EXPECT_TRUE(in_range(0.2f, c.saturation_f()));
        EXPECT_TRUE(in_range(0.3f, c.value_f()));
        EXPECT_TRUE(in_range(1.0f, c.alpha_f()));
    }
}

TEST(HSVColor, constructor_exceptions)
{
    try
    {
        HSVColor c(-1, 0, 0 , 0);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid hue outside [0-359] range: -1",
                  e.to_string());
    }

    try
    {
        HSVColor c(1, -1, 256 , 300);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0-255] range:"
                  " -1 (saturation), 256 (value), 300 (alpha)",
                  e.to_string());
    }

    try
    {
        HSVColor c(360, -1, 256 , 300);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid hue outside [0-359] range: 360;"
                  " invalid color arguments outside [0-255] range: -1 (saturation), 256 (value), 300 (alpha)",
                  e.to_string());
    }

    try
    {
        HSVColor c(-1.0f, 0.0f, 0.0f, 0.0f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid hue outside [0.0-1.0] range: -1",
                  e.to_string());
    }

    try
    {
        HSVColor c(1.0f, 0.0f, -1.01f, 1.01f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color arguments outside [0.0-1.0] range:"
                  " -1.010000 (value), 1.010000 (alpha)",
                  e.to_string());
    }

    try
    {
        HSVColor c(1.0f, -0.1f, 0.0f, 0.0f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range:"
                  " -0.100000 (saturation)",
                  e.to_string());
    }
}

TEST(HSVColor, mutators)
{
    {
        HSVColor c(1, 2, 3, 4);

        EXPECT_EQ(1, c.set_hue(11));
        EXPECT_EQ(11, c.hue());
        EXPECT_EQ(2, c.set_saturation(12));
        EXPECT_EQ(12, c.saturation());
        EXPECT_EQ(3, c.set_value(13));
        EXPECT_EQ(13, c.value());
        EXPECT_EQ(4, c.set_alpha(14));
        EXPECT_EQ(14, c.alpha());

        EXPECT_EQ(11, c.hue());
        EXPECT_EQ(12, c.saturation());
        EXPECT_EQ(13, c.value());
        EXPECT_EQ(14, c.alpha());
    }

    {
        HSVColor c(0.1f, 0.2f, 0.3f, 0.4f);

        EXPECT_TRUE(in_range(0.1f, c.set_hue_f(0.11f)));
        EXPECT_TRUE(in_range(0.11f, c.hue_f()));
        EXPECT_TRUE(in_range(0.2f, c.set_saturation_f(0.12f)));
        EXPECT_TRUE(in_range(0.12f, c.saturation_f()));
        EXPECT_TRUE(in_range(0.3f, c.set_value_f(0.13f)));
        EXPECT_TRUE(in_range(0.13f, c.value_f()));
        EXPECT_TRUE(in_range(0.4f, c.set_alpha_f(0.14f)));
        EXPECT_TRUE(in_range(0.14f, c.alpha_f()));

        EXPECT_TRUE(in_range(0.11f, c.hue_f()));
        EXPECT_TRUE(in_range(0.12f, c.saturation_f()));
        EXPECT_TRUE(in_range(0.13f, c.value_f()));
        EXPECT_TRUE(in_range(0.14f, c.alpha_f()));

        EXPECT_EQ(40, c.hue());
        EXPECT_EQ(31, c.saturation());
        EXPECT_EQ(33, c.value());
        EXPECT_EQ(36, c.alpha());
    }

    // Changing one compoment must not change any other component. This is important for V = 0, which leaves
    // H and S undefined. In that case, H and S don't change their previous value.
    // Similary for S = 0, which leaves leaves H undefined. In that case H does not change it's previous value.

    {
        HSVColor c(40, 40, 40, 40);

        EXPECT_EQ(40, c.hue());
        EXPECT_EQ(40, c.saturation());
        EXPECT_EQ(40, c.value());
        EXPECT_EQ(40, c.alpha());

        EXPECT_EQ(40, c.set_saturation(0));

        EXPECT_EQ(40, c.hue());
        EXPECT_EQ(0, c.saturation());
        EXPECT_EQ(40, c.value());
        EXPECT_EQ(40, c.alpha());

        EXPECT_EQ(0, c.set_saturation(50));

        EXPECT_EQ(40, c.hue());
        EXPECT_EQ(50, c.saturation());
        EXPECT_EQ(40, c.value());
        EXPECT_EQ(40, c.alpha());

        EXPECT_EQ(40, c.set_value(0));

        EXPECT_EQ(40, c.hue());
        EXPECT_EQ(50, c.saturation());
        EXPECT_EQ(0, c.value());
        EXPECT_EQ(40, c.alpha());

        EXPECT_EQ(0, c.set_value(60));

        EXPECT_EQ(40, c.hue());
        EXPECT_EQ(50, c.saturation());
        EXPECT_EQ(60, c.value());
        EXPECT_EQ(40, c.alpha());
    }
}

TEST(HSVColor, mutator_exceptions)
{
    HSVColor c;

    try
    {
        c.set_hue(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid hue argument outside [0-359] range: -1",
                  e.to_string());
    }

    try
    {
        c.set_hue(360);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid hue argument outside [0-359] range: 360",
                  e.to_string());
    }

    try
    {
        c.set_saturation(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (saturation)",
                  e.to_string());
    }

    try
    {
        c.set_saturation(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (saturation)",
                  e.to_string());
    }

    try
    {
        c.set_value(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (value)",
                  e.to_string());
    }

    try
    {
        c.set_value(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (value)",
                  e.to_string());
    }

    try
    {
        c.set_alpha(-1);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: -1 (alpha)",
                  e.to_string());
    }

    try
    {
        c.set_alpha(256);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0-255] range: 256 (alpha)",
                  e.to_string());
    }

    try
    {
        c.set_hue_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (hue)",
                  e.to_string());
    }

    try
    {
        c.set_hue_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (hue)",
                  e.to_string());
    }

    try
    {
        c.set_saturation_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (saturation)",
                  e.to_string());
    }

    try
    {
        c.set_saturation_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (saturation)",
                  e.to_string());
    }

    try
    {
        c.set_value_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (value)",
                  e.to_string());
    }

    try
    {
        c.set_value_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (value)",
                  e.to_string());
    }

    try
    {
        c.set_alpha_f(-0.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: -0.1 (alpha)",
                  e.to_string());
    }

    try
    {
        c.set_alpha_f(1.1f);
        FAIL();
    }
    catch (const InvalidArgumentException& e)
    {
        EXPECT_EQ("glint::InvalidArgumentException: Invalid color argument outside [0.0-1.0] range: 1.1 (alpha)",
                  e.to_string());
    }
}

TEST(HSVColor, less)
{
    EXPECT_FALSE(HSVColor(0, 0, 0, 0) < HSVColor(0, 0, 0, 0));

    EXPECT_TRUE(HSVColor(0, 0, 0, 0) < HSVColor(1, 0, 0, 0));
    EXPECT_TRUE(HSVColor(0, 0, 0, 0) < HSVColor(0, 1, 0, 0));
    EXPECT_TRUE(HSVColor(0, 0, 0, 0) < HSVColor(0, 0, 1, 0));
    EXPECT_TRUE(HSVColor(0, 0, 0, 0) < HSVColor(0, 0, 0, 1));

    EXPECT_FALSE(HSVColor(1, 0, 0, 0) < HSVColor(0, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 1, 0, 0) < HSVColor(0, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 0, 1, 0) < HSVColor(0, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 0, 0, 1) < HSVColor(0, 0, 0, 0));
}


TEST(HSVColor, equals)
{
    EXPECT_TRUE(HSVColor(0, 0, 0, 0) == HSVColor(0, 0, 0, 0));

    EXPECT_FALSE(HSVColor(0, 0, 0, 0) == HSVColor(1, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 0, 0, 0) == HSVColor(0, 1, 0, 0));
    EXPECT_FALSE(HSVColor(0, 0, 0, 0) == HSVColor(0, 0, 1, 0));
    EXPECT_FALSE(HSVColor(0, 0, 0, 0) == HSVColor(0, 0, 0, 1));

    EXPECT_FALSE(HSVColor(1, 0, 0, 0) == HSVColor(0, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 1, 0, 0) == HSVColor(0, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 0, 1, 0) == HSVColor(0, 0, 0, 0));
    EXPECT_FALSE(HSVColor(0, 0, 0, 1) == HSVColor(0, 0, 0, 0));
}

TEST(HSVColor, hash)
{
    unordered_map<HSVColor, string> m;
    m[HSVColor()] = "zero";
    m[HSVColor(1, 0, 0, 0)] = "one";
    EXPECT_EQ(m.find(HSVColor())->second, "zero");
    EXPECT_EQ(m.find(HSVColor(1, 0, 0, 0))->second, "one");
}

TEST(HSVColor, RGB_TO_HSV)
{
    // Black sets hue and saturation zero.
    {
        Color c;
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(0, hsv.hue());
        EXPECT_EQ(0, hsv.saturation());
        EXPECT_EQ(0, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // White sets hue to zero.
    {
        Color c(255, 255, 255);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(0, hsv.hue());
        EXPECT_EQ(0, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Grey sets hue and saturation to zero.
    {
        Color c(127, 127, 127);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(0, hsv.hue());
        EXPECT_EQ(0, hsv.saturation());
        EXPECT_EQ(127, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Just before red
    {
        Color c(255, 0, 4);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(359, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Red
    {
        Color c(255, 0, 0);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(0, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Just before green
    {
        Color c(4, 255, 0);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(119, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Green
    {
        Color c(0, 255, 0);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(120, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Just before blue
    {
        Color c(0, 4, 255);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(239, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Blue
    {
        Color c(0, 0, 255);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(240, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Just before yellow
    {
        Color c(255, 251, 0);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(59, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Yellow
    {
        Color c(255, 255, 0);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(60, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Just before cyan
    {
        Color c(0, 255, 251);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(179, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Cyan
    {
        Color c(0, 255, 255);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(180, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Just before magenta
    {
        Color c(251, 0, 255);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(299, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }

    // Magenta
    {
        Color c(255, 0, 255);
        HSVColor hsv = RGB_to_HSV(c);
        EXPECT_EQ(300, hsv.hue());
        EXPECT_EQ(255, hsv.saturation());
        EXPECT_EQ(255, hsv.value());
        EXPECT_EQ(255, hsv.alpha());
    }
}

TEST(HSVColor, HSV_TO_RGB)
{
    // Black
    {
        HSVColor c;
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(0, rgb.red());
        EXPECT_EQ(0, rgb.green());
        EXPECT_EQ(0, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    // White
    {
        HSVColor c(60, 0, 255);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(255, rgb.red());
        EXPECT_EQ(255, rgb.green());
        EXPECT_EQ(255, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    // Grey
    {
        HSVColor c(90, 0, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(127, rgb.red());
        EXPECT_EQ(127, rgb.green());
        EXPECT_EQ(127, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    // Test color in all six sectors either side of boundary.
    {
        HSVColor c(0, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(127, rgb.red());
        EXPECT_EQ(64, rgb.green());
        EXPECT_EQ(64, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(59, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(127, rgb.red());
        EXPECT_EQ(126, rgb.green());
        EXPECT_EQ(64, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(60, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(127, rgb.red());
        EXPECT_EQ(127, rgb.green());
        EXPECT_EQ(64, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(119, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(65, rgb.red());
        EXPECT_EQ(127, rgb.green());
        EXPECT_EQ(64, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(120, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(64, rgb.red());
        EXPECT_EQ(127, rgb.green());
        EXPECT_EQ(64, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(179, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(64, rgb.red());
        EXPECT_EQ(127, rgb.green());
        EXPECT_EQ(126, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(180, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(64, rgb.red());
        EXPECT_EQ(127, rgb.green());
        EXPECT_EQ(127, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(239, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(64, rgb.red());
        EXPECT_EQ(65, rgb.green());
        EXPECT_EQ(127, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(240, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(64, rgb.red());
        EXPECT_EQ(64, rgb.green());
        EXPECT_EQ(127, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(299, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(126, rgb.red());
        EXPECT_EQ(64, rgb.green());
        EXPECT_EQ(127, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(300, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(127, rgb.red());
        EXPECT_EQ(64, rgb.green());
        EXPECT_EQ(127, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }

    {
        HSVColor c(359, 127, 127);
        Color rgb = HSV_to_RGB(c);
        EXPECT_EQ(127, rgb.red());
        EXPECT_EQ(64, rgb.green());
        EXPECT_EQ(65, rgb.blue());
        EXPECT_EQ(255, rgb.alpha());
    }
}
