/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Program.h>
#include <glint/Shader.h>
#include <glint/Runtime.h>
#include <glint/GlintExceptions.h>

#include <gtest/gtest.h>
#include <boost/regex.hpp>  // BUGFIX: We don't use std::regex because, as of gcc 4.7.2, it is still broken

#include <array>

using namespace std;
using namespace glint;

TEST(Program, basic)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "void main()"
                           "{"
                           "    gl_Position = vec4(0);"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    auto v = VertexShader::create(v_source);
    auto f = FragmentShader::create(f_source);
    auto p = Program::create(*v, *f);

    EXPECT_NE(0, p->id());
    EXPECT_EQ(v->id(), p->vertex_shader_id());
    EXPECT_EQ(f->id(), p->fragment_shader_id());

    EXPECT_FALSE(p->in_use());
    p->use();
    EXPECT_TRUE(p->in_use());

    EXPECT_TRUE(p->uniforms().empty());
}

TEST(Program, bad_vertex_shader)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "void main()"
                           "{"
                           "/* Invalid, no output */"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    // We deliberately enclose the entire sequence in the try block because
    // at what stage (compile, link, or use) things fail depends on the implementation.
    try
    {
        auto v = VertexShader::create(v_source);
        auto f = FragmentShader::create(f_source);
        auto p = Program::create(*v, *f);
        p->use();
        FAIL();
    }
    catch (const OpenGLException& e)
    {
        // No string comparison here for the error message because the message is implementation-dependent
    }
}

TEST(Program, bad_link)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "float foo() { return 1.0; }";

    const char* f_source = "float foo() { return 2.0; }";

    // We deliberately enclose the entire sequence in the try block because
    // at what stage (compile, link, or use) things fail depends on the implementation.
    try
    {
        auto v = VertexShader::create(v_source);
        auto f = FragmentShader::create(f_source);
        auto p = Program::create(*v, *f);
        p->use();
        FAIL();
    }
    catch (const Exception& e)
    {
        // No string comparison here for the error message because the message is implementation-dependent
    }
}

TEST(Program, uniform_lookup)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "uniform float my_float[2];"
                           "void main()"
                           "{"
                           "    gl_Position = vec4(my_float[0] + my_float[1]);"
                           "}";

    const char* f_source = "void main()"
                           "{"
                           "    gl_FragColor = vec4(0);"
                           "}";

    auto v = VertexShader::create(v_source);
    auto f = FragmentShader::create(f_source);
    auto p = Program::create(*v, *f);

    // Setting a variable before the program is installed should fail, but reading it is OK.
    try
    {
        UniformVariable& uv = p->uniform("my_float");
        FloatArrayUniform& iu = dynamic_cast<FloatArrayUniform&>(uv);
        iu.set({ { 1.0f, 2.0f }});
        FAIL();
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: cannot set uniform variable \"my_float\" at location [0-9]+;"
                       " size = 2, program id = [0-9]+: program is not currently in use",
                       boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }

    {
        UniformVariable& uv = p->uniform("my_float");
        FloatArrayUniform& iu = dynamic_cast<FloatArrayUniform&>(uv);
        iu.get();
    }

    p->use();

    {
        UniformVariable& uv = p->uniform("my_float");
        FloatArrayUniform& iu = dynamic_cast<FloatArrayUniform&>(uv);
        EXPECT_EQ(p->id(), iu.program_id());
        EXPECT_NE(-1, iu.location());
        EXPECT_EQ("my_float", iu.name());
        EXPECT_EQ(GL_FLOAT, iu.type());
        EXPECT_EQ(2, iu.size());
    }

    // Repeat for the same variable, but this time with an array name.
    {
        UniformVariable& uv = p->uniform("my_float[0]");
        FloatUniform& iu = dynamic_cast<FloatUniform&>(uv);
        EXPECT_EQ(p->id(), iu.program_id());
        EXPECT_NE(-1, iu.location());
        EXPECT_EQ("my_float[0]", iu.name());
        EXPECT_EQ(GL_FLOAT, iu.type());
        EXPECT_EQ(1, iu.size());
    }

    // Once more for coverage testing. This time, the variable should come out of the cache.
    {
        UniformVariable& uv = p->uniform("my_float[0]");
        FloatUniform& iu = dynamic_cast<FloatUniform&>(uv);
        EXPECT_EQ(p->id(), iu.program_id());
        EXPECT_NE(-1, iu.location());
        EXPECT_EQ("my_float[0]", iu.name());
        EXPECT_EQ(GL_FLOAT, iu.type());
        EXPECT_EQ(1, iu.size());
    }

    try
    {
        p->uniform("Michi");
        FAIL();
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: no uniform variable \"Michi\" found for program id [0-9]+",
                       boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }
}

// Check that complaints about vector size are correct in terms of singular and plural.

TEST(Program, size_exception)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    const char* v_source = "uniform float fa2[2];"
                           "void main()"
                           "{"
                           "    gl_Position = vec4(fa2[1]);"
                           "}";

    const char* f_source = "uniform sampler2D sa[2];"
                           "uniform samplerCube ca[2];"
                           "void main()"
                           "{"
                           "    gl_FragColor = texture2D(sa[1], vec2(0));"
                           "    gl_FragColor = gl_FragColor + textureCube(ca[1], vec3(0));"
                           "}";

    Program::UPtr p;
    try
    {
        auto v = VertexShader::create(v_source);
        auto f = FragmentShader::create(f_source);
        p = Program::create(*v, *f);
        p->use();
    }
    catch (const Exception& e)
    {
        FAIL() << e.to_string();
    }

    UniformVariable& uv = p->uniform("fa2");
    FloatArrayUniform& u = dynamic_cast<FloatArrayUniform&>(uv);
    FloatArray a;

    try
    {
        u.set(a);
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: cannot set array uniform variable \"[a-zA-Z0-9_]+\" at location"
                       " [0-9]+ for program id [0-9]+: expected size is 2 elements"
                       " but vector contains 0 elements",
               boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }

    a.resize(1);
    try
    {
        u.set(a);
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: cannot set array uniform variable \"[a-zA-Z0-9_]+\" at location"
                       " [0-9]+ for program id [0-9]+: expected size is 2 elements"
                       " but vector contains 1 element",
               boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }

    a.resize(3);
    try
    {
        u.set(a);
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: cannot set array uniform variable \"[a-zA-Z0-9_]+\" at location"
                       " [0-9]+ for program id [0-9]+: expected size is 2 elements"
                       " but vector contains 3 elements",
               boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }
}

//
// size_test() checks that setting from a vector with the wrong number of elements throws.
// We call this for each possible type, to make sure that the implementation of the
// set() method in each class actually enforces the correct number of elements.
//

template<typename T, typename U>
void
size_test(U& u)
{
    vector<T> v;
    v.resize(3);
    try
    {
        u.set(v);
        FAIL();
    }
    catch (const LogicException& e)
    {
        boost::regex r("glint::LogicException: cannot set array uniform variable \"[a-zA-Z0-9_]+\" at location"
                       " [0-9]+ for program id [0-9]+: expected size is 2 elements"
                       " but vector contains 3 elements",
               boost::regex::egrep);
        EXPECT_TRUE(boost::regex_match(e.to_string(), r));
    }
}

template<typename T, typename U>
void
assign_test(U& u)
{
    // Type-specific test
    T val;
    val = u;
    EXPECT_EQ(u.get(), val);
    u = val;
    u = u;
    EXPECT_EQ(u.get(), val);

    // Assignment to base.

    UniformVariable& v(u);
    v = val;
    EXPECT_EQ(u.get(), val);
    v.set(val);
    EXPECT_EQ(u.get(), val);
}

TEST(Program, uniforms)
{
    auto rt = Runtime::create();
    rt->create_display(200, 200);

    // Shader with all possible types of uniform. Silly use of these in the body of the shader, so they
    // are not optimized out of existence.
    const char* v_source = "uniform float f;"
                           "uniform vec2 fv2;"
                           "uniform vec3 fv3;"
                           "uniform vec4 fv4;"
                           "uniform mat2 fm2;"
                           "uniform mat3 fm3;"
                           "uniform mat4 fm4;"
                           "uniform float fa[2];"
                           "uniform vec2 fv2a[2];"
                           "uniform vec3 fv3a[2];"
                           "uniform vec4 fv4a[2];"
                           "uniform mat2 fm2a[2];"
                           "uniform mat3 fm3a[2];"
                           "uniform mat4 fm4a[2];"
                           "uniform int i;"
                           "uniform ivec2 iv2;"
                           "uniform ivec3 iv3;"
                           "uniform ivec4 iv4;"
                           "uniform int ia[2];"
                           "uniform ivec2 iv2a[2];"
                           "uniform ivec3 iv3a[2];"
                           "uniform ivec4 iv4a[2];"
                           "uniform bool b;"
                           "uniform bvec2 bv2;"
                           "uniform bvec3 bv3;"
                           "uniform bvec4 bv4;"
                           "uniform bool ba[2];"
                           "uniform bvec2 bv2a[2];"
                           "uniform bvec3 bv3a[2];"
                           "uniform bvec4 bv4a[2];"
                           "void main()"
                           "{"
                           "    gl_Position = vec4(f + fv2.x + fv3.x + fv4.x);"
                           "    gl_Position = gl_Position + vec4(fm2[0][0] + fm3[0][0] + fm4[0][0]);"
                           "    gl_Position = gl_Position + vec4(fa[1] + fv2a[1].x + fv3a[1].x + fv4a[1].x);"
                           "    gl_Position = gl_Position + vec4(fm2a[1][0][0] + fm3a[1][0][0] + fm4a[1][0][0]);"
                           "    gl_Position = gl_Position + vec4(i + iv2.x + iv3.x + iv4.x);"
                           "    gl_Position = gl_Position + vec4(ia[1] + iv2a[1].x + iv3a[1].x + iv4a[1].x);"
                           "    gl_Position = gl_Position + vec4(int(b) + int(bv2.x) + int(bv3.x) + int(bv4.x));"
                           "    gl_Position = gl_Position + vec4(int(ba[1]) + int(bv2a[1]) + int(bv3a[1]) + int(bv4a[1]));"
                           "}";

    const char* f_source = "uniform sampler2D s;"
                           "uniform sampler2D sa[2];"
                           "uniform samplerCube c;"
                           "uniform samplerCube ca[2];"
                           "void main()"
                           "{"
                           "    gl_FragColor = texture2D(s, vec2(0));"
                           "    gl_FragColor = gl_FragColor + texture2D(sa[1], vec2(0));"
                           "    gl_FragColor = gl_FragColor + textureCube(c, vec3(0));"
                           "    gl_FragColor = gl_FragColor + textureCube(ca[1], vec3(0));"
                           "}";

    Program::UPtr p;
    try
    {
        auto v = VertexShader::create(v_source);
        auto f = FragmentShader::create(f_source);
        p = Program::create(*v, *f);
        p->use();
    }
    catch (const Exception& e)
    {
        FAIL() << e.to_string();
    }

    try
    {
        // Float types

        {
            UniformVariable& uv = p->uniform("f");
            FloatUniform& u = dynamic_cast<FloatUniform&>(uv);
            u = 0.5f;
            EXPECT_EQ(0.5f, u.get());

            assign_test<GLfloat>(u);
        }

        {
            UniformVariable& uv = p->uniform("fv2");
            FloatVec2Uniform& u = dynamic_cast<FloatVec2Uniform&>(uv);
            u = { { 0.1f, 0.2f } };
            EXPECT_EQ(0.1f, u.get()[0]);
            EXPECT_EQ(0.2f, u.get()[1]);

            u.set(0.0f, 0.1f);              // Separate parameter interface
            float x, y;
            u.get(x, y);
            EXPECT_EQ(0.0f, x);
            EXPECT_EQ(0.1f, y);

            assign_test<FloatVec2>(u);
        }

        {
            UniformVariable& uv = p->uniform("fv3");
            FloatVec3Uniform& u = dynamic_cast<FloatVec3Uniform&>(uv);
            u.set({ { 0.1f, 0.2f, 0.3f } });
            EXPECT_EQ(0.1f, u.get()[0]);
            EXPECT_EQ(0.2f, u.get()[1]);
            EXPECT_EQ(0.3f, u.get()[2]);

            u.set(0.0f, 0.1f, 0.2f);        // Separate parameter interface
            float x, y, z;
            u.get(x, y, z);
            EXPECT_EQ(0.0f, x);
            EXPECT_EQ(0.1f, y);
            EXPECT_EQ(0.2f, z);

            FloatVec3 val = u;
            EXPECT_EQ(u.get(), val);
            u = val;
            u = u;
            EXPECT_EQ(u.get(), val);

            assign_test<FloatVec3>(u);
        }

        {
            UniformVariable& uv = p->uniform("fv4");
            FloatVec4Uniform& u = dynamic_cast<FloatVec4Uniform&>(uv);
            u.set({ { 0.1f, 0.2f, 0.3f, 0.4f } });
            EXPECT_EQ(0.1f, u.get()[0]);
            EXPECT_EQ(0.2f, u.get()[1]);
            EXPECT_EQ(0.3f, u.get()[2]);
            EXPECT_EQ(0.4f, u.get()[3]);

            u.set(0.0f, 0.1f, 0.2f, 0.3f);   // Separate parameter interface
            float x, y, z, w;
            u.get(x, y, z, w);
            EXPECT_EQ(0.0f, x);
            EXPECT_EQ(0.1f, y);
            EXPECT_EQ(0.2f, z);
            EXPECT_EQ(0.3f, w);

            assign_test<FloatVec4>(u);
        }

        {
            UniformVariable& uv = p->uniform("fm2");
            FloatMat2Uniform& u = dynamic_cast<FloatMat2Uniform&>(uv);
            FloatMat2 m = { { { { 1.0f, 2.0f } },
                              { { 3.0f, 4.0f } } } };
            u.set(m);
            EXPECT_EQ(m, u.get());

            assign_test<FloatMat2>(u);
        }

        {
            UniformVariable& uv = p->uniform("fm3");
            FloatMat3Uniform& u = dynamic_cast<FloatMat3Uniform&>(uv);
            FloatMat3 m = { { { { 1.0f, 2.0f, 3.0f } },
                              { { 4.0f, 5.0f, 6.0f } },
                              { { 7.0f, 8.0f, 9.0f } } } };
            u.set(m);
            EXPECT_EQ(m, u.get());

            assign_test<FloatMat3>(u);
        }

        {
            UniformVariable& uv = p->uniform("fm4");
            FloatMat4Uniform& u = dynamic_cast<FloatMat4Uniform&>(uv);
            FloatMat4 m = { { { {  1.0f,  2.0f,  3.0f,  4.0f } },
                              { {  5.0f,  6.0f,  7.0f,  8.0f } },
                              { {  9.0f, 10.0f, 11.0f, 12.0f } },
                              { { 13.0f, 14.0f, 15.0f, 16.0f } } } };
            u.set(m);
            EXPECT_EQ(m, u.get());

            assign_test<FloatMat4>(u);
        }

        // Float array types

        {
            UniformVariable& uv = p->uniform("fa");
            FloatArrayUniform& u = dynamic_cast<FloatArrayUniform&>(uv);
            u.set({ { 0.1f, 0.2f } });
            EXPECT_EQ(0.1f, u.get()[0]);
            EXPECT_EQ(0.2f, u.get()[1]);

            assign_test<FloatArray>(u);
            size_test<GLfloat>(u);
        }

        {
            UniformVariable& uv = p->uniform("fv2a");
            FloatVec2ArrayUniform& u = dynamic_cast<FloatVec2ArrayUniform&>(uv);
            FloatVec2Array a { { { { 1.0f, 2.0f } },
                                 { { 3.0f, 4.0f } } } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<FloatVec2Array>(u);
            size_test<FloatVec2>(u);
        }

        {
            UniformVariable& uv = p->uniform("fv3a");
            FloatVec3ArrayUniform& u = dynamic_cast<FloatVec3ArrayUniform&>(uv);
            FloatVec3Array a { { { { 1.0f, 2.0f, 3.0f } },
                                 { { 4.0f, 5.0f, 6.0f } } } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<FloatVec3Array>(u);
            size_test<FloatVec3>(u);
        }

        {
            UniformVariable& uv = p->uniform("fv4a");
            FloatVec4ArrayUniform& u = dynamic_cast<FloatVec4ArrayUniform&>(uv);
            FloatVec4Array a { { { { 1.0f, 2.0f, 3.0f, 4.0f } },
                                 { { 3.0f, 4.0f, 5.0f, 8.0f } } } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<FloatVec4Array>(u);
            size_test<FloatVec4>(u);
        }

        {
            UniformVariable& uv = p->uniform("fm2a");
            FloatMat2ArrayUniform& u = dynamic_cast<FloatMat2ArrayUniform&>(uv);
            FloatMat2Array a { {
                                 { { { { 1.0f, 2.0f } },
                                     { { 3.0f, 4.0f } } } },
                                 { { { { 5.0f, 6.0f } },
                                     { { 7.0f, 8.0f } } } }
                             } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<FloatMat2Array>(u);
            size_test<FloatMat2>(u);
        }

        {
            UniformVariable& uv = p->uniform("fm3a");
            FloatMat3ArrayUniform& u = dynamic_cast<FloatMat3ArrayUniform&>(uv);
            FloatMat3Array a { {
                                 { { { {  1.0f,  2.0f,  3.0f } },
                                     { {  4.0f,  5.0f,  6.0f } },
                                     { {  7.0f,  8.0f,  9.0f } } } },
                                 { { { { 10.0f, 11.0f, 12.0f } },
                                     { { 13.0f, 14.0f, 15.0f } },
                                     { { 16.0f, 17.0f, 18.0f } } } }
                             } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<FloatMat3Array>(u);
            size_test<FloatMat3>(u);
        }

        {
            UniformVariable& uv = p->uniform("fm4a");
            FloatMat4ArrayUniform& u = dynamic_cast<FloatMat4ArrayUniform&>(uv);
            FloatMat4Array a { {
                                 { { { {  1.0f,  2.0f,  3.0f,  4.0f } },
                                     { {  5.0f,  6.0f,  7.0f,  8.0f } },
                                     { {  9.0f, 10.0f, 11.0f, 12.0f } },
                                     { { 13.0f, 14.0f, 15.0f, 16.0f } } } },
                                 { { { { 21.0f, 22.0f, 23.0f, 24.0f } },
                                     { { 25.0f, 26.0f, 27.0f, 28.0f } },
                                     { { 29.0f, 20.0f, 21.0f, 22.0f } },
                                     { { 23.0f, 24.0f, 25.0f, 26.0f } } } },
                             } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<FloatMat4Array>(u);
            size_test<FloatMat4>(u);
        }

        // Integer types

        {
            UniformVariable& uv = p->uniform("i");
            IntUniform& u = dynamic_cast<IntUniform&>(uv);
            u.set(99);
            EXPECT_EQ(99, u.get());

            assign_test<GLint>(u);
        }

        {
            UniformVariable& uv = p->uniform("iv2");
            IntVec2Uniform& u = dynamic_cast<IntVec2Uniform&>(uv);
            u.set({ { 1, 2 } });
            EXPECT_EQ(1, u.get()[0]);
            EXPECT_EQ(2, u.get()[1]);

            u.set(0, 1);                // Separate parameter interface
            int x, y;
            u.get(x, y);
            EXPECT_EQ(0, x);
            EXPECT_EQ(1, y);

            assign_test<IntVec2>(u);
        }

        {
            UniformVariable& uv = p->uniform("iv3");
            IntVec3Uniform& u = dynamic_cast<IntVec3Uniform&>(uv);
            u.set({ { 1, 2, 3 } });
            EXPECT_EQ(1, u.get()[0]);
            EXPECT_EQ(2, u.get()[1]);
            EXPECT_EQ(3, u.get()[2]);

            u.set(0, 1, 2);             // Separate parameter interface
            int x, y, z;
            u.get(x, y, z);
            EXPECT_EQ(0, x);
            EXPECT_EQ(1, y);
            EXPECT_EQ(2, z);

            assign_test<IntVec3>(u);
        }

        {
            UniformVariable& uv = p->uniform("iv4");
            IntVec4Uniform& u = dynamic_cast<IntVec4Uniform&>(uv);
            u.set({ { 1, 2, 3, 4 } });
            EXPECT_EQ(1, u.get()[0]);
            EXPECT_EQ(2, u.get()[1]);
            EXPECT_EQ(3, u.get()[2]);
            EXPECT_EQ(4, u.get()[3]);

            u.set(0, 1, 2, 3);          // Separate parameter interface
            int x, y, z, w;
            u.get(x, y, z, w);
            EXPECT_EQ(0, x);
            EXPECT_EQ(1, y);
            EXPECT_EQ(2, z);
            EXPECT_EQ(3, w);

            assign_test<IntVec4>(u);
        }

        // Int array types

        {
            UniformVariable& uv = p->uniform("ia");
            IntArrayUniform& u = dynamic_cast<IntArrayUniform&>(uv);
            u.set({ { 1, 2 } });
            EXPECT_EQ(1, u.get()[0]);
            EXPECT_EQ(2, u.get()[1]);

            assign_test<IntArray>(u);
            size_test<GLint>(u);
        }

        {
            UniformVariable& uv = p->uniform("iv2a");
            IntVec2ArrayUniform& u = dynamic_cast<IntVec2ArrayUniform&>(uv);
            IntVec2Array a { {
                               { { 1, 2 } },
                               { { 3, 4 } }
                           } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<IntVec2Array>(u);
            size_test<IntVec2>(u);
        }

        {
            UniformVariable& uv = p->uniform("iv3a");
            IntVec3ArrayUniform& u = dynamic_cast<IntVec3ArrayUniform&>(uv);
            IntVec3Array a { {
                               { { 1, 2, 3 } },
                               { { 4, 5, 6 } }
                           } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<IntVec3Array>(u);
            size_test<IntVec3>(u);
        }

        {
            UniformVariable& uv = p->uniform("iv4a");
            IntVec4ArrayUniform& u = dynamic_cast<IntVec4ArrayUniform&>(uv);
            IntVec4Array a { {
                               { { 1, 2, 3, 4 } },
                               { { 5, 6, 7, 8 } }
                           } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<IntVec4Array>(u);
            size_test<IntVec4>(u);
        }

        // Bool types

        {
            UniformVariable& uv = p->uniform("b");
            BoolUniform& u = dynamic_cast<BoolUniform&>(uv);
            u.set(GL_TRUE);
            EXPECT_EQ(GL_TRUE, u.get());
            u.set(GL_FALSE);
            EXPECT_EQ(GL_FALSE, u.get());

            assign_test<GLboolean>(u);
        }

        {
            UniformVariable& uv = p->uniform("bv2");
            BoolVec2Uniform& u = dynamic_cast<BoolVec2Uniform&>(uv);
            u.set({ { GL_TRUE, GL_TRUE } });
            EXPECT_EQ(GL_TRUE, u.get()[0]);
            EXPECT_EQ(GL_TRUE, u.get()[1]);
            u.set({ { GL_FALSE, GL_FALSE } });
            EXPECT_EQ(GL_FALSE, u.get()[0]);
            EXPECT_EQ(GL_FALSE, u.get()[1]);

            u.set(true, false);                         // Separate parameter interface
            GLboolean x, y;
            u.get(x, y);
            EXPECT_EQ(GL_TRUE, x);
            EXPECT_EQ(GL_FALSE, y);

            assign_test<BoolVec2>(u);
        }

        {
            UniformVariable& uv = p->uniform("bv3");
            BoolVec3Uniform& u = dynamic_cast<BoolVec3Uniform&>(uv);
            u.set({ { GL_TRUE, GL_TRUE , GL_TRUE } });
            EXPECT_EQ(GL_TRUE, u.get()[0]);
            EXPECT_EQ(GL_TRUE, u.get()[1]);
            EXPECT_EQ(GL_TRUE, u.get()[2]);
            u.set({ { GL_FALSE, GL_FALSE , GL_FALSE } });
            EXPECT_EQ(GL_FALSE, u.get()[0]);
            EXPECT_EQ(GL_FALSE, u.get()[1]);
            EXPECT_EQ(GL_FALSE, u.get()[2]);

            u.set(true, false, true);                   // Separate parameter interface
            GLboolean x, y, z;
            u.get(x, y, z);
            EXPECT_EQ(GL_TRUE, x);
            EXPECT_EQ(GL_FALSE, y);
            EXPECT_EQ(GL_TRUE, z);

            assign_test<BoolVec3>(u);
        }

        {
            UniformVariable& uv = p->uniform("bv4");
            BoolVec4Uniform& u = dynamic_cast<BoolVec4Uniform&>(uv);
            u.set({ { GL_TRUE, GL_TRUE , GL_TRUE, GL_TRUE } });
            EXPECT_EQ(GL_TRUE, u.get()[0]);
            EXPECT_EQ(GL_TRUE, u.get()[1]);
            EXPECT_EQ(GL_TRUE, u.get()[2]);
            EXPECT_EQ(GL_TRUE, u.get()[3]);
            u.set({ { GL_FALSE, GL_FALSE , GL_FALSE, GL_FALSE } });
            EXPECT_EQ(GL_FALSE, u.get()[0]);
            EXPECT_EQ(GL_FALSE, u.get()[1]);
            EXPECT_EQ(GL_FALSE, u.get()[2]);
            EXPECT_EQ(GL_FALSE, u.get()[3]);

            u.set(true, false, true, true);             // Separate parameter interface
            GLboolean x, y, z, w;
            u.get(x, y, z, w);
            EXPECT_EQ(GL_TRUE, x);
            EXPECT_EQ(GL_FALSE, y);
            EXPECT_EQ(GL_TRUE, z);
            EXPECT_EQ(GL_TRUE, w);

            assign_test<BoolVec4>(u);
        }

        // Bool array types

        {
            UniformVariable& uv = p->uniform("ba");
            BoolArrayUniform& u = dynamic_cast<BoolArrayUniform&>(uv);
            u.set({ { GL_FALSE, GL_TRUE } });
            EXPECT_EQ(GL_FALSE, u.get()[0]);
            EXPECT_EQ(GL_TRUE, u.get()[1]);
            u.set({ { GL_TRUE, GL_FALSE } });
            EXPECT_EQ(GL_TRUE, u.get()[0]);
            EXPECT_EQ(GL_FALSE, u.get()[1]);

            assign_test<BoolArray>(u);
            size_test<GLboolean>(u);
        }

        {
            UniformVariable& uv = p->uniform("bv2a");
            BoolVec2ArrayUniform& u = dynamic_cast<BoolVec2ArrayUniform&>(uv);
            BoolVec2Array a { {
                               { { GL_FALSE, GL_FALSE } },
                               { { GL_FALSE, GL_FALSE } }
                            } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            BoolVec2Array a2 { {
                                { { GL_TRUE, GL_TRUE } },
                                { { GL_TRUE, GL_TRUE } }
                             } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<BoolVec2Array>(u);
            size_test<BoolVec2>(u);
        }

        {
            UniformVariable& uv = p->uniform("bv3a");
            BoolVec3ArrayUniform& u = dynamic_cast<BoolVec3ArrayUniform&>(uv);
            BoolVec3Array a { {
                               { { GL_FALSE, GL_FALSE, GL_FALSE } },
                               { { GL_FALSE, GL_FALSE, GL_FALSE } }
                            } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            BoolVec3Array a2 { {
                                { { GL_TRUE, GL_TRUE, GL_TRUE } },
                                { { GL_TRUE, GL_TRUE, GL_TRUE } }
                             } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<BoolVec3Array>(u);
            size_test<BoolVec3>(u);
        }

        {
            UniformVariable& uv = p->uniform("bv4a");
            BoolVec4ArrayUniform& u = dynamic_cast<BoolVec4ArrayUniform&>(uv);
            BoolVec4Array a { {
                               { { GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE } },
                               { { GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE } }
                            } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            BoolVec4Array a2 { {
                                { { GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE } },
                                { { GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE } }
                             } };
            u.set(a);
            EXPECT_EQ(a, u.get());

            assign_test<BoolVec4Array>(u);
            size_test<BoolVec4>(u);
        }

        // Sampler2D

        {
            UniformVariable& uv = p->uniform("s");
            Sampler2DUniform& u = dynamic_cast<Sampler2DUniform&>(uv);
            u.set(0);
            EXPECT_EQ(0, u.get());

            assign_test<GLint>(u);
        }

        // Sampler2D array

        {
            UniformVariable& uv = p->uniform("sa");
            Sampler2DArrayUniform& u = dynamic_cast<Sampler2DArrayUniform&>(uv);
            u.set({ { 1, 2 } });
            EXPECT_EQ(1, u.get()[0]);
            EXPECT_EQ(2, u.get()[1]);

            assign_test<Sampler2DArray>(u);
            size_test<GLint>(u);
        }

        // SamplerCube

        {
            UniformVariable& uv = p->uniform("c");
            SamplerCubeUniform& u = dynamic_cast<SamplerCubeUniform&>(uv);
            u.set(0);
            EXPECT_EQ(0, u.get());

            assign_test<GLint>(u);
        }

        // SamplerCube array

        {
            UniformVariable& uv = p->uniform("ca");
            SamplerCubeArrayUniform& u = dynamic_cast<SamplerCubeArrayUniform&>(uv);
            u.set({ { 1, 2 } });
            EXPECT_EQ(1, u.get()[0]);
            EXPECT_EQ(2, u.get()[1]);

            assign_test<SamplerCubeArray>(u);
            size_test<GLint>(u);
        }
    }
    catch (const Exception& e)
    {
        FAIL() << e.to_string();
    }
}

TEST(Program, assignment)
{
}
