/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <glint/Runtime.h>
#include <glint/GlintExceptions.h>

#include <boost/regex.hpp>  // BUGFIX: We don't use std::regex because, as of gcc 4.7.2, it is still broken
#include <gtest/gtest.h>

using namespace std;
using namespace glint;

TEST(Runtime, basic)
{
    {
        Runtime::Ptr rt = Runtime::create();
    } // No leak here

    {
        Runtime::Ptr rt = Runtime::create();
        rt->shutdown();
        rt->shutdown(); // OK to shut down twice
    }

    {
        Runtime::Ptr rt = Runtime::create();
        rt->create_display();
    } // No leak here

    {
        Runtime::Ptr rt = Runtime::create();
        rt->create_display(200, 300, 50, 80);
    } // No leak here
}

TEST(Runtime, exceptions)
{
    {
        Runtime::Ptr rt = Runtime::create();

        try
        {
            rt->create_display(-1, 300, 50, 80);
            FAIL();
        }
        catch (const InvalidArgumentException& e)
        {
            EXPECT_EQ("glint::InvalidArgumentException: Invalid window width: -1"
                      " (width must be greater than zero)", e.to_string());
        }

        try
        {
            rt->create_display(200, 0, 50, 80);
            FAIL();
        }
        catch (const InvalidArgumentException& e)
        {
            EXPECT_EQ("glint::InvalidArgumentException: Invalid window height: 0"
                      " (height must be greater than zero)", e.to_string());
        }

        try
        {
            rt->create_display(200, 300, -1, 80);
            FAIL();
        }
        catch (const InvalidArgumentException& e)
        {
            EXPECT_EQ("glint::InvalidArgumentException: Invalid x position: -1"
                      " (x position cannot be negative)" , e.to_string());
        }

        try
        {
            rt->create_display(200, 300, 50, -1);
            FAIL();
        }
        catch (const InvalidArgumentException& e)
        {
            EXPECT_EQ("glint::InvalidArgumentException: Invalid y position: -1"
                      " (y position cannot be negative)" , e.to_string());
        }

        try
        {
            const int width = 100;
            int max_width = rt->max_display_width();
            int xpos = max_width - width + 1;
            rt->create_display(width, 100, xpos, 0);
            FAIL();
        }
        catch (const InvalidArgumentException& e)
        {
            boost::regex r("glint::InvalidArgumentException: Invalid x position: [0-9]+ \\(x position plus"
                           " width \\(100\\) must not exceed display width \\([0-9]+\\)\\)", boost::regex::egrep);
            EXPECT_TRUE(boost::regex_match(e.to_string(), r));
        }

        try
        {
            const int height = 150;
            int max_height = rt->max_display_height();
            int ypos = max_height - height + 1;
            rt->create_display(200, height, 0, ypos);
            FAIL();
        }
        catch (const InvalidArgumentException& e)
        {
            boost::regex r("glint::InvalidArgumentException: Invalid y position: [0-9]+ \\(y position plus"
                           " height \\(150\\) must not exceed display height \\([0-9]+\\)\\)", boost::regex::egrep);
            EXPECT_TRUE(boost::regex_match(e.to_string(), r));
        }

        rt->create_display(200, 300, 50, 80);   // This must work, despite the preceding errors.
    }

    {
        Runtime::Ptr rt = Runtime::create();
        rt->create_display(200, 300, 50, 80);
        try
        {
            rt->create_display(100, 200, 40, 70);
            FAIL();
        }
        catch (const LogicException& e)
        {
            EXPECT_EQ("glint::LogicException: Cannot create more than one display", e.to_string());
        }
    }

    {
        Runtime::Ptr rt = Runtime::create();
        try
        {
            rt->display();
            FAIL();
        }
        catch (const LogicException& e)
        {
            EXPECT_EQ("glint::LogicException: Cannot access display before a display was"
                      " created with create_display()", e.to_string());
        }
        GraphicsDisplay* gd = rt->create_display(200, 300, 50, 80);
        EXPECT_TRUE(gd);
        EXPECT_TRUE(rt->display());
    }
}
