/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <gtest/gtest.h>
#include <glint/genv/egl/EGLException.h>
#include <glint/genv/egl/internal/EGLExceptionImpl.h>

using namespace std;
using namespace glint::genv::egl;

//
// Check EGLException basic functionality (copy, assignment, accessor methods, etc.)
//

TEST(EGLException, basic)
{
    EGLException e("Hello");
    EXPECT_EQ("Hello: egl_error = 0x3000 (EGL_SUCCESS: No error)", e.reason());
    EXPECT_STREQ("glint::genv::egl::EGLException", e.what());
    EXPECT_EQ(0x3000, e.egl_error());
    EXPECT_THROW(rethrow_exception(e.self()), EGLException);

    EGLException e2(e);
    EXPECT_EQ(e2.reason(), e.reason());
    EXPECT_STREQ(e2.what(), e.what());
    EXPECT_EQ(e2.egl_error(), e.egl_error());

    EGLException e3("blah");
    e2 = e3;
    EXPECT_EQ(e2.reason(), e3.reason());
    EXPECT_STREQ(e2.what(), e3.what());
    EXPECT_EQ(e2.egl_error(), e3.egl_error());
}

TEST(EGLException, error_codes)
{
    {
        EGLException e("Hello", 0);
        EXPECT_EQ("Hello: egl_error = 0x0000 (Unknown EGL error code)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3001);
        EXPECT_EQ("Hello: egl_error = 0x3001 (EGL_NOT_INITIALIZED:"
                  " EGL is not or could not be initialized for the specified display.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3002);
        EXPECT_EQ("Hello: egl_error = 0x3002 (EGL_BAD_ACCESS:"
                  " EGL cannot access a requested resource, such as a context bound in another thread.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3003);
        EXPECT_EQ("Hello: egl_error = 0x3003 (EGL_BAD_ALLOC:"
                  " EGL failed to allocate resources for the requested operation.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3004);
        EXPECT_EQ("Hello: egl_error = 0x3004 (EGL_BAD_ATTRIBUTE:"
                  " An unrecognized attribute or attribute value was passed in an attribute list.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3005);
        EXPECT_EQ("Hello: egl_error = 0x3005 (EGL_BAD_CONFIG:"
                  " An EGLConfig argument does not name a valid EGLConfig.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3006);
        EXPECT_EQ("Hello: egl_error = 0x3006 (EGL_BAD_CONTEXT:"
                  " An EGLContext argument does not name a valid EGLContext.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3007);
        EXPECT_EQ("Hello: egl_error = 0x3007 (EGL_BAD_CURRENT_SURFACE:"
                  " The current surface of the calling thread is a window, pbuffer, or pixmap that is no longer valid.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3008);
        EXPECT_EQ("Hello: egl_error = 0x3008 (EGL_BAD_DISPLAY:"
                  " An EGLDisplay argument does not name a valid EGLDisplay.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x3009);
        EXPECT_EQ("Hello: egl_error = 0x3009 (EGL_BAD_MATCH:"
                  " Arguments are inconsistent; for example, an otherwise valid context requires buffers"
                  " (e.g. depth or stencil) not allocated by an otherwise valid surface.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x300a);
        EXPECT_EQ("Hello: egl_error = 0x300a (EGL_BAD_NATIVE_PIXMAP:"
                  " An EGLNativePixmapType argument does not refer to a valid native pixmap.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x300b);
        EXPECT_EQ("Hello: egl_error = 0x300b (EGL_BAD_NATIVE_WINDOW:"
                  " An EGLNativeWindowType argument does not refer to a valid native window.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x300c);
        EXPECT_EQ("Hello: egl_error = 0x300c (EGL_BAD_PARAMETER:"
                  " One or more argument values are invalid.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x300d);
        EXPECT_EQ("Hello: egl_error = 0x300d (EGL_BAD_SURFACE:"
                  " An EGLSurface argument does not name a valid surface (window, pbuffer, or pixmap)"
                  " configured for rendering.)",
                  e.reason());
    }
    {
        EGLException e("Hello", 0x300e);
        EXPECT_EQ("Hello: egl_error = 0x300e (EGL_CONTEXT_LOST:"
                  " A power management event has occurred. This application must destroy all contexts"
                  " and reinitialize client API state and objects to continue rendering.)",
                  e.reason());
    }
}

// Dynamic allocataion to get around bogus function coverage reports by gcov.

TEST(EGLException, dynamic)
{
    EGLException* ep = new EGLException("Hello");
    delete ep;
    internal::EGLExceptionImpl* eip = new internal::EGLExceptionImpl("Hello", 1);
    delete eip;
}
