/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <gtest/gtest.h>
#include <glint/genv/glx/internal/GLXGraphicsDisplay.h>
#include <glint/genv/glx/internal/GLXGEnv.h>
#include <glint/genv/x11/X11Window.h>
#include <glint/Exception.h>

#include <boost/thread.hpp>
#include <GL/gl.h>

#include <iostream>

using namespace std;
using namespace glint;
using namespace glint::genv::internal;
using namespace glint::genv::glx::internal;
using namespace glint::genv::x11;

TEST(GLXGraphicsDisplay, basic)
{
    {
        AbstractGEnv::UPtr genv = create_concrete_genv();
        EXPECT_TRUE(genv.get());

        GraphicsDisplay* gd = genv->create_display(100, 110, 5, 10);
        EXPECT_TRUE(gd);

        EXPECT_EQ(100, gd->width());
        EXPECT_EQ(110, gd->height());
        EXPECT_EQ(5, gd->xpos());
        EXPECT_EQ(10, gd->ypos());

        GLXGraphicsDisplay* egd = dynamic_cast<GLXGraphicsDisplay*>(gd);
        EXPECT_TRUE(egd);

        AbstractWindow* aw = gd->window();
        EXPECT_TRUE(aw);

        X11Window* xwin = dynamic_cast<X11Window*>(aw);
        EXPECT_TRUE(xwin);

        EXPECT_TRUE(xwin->display());
        EXPECT_NE(0, xwin->window());

        // Called so we get coverage; no actual test here
        //
        EXPECT_NE(0, gd->gl_major_version());
        gd->gl_minor_version();
    }
}

namespace
{

//
// Roundabout way of testing whether the current thread holds the context.
// glGetError() returns GL_NO_ERROR whenever it is called without a
// current context, but it does return an error for an incorrect API
// call when there *is* a current context. So, if we make a non-sensical
// API call *without* a current context, glGetError() returns GL_NO_ERROR,
// proving that we indeed don't have a context. If we make a non-sensical API
// call *with* a current context, glGetError() returns a real error,
// proving that the context was indeed current.
//

bool have_current()
{
    GLboolean b[1];
    glGetBooleanv(-1, b);           // Non-existent enum value
    return glGetError() != GL_NO_ERROR;
}

} // namespace

TEST(GLXGraphicsDisplay, current)
{
    AbstractGEnv::UPtr genv = create_concrete_genv();
    GraphicsDisplay* gd = genv->create_display(200, 200, 0, 0);

    gd->release_current();
    gd->release_current();  // OK to call twice, second call is a no-op
    EXPECT_FALSE(have_current());

    gd->make_current();
    gd->make_current();     // OK to call twice, second call is a no-op
    EXPECT_TRUE(have_current());
}

namespace
{

enum ThreadType { left, right };

class CurrentTester
{
public:
    CurrentTester(GraphicsDisplay* display, ThreadType type)
        : gd_(display)
        , type_(type)
    {
    }

    void operator()()
    {
        while (count_ > 0)
        {
            boost::unique_lock<boost::mutex> lock(m_);

            // Force the two threads to take turns calling make_current()/release_current()
            //
            if (type_ == left)
            {
                while (count_ % 2 != 0)
                {
                    cond_.wait(lock);
                }
            }
            else
            {
                while (count_ % 2 == 0)
                {
                    cond_.wait(lock);
                }
            }

            try
            {
                gd_->make_current();
                EXPECT_TRUE(have_current());
                gd_->release_current();
                EXPECT_FALSE(have_current());
            }
            catch (const glint::Exception& e)
            {
                cerr << e.to_string() << endl;
                abort();
            }
            --count_;
            cond_.notify_all();
            boost::this_thread::yield();
        }
    }

private:
    GraphicsDisplay* gd_;
    static boost::condition_variable cond_;
    static boost::mutex m_;
    static int count_;
    ThreadType type_;
};

boost::condition_variable CurrentTester::cond_;
boost::mutex CurrentTester::m_;
int CurrentTester::count_ = 20;

}

TEST(GLXGraphicsDisplay, pass_current)
{
    AbstractGEnv::UPtr genv = create_concrete_genv();
    GraphicsDisplay* gd = genv->create_display(200, 200, 0, 0);
    gd->release_current();

    // Start two threads that repeatedly pass the current back and forth.
    //
    boost::thread t1 = CurrentTester(gd, ThreadType::left);
    boost::thread t2 = CurrentTester(gd, ThreadType::right);

    t1.join();
    t2.join();
}

namespace
{

#if GLINT_MANUAL_TESTS
void sleep(int seconds)
{
    ::sleep(seconds);
}
#else
void sleep(int)
{
}
#endif

} // namespace

//
// It appears to be impossible to programmatically test whether swap_buffers() works
// because exchaning the front and back buffer leaves the new back buffer in an
// undefined state. In practice, this means that, if we make a black front buffer,
// swap, make a white front buffer, and swap again, on the next swap, the previously
// black front buffer (which was moved to the back and then to the front again), is
// suddently white.
//
// The code below allows visual inspection of the buffer swapping, by slowing things
// down enough to follow.
//

TEST(GLXGraphicsDisplay, swap_buffers)
{
    AbstractGEnv::UPtr genv = create_concrete_genv();
    GraphicsDisplay* gd = genv->create_display(200, 200, 0, 0);

    GLint pixel;

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    cout << "Showing black in one second" << endl;
    sleep(1);
    gd->swap_buffers();
    pixel = 0;
    glReadPixels(0, 0, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, &pixel);
    EXPECT_EQ(0, pixel);

    sleep(3);

    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    cout << "Showing white in one second" << endl;
    sleep(1);
    gd->swap_buffers();
    pixel = 0;
    glReadPixels(0, 0, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, &pixel);
    EXPECT_EQ(0xffffff, pixel);

    sleep(3);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    cout << "Showing black in one second" << endl;
    sleep(1);
    gd->swap_buffers();
    pixel = 0;
    glReadPixels(0, 0, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, &pixel);
    EXPECT_EQ(0, pixel);

    sleep(3);

    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    cout << "Showing white in one second" << endl;
    sleep(1);
    gd->swap_buffers();
    pixel = 0;
    glReadPixels(0, 0, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, &pixel);
    EXPECT_EQ(0xffffff, pixel);

    sleep(3);
}

TEST(GLXGraphicsDisplay, visibility)
{
    AbstractGEnv::UPtr genv = create_concrete_genv();
    GraphicsDisplay* gd = genv->create_display(200, 200, 0, 0);

    EXPECT_EQ(GraphicsDisplay::Vis::invisible, gd->visibility());

    EXPECT_EQ(GraphicsDisplay::Vis::invisible, gd->set_visibility(GraphicsDisplay::Vis::visible));
    EXPECT_EQ(GraphicsDisplay::Vis::visible, gd->visibility());

    EXPECT_EQ(GraphicsDisplay::Vis::visible, gd->set_visibility(GraphicsDisplay::Vis::visible));
    EXPECT_EQ(GraphicsDisplay::Vis::visible, gd->visibility());

    EXPECT_EQ(GraphicsDisplay::Vis::visible, gd->set_visibility(GraphicsDisplay::Vis::invisible));
    EXPECT_EQ(GraphicsDisplay::Vis::invisible, gd->visibility());

    EXPECT_EQ(GraphicsDisplay::Vis::invisible, gd->set_visibility(GraphicsDisplay::Vis::invisible));
    EXPECT_EQ(GraphicsDisplay::Vis::invisible, gd->visibility());
}

TEST(GLXGraphicsDisplay, stacking_order)
{
    AbstractGEnv::UPtr genv = create_concrete_genv();
    GraphicsDisplay* gd = genv->create_display(200, 200, 0, 0);

    // No test here, but we exercise the call for coverage analysis.
    gd->set_stacking_order(GraphicsDisplay::StackingOrder::bottom);
    gd->set_stacking_order(GraphicsDisplay::StackingOrder::top);
}
