/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <gtest/gtest.h>
#include <glint/genv/x11/X11Exception.h>
#include <glint/genv/x11/internal/X11WEnv.h>
#include <glint/genv/x11/X11Window.h>
#include <glint/genv/NativeDefs.h>

#ifdef GLINT_GLX
#include <GL/glx.h>
#else
#include <glint/genv/egl/EGLException.h>
#include <EGL/egl.h>
#endif

using namespace std;
using namespace glint;
using namespace glint::genv::internal;
using namespace glint::genv::x11;
using namespace glint::genv::x11::internal;

TEST(X11WEnv, basic)
{
    {
        X11WEnv env;  // Leak test, destructor should clean up.
    }

    {
        X11WEnv env;
        env.shutdown();
        env.shutdown(); // OK to shutdown twice
    }
}

namespace
{

class XVisual
{
public:
    int visual_id_;

#ifdef GLINT_GLX

    XVisual(X11WEnv& env)
        : visual_id_(-1)
        , vi_(nullptr, XFree)
    {
        int attr[] =
        {
            GLX_RGBA,
            GLX_DOUBLEBUFFER,
            GLX_RED_SIZE,     8,
            GLX_GREEN_SIZE,   8,
            GLX_BLUE_SIZE,    8,
            GLX_ALPHA_SIZE,   8,
            GLX_DEPTH_SIZE,   24,
            GLX_STENCIL_SIZE, 8,
            None
        };
        vi_.reset(glXChooseVisual(env.display(), XDefaultScreen(env.display()), attr));
        EXPECT_TRUE(vi_.get());
        visual_id_ = vi_->visualid;
    }

private:
    unique_ptr<
    XVisualInfo,
    decltype(&XFree)
    > vi_;

#else

    XVisual(X11WEnv& /* env */)
        : visual_id_(-1)
        , egl_display_([this](EGLDisplay d) { this->terminate_egl(d); })
    {
        EGLint const attr[] = {
            EGL_RED_SIZE,     8,
            EGL_GREEN_SIZE,   8,
            EGL_BLUE_SIZE,    8,
            EGL_ALPHA_SIZE,   8,
            EGL_DEPTH_SIZE,  24,
            EGL_STENCIL_SIZE, 8,
            EGL_RENDERABLE_TYPE,
            EGL_OPENGL_ES2_BIT,
            EGL_NONE
        };

        egl_display_.reset(eglGetDisplay(EGL_DEFAULT_DISPLAY));
        if (egl_display_.get() == EGL_NO_DISPLAY)
        {
            throw glint::genv::egl::EGLException("eglGetDisplay() failed");
        }

        EGLint egl_major;
        EGLint egl_minor;
        if (!eglInitialize(egl_display_.get(), &egl_major, &egl_minor))
        {
            throw glint::genv::egl::EGLException("eglInitialize() failed");
        }

        EGLConfig egl_config;
        EGLint num_config;
        if (!eglChooseConfig(egl_display_.get(), attr, &egl_config, 1, &num_config))
        {
            throw glint::genv::egl::EGLException("eglChooseConfig() failed");
        }

        if (!eglGetConfigAttrib(egl_display_.get(), egl_config, EGL_NATIVE_VISUAL_ID, &visual_id_))
        {
            throw glint::genv::egl::EGLException("eglGetConfigAttrib() failed");
        }
    }

    void terminate_egl(EGLDisplay d)
    {
        if (!eglTerminate(d))
        {
            throw glint::genv::egl::EGLException("eglTerminate() failed");
        }
    }

private:
    util::ResourcePtr<EGLDisplay, std::function<void(EGLDisplay)>> egl_display_;
#endif
};

} // namespace

TEST(X11WEnv, window)
{
    {
        X11WEnv env;

        EXPECT_TRUE(env.display());

        XVisual xv(env);
        env.create_window(xv.visual_id_, env.max_display_width(), env.max_display_height(), 0, 0);

        EXPECT_EQ(env.max_display_width(), env.width());
        EXPECT_EQ(env.max_display_height(), env.height());
        EXPECT_EQ(0, env.xpos());
        EXPECT_EQ(0, env.ypos());
    }

    {
        X11WEnv env;

        EXPECT_TRUE(env.display());

        XVisual xv(env);
        AbstractWindow* win = env.create_window(xv.visual_id_, 100, 200, 50, 80);

        EXPECT_EQ(100, env.width());
        EXPECT_EQ(200, env.height());
        EXPECT_EQ(50, env.xpos());
        EXPECT_EQ(80, env.ypos());

        EXPECT_EQ(win, env.window());

        X11Window* xwin = dynamic_cast<X11Window*>(win);
        EXPECT_TRUE(xwin);
        EXPECT_EQ(env.display(), xwin->display());
        EXPECT_NE(0, xwin->window());
    }
}
