/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michi Henning <michi.henning@canonical.com>
 */

#include <gtest/gtest.h>
#include <glint/GlintExceptions.h>
#include <glint/util/ResourcePtr.h>

#include <set>

using namespace std;
using namespace glint;
using namespace glint::util;

namespace
{

std::set<int*> allocated; // Tracks allocated resources

}

//
// Fake allocation and deallocation functions.
//

int* alloc_int(int p)
{
    int* pointer = reinterpret_cast<int*>(p);
    allocated.insert(pointer);
    return pointer;
}

int dealloc_int(int* i)
{
    auto it = allocated.find(i);
    EXPECT_NE(it, allocated.end());
    allocated.erase(it);
    return 0;
}

//
// Check that allocator and deallocator are called with the correct argument
// and that get() returns the correct value.
//

TEST(ResourcePtr, basic)
{
    {
        ResourcePtr<int*, decltype(&dealloc_int)> rp(dealloc_int);
        EXPECT_FALSE(rp.has_resource());
    }

    {
        ResourcePtr<int*, decltype(&dealloc_int)> rp(alloc_int(99), dealloc_int);
        EXPECT_EQ(allocated.size(), 1);
        EXPECT_NE(allocated.find(rp.get()), allocated.end());
        EXPECT_TRUE(rp.has_resource());
    }
    EXPECT_TRUE(allocated.empty());
}

//
// A deallocator that expects two arguments. We use std::bind to call it as if it expected only one.
//

int dealloc_string_int(const string s, int* i)
{
    EXPECT_EQ(s, "Hello");
    dealloc_int(i);
    return 0;
}

//
// Check that calling with bind works as expected.
//

TEST(ResourcePtr, bind)
{
    {
        auto call_dealloc = std::bind(&dealloc_string_int, "Hello", std::placeholders::_1);
        ResourcePtr<int*, decltype(call_dealloc)> rp(alloc_int(42), call_dealloc);
        EXPECT_EQ(allocated.size(), 1);
        EXPECT_NE(allocated.find(rp.get()), allocated.end());
    }
    EXPECT_TRUE(allocated.empty());
}

//
// Check that reset assigns the new pointer value.
//

TEST(ResourcePtr, reset)
{
    ResourcePtr<int*, decltype(&dealloc_int)> rp(alloc_int(77), dealloc_int);
    EXPECT_EQ(allocated.size(), 1);
    EXPECT_NE(allocated.find(rp.get()), allocated.end());

    rp.reset(alloc_int(33));
    EXPECT_EQ(allocated.size(), 1);
    EXPECT_EQ(allocated.find((int*)77), allocated.end());
    EXPECT_NE(allocated.find((int*)33), allocated.end());

    rp.reset(alloc_int(10));
    rp.reset(alloc_int(20));    // Will deallocate the previous resource, so this will not cause a leak.
}

//
// Check that dealloc deallocates and can be called multiple times without ill effects.
//

TEST(ResourcePtr, dealloc)
{
    ResourcePtr<int*, decltype(&dealloc_int)> rp(dealloc_int);

    rp.reset(alloc_int(77));
    EXPECT_EQ(allocated.size(), 1);
    EXPECT_NE(allocated.find(rp.get()), allocated.end());

    rp.dealloc();
    EXPECT_EQ(allocated.size(), 0);
    EXPECT_FALSE(rp.has_resource());

    rp.dealloc();
    EXPECT_EQ(allocated.size(), 0);
    EXPECT_FALSE(rp.has_resource());

    rp.reset(alloc_int(33));
    EXPECT_EQ(allocated.size(), 1);
    EXPECT_NE(allocated.find((int*)33), allocated.end());
    EXPECT_TRUE(rp.has_resource());

    rp.dealloc();
    EXPECT_EQ(allocated.size(), 0);
    EXPECT_FALSE(rp.has_resource());

    rp.dealloc();
    EXPECT_EQ(allocated.size(), 0);
    EXPECT_FALSE(rp.has_resource());
}

int dealloc_throw(int* i)
{
    auto it = allocated.find(i);
    EXPECT_NE(it, allocated.end());
    allocated.erase(it);
    throw 99;
}

//
// Test that things work correctly if the deleter throws.
//

TEST(ResourcePtr, dealloc_throw)
{
    {
        ResourcePtr<int*, decltype(&dealloc_throw)> rp(dealloc_throw);
        rp.reset(alloc_int(77));

        try
        {
            rp.dealloc();
        }
        catch (int i)
        {
            EXPECT_EQ(99, i);
            EXPECT_FALSE(rp.has_resource());
        }
    }

    {
        // Same test again, but this time by causing the deallocation via reset().
        ResourcePtr<int*, decltype(&dealloc_throw)> rp(dealloc_throw);
        rp.reset(alloc_int(77));

        try
        {
            rp.reset(alloc_int(66));
        }
        catch (int i)
        {
            EXPECT_EQ(99, i);
            EXPECT_TRUE(rp.has_resource()); // True because rp now owns the new resource
        }
    }

    {
        // Same test again, but this time by letting the destructor do
        // the deallocation.
        try
        {
            ResourcePtr<int*, decltype(&dealloc_throw)> rp(dealloc_throw);
            rp.reset(alloc_int(77));
        }
        catch (...)
        {
            FAIL();
        }
    }
}

//
// get() without a resource throws.
//

TEST(ResourcePtr, get_throw)
{
    ResourcePtr<int*, decltype(&dealloc_int)> rp(dealloc_int);

    try
    {
        rp.get();
        FAIL();
    }
    catch (const std::logic_error& e)
    {
        EXPECT_STREQ("get() called on ResourcePtr without resource", e.what());
    }

    rp.reset(alloc_int(99));
    rp.dealloc();

    try
    {
        rp.get();
        FAIL();
    }
    catch (const std::logic_error& e)
    {
        EXPECT_STREQ("get() called on ResourcePtr without resource", e.what());
    }
}

//
// Check that release() releases the resource and throws if there is no resource to release.
//

TEST(ResourcePtr, release)
{
    ResourcePtr<int*, decltype(&dealloc_int)> rp(dealloc_int);

    rp.reset(alloc_int(99));
    EXPECT_TRUE(rp.has_resource());
    rp.release();
    EXPECT_FALSE(rp.has_resource());

    try
    {
        rp.release();
        FAIL();
    }
    catch (const std::logic_error& e)
    {
        EXPECT_STREQ("release() called on ResourcePtr without resource", e.what());
    }
}
